//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import sdms.impl.Factory;
import sdms.impl.VA27_hack;

final class SolutionWithViolation {
    
    final double viol;
    final Solution s;
    
    public SolutionWithViolation(Solution s, double viol) {
        this.s = s; 
        this.viol = viol;
    }
    
    static final class CmpByViol implements Comparator<SolutionWithViolation> {
        @Override
        public int compare(SolutionWithViolation o1, SolutionWithViolation o2) {
            return Double.compare(o1.viol, o2.viol);
        }
    }
}

// TODO The code is before a major refactoring: all the funny looking static
//      functions should be moved into a separate class;
//      *all* creation should  be moved into the Factory class;
//      code segments between //---- should be moved to separate functions

public class Algorithm {

    private static final Logger log = LogManager.getLogger();
    
    private boolean FULL_PROBLEM_ONLY;
    private boolean SHOW_CONDITION_NUMBER;
    private int POINTS_TO_PROPAGATE;
    private int POINTS_PER_FRONT;
    private double INFEASIBILITY_THRESHOLD;
    
    private double INFEAS_REG; // seems to be ignored, used only to print statistics
    
    private int FIXED_LAG;
    
    private int EXTENSION_ATTEMPTS;
    private int EXPECTED_OUTPUT;
    private double BOUND_RELAX_FACTOR;
    private ReparamOpts reparamOpts;
    
    private String SUBPROB_LOG;
    private String FULLPROB_LOG;
 
    private NLPProblem problem;  
    private SubNLPProblem subProblem;
    private SubproblemSelector selector;
    private StartingpointGenerator startingpointGenerator;
    private Solver solver;
    private FullProblemSolving fullProblemSolution; 
    private Options options;
    private Plotter plotter;
    
    private Random rnd;
    
    private Solution currentPartialSolution;
    private Solution solution;
    private ArrayList<Solution> allPartialSolutions;
    private ArrayList<Solution> extendedPartialSolutions;

    private void initialize(String file) {

        POINTS_TO_PROPAGATE = options.get(IntOptKey.POINTS_TO_PROPAGATE);
        POINTS_PER_FRONT    = options.get(IntOptKey.POINTS_PER_FRONT);
        FIXED_LAG           = options.get(IntOptKey.FIXED_LAG);
        
        EXTENSION_ATTEMPTS = options.get(IntOptKey.EXTENSION_ATTEMPTS);
        EXPECTED_OUTPUT    = options.get(IntOptKey.EXPECTED_OUTPUT);
        BOUND_RELAX_FACTOR = options.get(DblOptKey.BOUND_RELAX_FACTOR);
        
        reparamOpts = new ReparamOpts(POINTS_TO_PROPAGATE, EXTENSION_ATTEMPTS, EXPECTED_OUTPUT, BOUND_RELAX_FACTOR);        
        
        FULL_PROBLEM_ONLY = options.get(StrOptKey.FULL_PROBLEM_ONLY).equals("yes");
        SHOW_CONDITION_NUMBER = options.get(StrOptKey.SHOW_CONDITION_NUMBER).equals("yes");

        INFEASIBILITY_THRESHOLD  = options.get(DblOptKey.INFEASIBILITY_THRESHOLD);
        INFEAS_REG   = options.get(DblOptKey.INFEAS_REG);

        
        SUBPROB_LOG  = options.get(StrOptKey.SUBPROB_LOG);
        FULLPROB_LOG = options.get(StrOptKey.FULLPROB_LOG);
        
        // Maybe pack them into one object or the Factory should set them?
        // Or make a polymorphic factory, and get the factory from a factory method receiving args?
        // Or call the factory from main and use dependency injection?
        
        String scale = options.get(StrOptKey.SCALING);
        
        problem = Factory.createNLPProblem(file, scale);
        selector = Factory.createSubproblemSelector(file);
        startingpointGenerator = Factory.createStartingpointGenerator();
        solver = createSubproblemSolver();
        fullProblemSolution = createFullproblemSolvingLogic(options.get(StrOptKey.FULLPROBLEM_SOLVER));
        plotter = Factory.createPlotter(file, options.get(StrOptKey.PLOT));
        
        rnd = new Random(31L);

        allPartialSolutions = new ArrayList<>();
        extendedPartialSolutions = new ArrayList<>();
    }
   
    private Solver createSubproblemSolver() {
        
        if (FULL_PROBLEM_ONLY) // Sign of a design flaw
            return null;
        
        return Factory.createSolver(options.get(StrOptKey.SUBPROBLEM_SOLVER));
    }

    private Solver createFullproblemSolver() {
        
        return Factory.createSolver(options.get(StrOptKey.FULLPROBLEM_SOLVER));
    }
      
    public void run(Options opts, String file) {

        this.options = opts;

        initialize(file);

        if (FULL_PROBLEM_ONLY) {

            runOnFullProblemOnly();
        }
        else {

            runOnSubprobs();

            ArrayList<Solution> startingPoints = new ArrayList<>(allPartialSolutions);

            runFullProb();
            
            // 1. allPartialSolutions contain unsuccessful solutions at this point.
            // 2. They have been reordered (sorted by constraint violation) in case of Newton's method.
            
            dumpThoseNearestToReferenceSols(startingPoints);
            
            removeUnsuccessfulSols();
        }
        
        writeSolutions();
        
        showConditionNumbers();
        
        plotter.writeReferenceSolutions(problem.getReferenceSolutions());
        
        plotter.done(SUBPROB_LOG, FULLPROB_LOG); // log file names needed for the cleanup script
    }
    
    private void runOnFullProblemOnly() {
        
        allPartialSolutions.addAll(Factory.createEmptyPartialSolutions(POINTS_TO_PROPAGATE));
        
        structureUnawareMultistartOnFullProblem();
        
        dumpFullProblemComputationalEffortStatistics();        
    }
    
    private void runOnSubprobs() {
        
        long start = System.nanoTime();
        
        // Implement strategies and add a factory method
        // to set the strategy on construction?            
        initializeFirstVariables();

        fixedLagIteration(); // iterateThroughSubproblems();
        
        long end = System.nanoTime();
        
        log.info("Generating starting points took: "+Util.toTimeStamp(end-start));
        
        log.info("Dumping computational effort statistics for subproblems");
        
        solver.dumpStatistics(SUBPROB_LOG);
    }
    
    private void runFullProb() {
        
        long startMS = System.nanoTime();

        fullProblemSolution.run();
        
        long endMS = System.nanoTime();
        
        log.info("Solving from each point took: "+Util.toTimeStamp(endMS-startMS));
        
        fullProblemSolution.dumpStatistics();        
    }
    
    private void removeUnsuccessfulSols() {
        
        for (Solution s : allPartialSolutions) {
            
            if (s.isSuccessful()) {
                
                extendedPartialSolutions.add(s);
            }
        }
        
        allPartialSolutions = extendedPartialSolutions;
        
        extendedPartialSolutions = new ArrayList<>();
    }
    
    private interface FullProblemSolving { 
        
        void run();
        
        void dumpStatistics();
    }
    
    private FullProblemSolving createFullproblemSolvingLogic(String option) {
        switch (option) {
        case "newton":
            return new FullProblemSolving() {
                @Override public void run() {  runNewtonFromAllPartialSolutions(); } 
                @Override public void dumpStatistics() { }
            };
        case "ipopt":
        case "lmbopt":
            return new FullProblemSolving() {
                @Override public void run() { runOnFullProblemFromAllPartialSolutions(); }
                @Override public void dumpStatistics() { dumpFullProblemComputationalEffortStatistics(); }
            };
        case "va27":
            return new FullProblemSolving() {
                @Override public void run() { run_va27(); }
                @Override public void dumpStatistics() { }
            };
        case "lmbopt_newton":
            return new FullProblemSolving() {
                @Override public void run() { run_lmbopt_newton(); }
                @Override public void dumpStatistics() { }
            };
        default:
            throw new RuntimeException("Unrecognized option for "+StrOptKey.FULLPROBLEM_SOLVER+": \""+option+"\"");       
        }
    }
    
    private void dumpFullProblemComputationalEffortStatistics() {

        log.info("Dumping computational effort statistics for full problems");
        
        solver.dumpStatistics(FULLPROB_LOG);
    }

    private void initializeFirstVariables() {

        TreeSet<String> firstVars = selector.getFirstVariables();
        
        if (firstVars.isEmpty()) {
            
            log.info("Starting with a completely random assigment");
            
            allPartialSolutions.addAll(Factory.createEmptyPartialSolutions(POINTS_TO_PROPAGATE));
        }
        else {
            throw new AssertionError("Used to be gridstart");
        }
    }
    
    private static final class ReparamOpts {
        final int POINTS_TO_PROPAGATE;
        final int EXTENSION_ATTEMPTS;  
        final int EXPECTED_OUTPUT;
        final double BOUND_RELAX_FACTOR; 
        public ReparamOpts(int POINTS_TO_PROPAGATE, int EXTENSION_ATTEMPTS, int EXPECTED_OUTPUT, double BOUND_RELAX_FACTOR) {
            this.POINTS_TO_PROPAGATE=POINTS_TO_PROPAGATE; 
            this.EXTENSION_ATTEMPTS =EXTENSION_ATTEMPTS; 
            this.EXPECTED_OUTPUT    =EXPECTED_OUTPUT; 
            this.BOUND_RELAX_FACTOR =BOUND_RELAX_FACTOR;
        }
    }

    private void fixedLagIteration() {
        
        Util.printLocation();
        
        System.out.println("Running with max step length: "+FIXED_LAG);
        
        int dbgSubproblemIndex = 0;
        
        while (selector.setNext()) {

            System.out.println("Started with " + allPartialSolutions.size() + " partial solutions");
            
            ++dbgSubproblemIndex;
            
            dbgShowInfeasibilites(allPartialSolutions, dbgSubproblemIndex);
            
            //--------------------------------------------------------------------------------------------
            
            TreeSet<String> xVars = selector.getFixedVars();
            
            TreeSet<String> yVars = selector.getVariablesToGuess();
            
            TreeSet<String> allSelectedVars = new TreeSet<String>(xVars);
            
            allSelectedVars.addAll(yVars);
            
            TreeSet<String> constraints = selector.getNextConstraints();
            
            // TODO Disable consistency checking in subProblem generation (NLPProblem.checkConsistency()) when ready
                                                                   // These 2 args are only for debugging 
            subProblem = problem.generateSubOptProblem(constraints, allSelectedVars, selector.dbgFullProbSparsity());

            TreeMap<String,Double> rowScales = problem.rowScales(constraints);
            
            TreeMap<String,Double> columnScales = problem.colScales(allSelectedVars);

            extendedPartialSolutions = Reparameterization.forwardProjection(subProblem, 
                                                                            solver, 
                                                                            allPartialSolutions, 
                                                                            xVars, 
                                                                            yVars, 
                                                                            rowScales, 
                                                                            columnScales, 
                                                                            POINTS_TO_PROPAGATE,
                                                                            EXTENSION_ATTEMPTS,
                                                                            EXPECTED_OUTPUT,
                                                                            BOUND_RELAX_FACTOR,
                                                                            rnd,
                                                                            dbgSubproblemIndex,
                                                                            selector.isLastSubproblem());
            
            plotter.writePartialSolutions(extendedPartialSolutions);
    
            //------------------------------------------------------------------------------------------------
            
            if (!xVars.isEmpty() && !selector.isLastSubproblem() /*&& dbgSubproblemIndex < 2*/) { // HACK
                
                extendedPartialSolutions.addAll( reparameterize(allPartialSolutions,
                                                                extendedPartialSolutions,
                                                                selector, 
                                                                FIXED_LAG, 
                                                                problem, 
                                                                solver, 
                                                                rnd, 
                                                                reparamOpts,
                                                                -dbgSubproblemIndex )
                                               );
            }
            
            //------------------------------------------------------------------------------------------------
            
            TreeSet<String> consSoFar = selector.getProcessedConstraintsSoFar();
            
            TreeSet<String> varsSoFar = new TreeSet<>(extendedPartialSolutions.get(0).getVarNames());
            
            SubNLPProblem processed = problem.generateSubOptProblem(consSoFar, varsSoFar, selector.dbgFullProbSparsity());
            
            TreeMap<String,Double> conScales = problem.rowScales(consSoFar);
            
            TreeMap<String,Double> varScales = problem.colScales(varsSoFar);
 
            //------------------------------------------------------------------------------------------------
            
            assert !extendedPartialSolutions.isEmpty() : "Lost all partial solutions?";
     
            ArrayList<Solution> tmp = new ArrayList<>(extendedPartialSolutions.size());
            
            ArrayList<SolutionWithViolation> solWithViol = new ArrayList<>(extendedPartialSolutions.size());

            for (Solution s : extendedPartialSolutions) {

                Violations viol = processed.allViolations(s.getAllVars(), conScales, varScales);
                
                // HACK The true solution would be to replace the interval evaluator with a 
                if (viol.hasNans()) {
                    log.error("NaN detected in F(x), where x = "+s.getAllVars());
                    continue;
                }
                
                s.setAllViolations(viol);
                
                solWithViol.add(new SolutionWithViolation(s, viol.LinfConsAndBound));
                
                //@SuppressWarnings("unused")
                double delta = 1.0d/POINTS_TO_PROPAGATE;

                if ((viol.L2SquaredConsAndBound<1e+300) && (viol.LinfBoundOnly<2.0*delta) && (viol.LinfConsAndBound<2.0*delta)) { // magic number on tolerated bound violation
                    // against infs and NaNs  ^^^ 
                    tmp.add(s);
                }
            }           
            
            assert !tmp.isEmpty() || selector.isLastSubproblem() : "Lost all partial solutions?";
            
            int DESIRED_POINTS = POINTS_TO_PROPAGATE + (dbgSubproblemIndex-1)*POINTS_PER_FRONT;
            
            if (tmp.size()<DESIRED_POINTS && selector.isLastSubproblem()) {
                
                extendedPartialSolutions = tryToRecover(solWithViol, DESIRED_POINTS);
            }
            else {
            
                extendedPartialSolutions = tmp;
            }
            
            tmp = null;
          
            //-------------------------------------------------------------------------------------------------
            
            plotThenDiscardSuperfluousCandidates(DESIRED_POINTS);

            //------------------------------------------------------------------------------------------------
            
            log.info("Finished with " + extendedPartialSolutions.size() + " partial solutions");

            allPartialSolutions = extendedPartialSolutions;

            extendedPartialSolutions = new ArrayList<>();
                        
            if (allPartialSolutions.isEmpty()) {
                
                log.warn("Lost all partial solutions");
                
                return;
            }
        }
        
        dbgShowInfeasibilites(allPartialSolutions, dbgSubproblemIndex);
    }
    
    static private void dbgShowInfeasibilites(ArrayList<Solution> allPartialSolutions, int dbgSubproblemIndex) {
        
        if (dbgSubproblemIndex==1) // No cumulative infeasibility set on the solution at the first subproblem
            return;
        
        double[] infNormConsAndBnds = new double[allPartialSolutions.size()];
        
        int k=0;
        
        for (Solution s : allPartialSolutions) {
            
            infNormConsAndBnds[k] = s.getInfeasibilityInfinityNorm();
            
            ++k;
        }
        
        Util.printLocation();
        
        System.out.println("At subproblem "+dbgSubproblemIndex);
        
        System.out.println("Scaled constraint and bound violation distribution, infinity norm");
        
        Util.sort_and_show_statistics(infNormConsAndBnds);
    }
    
    static private ArrayList<Solution> tryToRecover(ArrayList<SolutionWithViolation> solWithViol, int POINTS_TO_PROPAGATE) {
        
        Util.printLocation();
      
        Collections.sort(solWithViol, new SolutionWithViolation.CmpByViol());
        
        ArrayList<Solution> kept = new ArrayList<>(POINTS_TO_PROPAGATE);

        for (int i=0, n=Math.min(solWithViol.size(),POINTS_TO_PROPAGATE); i<n; ++i) {
            
            kept.add( solWithViol.get(i).s );
        }
        
        return kept;
    }
    
    static private ArrayList<Solution> reparameterize(ArrayList<Solution> allPartialSolutions, 
                                                      ArrayList<Solution> alreadyFwdProjected,
                                                      SubproblemSelector selector, 
                                                      int maxStepLength, 
                                                      NLPProblem problem, 
                                                      Solver solver, 
                                                      Random rnd,
                                                      ReparamOpts reparamOpts,
                                                      int dbgSubproblemIndex) 
    {
            System.out.println("Projecting back");
            
            TreeSet<String> consInBwd  = selector.allConsBetweenCurrentAndPast(maxStepLength);
            
            TreeSet<String> xVarsOrig = selector.hasPrecedingSubproblem(maxStepLength) ? selector.xVarsBetweenCurrentAndPast(maxStepLength) : new TreeSet<String>();
            
            TreeSet<String> yVarsOrig = selector.yVarsBetweenCurrentAndPast(maxStepLength);
            
            TreeSet<String> allVarsBwd = new TreeSet<>(xVarsOrig);

            allVarsBwd.addAll(yVarsOrig);

            SubNLPProblem subProbBwd = problem.generateSubOptProblem(consInBwd, allVarsBwd, selector.dbgFullProbSparsity());
            
            TreeMap<String,Double> rowScales = problem.rowScales(consInBwd);
            
            TreeMap<String,Double> columnScales = problem.colScales(allVarsBwd);
            
            return projectBack(subProbBwd, 
                               solver, 
                               allPartialSolutions, 
                               alreadyFwdProjected, 
                               rowScales, 
                               columnScales, 
                               consInBwd, 
                               xVarsOrig, 
                               yVarsOrig, 
                               selector.getVariablesToGuess(), 
                               1, // FIXME magic number 
                               rnd, 
                               reparamOpts,
                               selector.isLastSubproblem(),
                               dbgSubproblemIndex);
    }

    static private ArrayList<Solution> projectBack(SubNLPProblem subProbBwd, 
                                                   Solver solver,
                                                   ArrayList<Solution> allPartialSolutions,
                                                   ArrayList<Solution> alreadyFwdProjected,
                                                   TreeMap<String,Double> rowScales,
                                                   TreeMap<String,Double> columnScales,
                                                   TreeSet<String> consInBwd,
                                                   TreeSet<String> xVarsOrig, 
                                                   TreeSet<String> yVarsOrig, 
                                                   TreeSet<String> yLast,
                                                   int nToFix,
                                                   Random rnd,
                                                   ReparamOpts reparamOpts,
                                                   boolean isLast,
                                                   int dbgSubproblemIndex) 
    {
        ArrayList<Solution> result = new ArrayList<>();
        
        final int POINTS_TO_PROPAGATE = reparamOpts.POINTS_TO_PROPAGATE;
        
        if (xVarsOrig.isEmpty()) {
            
            // allPartialSolutions ignored, we simply introduce new ones
            
            TreeMap<String,Double> xPartEmpty = new TreeMap<>();
            
            // Similar to UD sampling but only yLast is sampled
            ArrayList<TreeMap<String,Double>> yLeastInfluential = subProbBwd.findLeastInfluentialWithQRP(rowScales, 
                                                                                                         columnScales, 
                                                                                                         xPartEmpty,
                                                                                                         yVarsOrig, 
                                                                                                         POINTS_TO_PROPAGATE, 
                                                                                                         yLast,
                                                                                                         nToFix,
                                                                                                         rnd);
            
            result.addAll( bwdProjection(subProbBwd, 
                                         solver, 
                                         alreadyFwdProjected, 
                                         rowScales, 
                                         columnScales, 
                                         Factory.createEmptySolution(),
                                         new TreeSet<String>(), // to be removed -> nothing
                                         yLeastInfluential, 
                                         consInBwd, 
                                         xVarsOrig, 
                                         yVarsOrig,
                                         reparamOpts,
                                         rnd, 
                                         dbgSubproblemIndex,
                                         isLast)
                         );
        }
        else {

            // Here we create potentially quadratic amount of new solutions
            
            TreeSet<String> toBeRemoved = new TreeSet<>(yVarsOrig);
            
            toBeRemoved.removeAll(yLast); // partial solutions of the *previous* front 
                                          // doesn't contain yVars of the current subproblem
            for (Solution s : allPartialSolutions) {
                
                TreeMap<String,Double> xPart = s.getValues(xVarsOrig);
                
                ArrayList<TreeMap<String,Double>> yLeastInfluential = subProbBwd.findLeastInfluentialWithQRP(rowScales, 
                                                                                                             columnScales, 
                                                                                                             xPart, 
                                                                                                             yVarsOrig, 
                                                                                                             POINTS_TO_PROPAGATE, 
                                                                                                             yLast,
                                                                                                             nToFix,
                                                                                                             rnd);
                result.addAll(  bwdProjection(subProbBwd, 
                                              solver, 
                                              alreadyFwdProjected, 
                                              rowScales, 
                                              columnScales, 
                                              s, 
                                              toBeRemoved,
                                              yLeastInfluential, 
                                              consInBwd, 
                                              xVarsOrig, 
                                              yVarsOrig, 
                                              reparamOpts,
                                              rnd, 
                                              dbgSubproblemIndex,
                                              isLast)
                             );                
            }
        }

        return result;
    }

    static private ArrayList<Solution> bwdProjection(SubNLPProblem subProbBwd, 
                                                     Solver solver,
                                                     ArrayList<Solution> alreadyFwdProjected,
                                                     TreeMap<String,Double> rowScales,
                                                     TreeMap<String,Double> columnScales,
                                                     Solution s,
                                                     TreeSet<String> toBeRemoved,
                                                     ArrayList<TreeMap<String,Double>> yLeastInfluential,
                                                     TreeSet<String> consInBwd,
                                                     TreeSet<String> xVarsOrig, 
                                                     TreeSet<String> yVarsOrig,
                                                     ReparamOpts reparamOpts,
                                                     Random rnd,
                                                     int dbgSubproblemIndex,
                                                     boolean isLast) 
    {
        ArrayList<Solution> extended = new ArrayList<>(yLeastInfluential.size());
        
        double delta = 1.0d / reparamOpts.POINTS_TO_PROPAGATE;
        
        for (TreeMap<String,Double> yFixedLHS : yLeastInfluential) {
            
            if (isDistantEnough(yFixedLHS, alreadyFwdProjected, columnScales, delta)) {
                
                //log.info("Least influential: "+yFixedLHS.firstKey());

                extended.addAll( projectFwd(subProbBwd, 
                                            solver, 
                                            rowScales, 
                                            columnScales, 
                                            s, 
                                            toBeRemoved,
                                            yFixedLHS, 
                                            consInBwd, 
                                            xVarsOrig, 
                                            yVarsOrig, 
                                            reparamOpts,
                                            rnd, 
                                            dbgSubproblemIndex,
                                            isLast )
                                );
            }
        }
        
        return extended;
    }
    
    static private boolean isDistantEnough(TreeMap<String,Double> yFixedLHS, 
                                           ArrayList<Solution>    alreadyFwdProjected,
                                           TreeMap<String,Double> columnScales,
                                           double delta ) 
    {
        double[] yFixed = Util.values(yFixedLHS);
        
        for (Solution s : alreadyFwdProjected) {
            
            SortedSet<String> yVarNames = yFixedLHS.navigableKeySet();
            
            double[] yFwd = s.values(yVarNames);
            
            double[] scales = Util.extractValues(columnScales, yVarNames);
            
            if (Util.scaledChebyshevDistance(yFixed, yFwd, scales) < 2.0*delta) { // magic number 
                
                return false;
            }
        }
        
        return true;
    }
    
    static private ArrayList<Solution> projectFwd(SubNLPProblem subProbBwd, 
                                                  Solver solver,
                                                  TreeMap<String,Double> rowScales,
                                                  TreeMap<String,Double> columnScales,
                                                  Solution s,
                                                  TreeSet<String> toBeRemoved,
                                                  TreeMap<String,Double> yFixedLHS,
                                                  @SuppressWarnings("unused")
                                                  TreeSet<String> consInBwd,
                                                  TreeSet<String> xVarsOrig, 
                                                  TreeSet<String> yVarsOrig,
                                                  ReparamOpts reparamOpts,
                                                  Random rnd,
                                                  int dbgSubproblemIndex,
                                                  boolean isLast) 
    {
        TreeSet<String> xVars = new TreeSet<>(xVarsOrig);
        TreeSet<String> yVars = new TreeSet<>(yVarsOrig);

        for (String yVarName : yFixedLHS.keySet()) {
            boolean removed = yVars.remove(yVarName);
            assert  removed;
            boolean added = xVars.add(yVarName);
            assert  added;
        }

        //System.out.println("Fixed:  "+xVars);
        //System.out.println("Sought: "+yVars);
        //System.out.println("Cons:   "+consInBwd);
        
        ArrayList<Solution> partialSolutions = new ArrayList<>();
        
        partialSolutions.add( s.extend(toBeRemoved, yFixedLHS, 0) );
        
        ArrayList<Solution> extended = Reparameterization.forwardProjection(subProbBwd, 
                                                                            solver, 
                                                                            partialSolutions,
                                                                            xVars,
                                                                            yVars,
                                                                            rowScales, 
                                                                            columnScales, 
                                                                            reparamOpts.POINTS_TO_PROPAGATE,
                                                                            reparamOpts.EXTENSION_ATTEMPTS,
                                                                            reparamOpts.EXPECTED_OUTPUT,
                                                                            reparamOpts.BOUND_RELAX_FACTOR,
                                                                            rnd,
                                                                            dbgSubproblemIndex,
                                                                            isLast);
        return extended;
    }
   
    private void plotThenDiscardSuperfluousCandidates(int pointsToKeep) {
        
        plotter.writePartialSolutions(extendedPartialSolutions);
   
        if (extendedPartialSolutions.size() <= pointsToKeep) {
            
            plotter.writePartialSolutions(extendedPartialSolutions);
            
            return;
        }
        
        discardSuperfluousCandidates(pointsToKeep);
        
        plotter.writePartialSolutions(extendedPartialSolutions);
    }
    
    private void discardSuperfluousCandidates(int pointsToKeep) {
        
        if (extendedPartialSolutions.size() <= pointsToKeep) {
            
            return;
        }
        
        double[] colScales = problem.columnScales(extendedPartialSolutions.get(0).getVarNames());
        
        CandidateSelector candidateSelector = Factory.createCumulativeEvaluator(colScales);
        
        //Collections.shuffle(extendedPartialSolutions, rnd);
        
        candidateSelector.selectMostPromisingOnlyByDistance(extendedPartialSolutions, pointsToKeep, INFEAS_REG);
        
        ArrayList<Solution> kept = candidateSelector.getSelectedOnes();
        
        log.info("Discarding "+(extendedPartialSolutions.size()-kept.size())+" partial solutions");
        
        extendedPartialSolutions = candidateSelector.getSelectedOnes();        
    }

    private void runOnFullProblemFromAllPartialSolutions() {

        selector.setFullProblem(); // all constraints selected, nothing fixed, all variables guessed
        
        solver = createFullproblemSolver();

        for (Solution s : allPartialSolutions) {
            
            assert !Double.isNaN(s.getCumulativeInfeasibility());
            
            currentPartialSolution = s;

            subProblem = problem.generateSubOptProblem(selector.getNextConstraints(),
                                                       currentPartialSolution.getValues(selector.getFixedVars()),
                                                       selector.getVariablesToGuess(),
                                                       selector.dbgFullProbSparsity());

            solution = solver.solveSubNLPProblem(subProblem, startingPointFromSolution(currentPartialSolution));

            // extended at this point contains unsuccessful solutions
            extendedPartialSolutions.add(solution);
        }

        allPartialSolutions = extendedPartialSolutions;

        extendedPartialSolutions = new ArrayList<>();    

    }
    
    private void runNewtonFromAllPartialSolutions() {

        selector.setFullProblem(); // all constraints selected, nothing fixed, all variables guessed
        
        subProblem = problem.generateSubOptProblem(selector.getNextConstraints(),
                                                   new TreeMap<String,Double>(),
                                                   selector.getVariablesToGuess(),
                                                   selector.dbgFullProbSparsity());

        // 1. extendedPartialSolutions contains unsuccessful solutions
        // 2. allPartialSolutions has been reordered (sorted by constraint violation
        extendedPartialSolutions = subProblem.denseNewtonIterate(allPartialSolutions);

        allPartialSolutions = extendedPartialSolutions;

        extendedPartialSolutions = new ArrayList<>();
    }
    
    private void run_va27() {
        
        int n_initial_points = allPartialSolutions.size();

        //allPartialSolutions = VA27_hack.perturb(problem.getReferenceSolutions());
        allPartialSolutions = VA27_hack.run(allPartialSolutions);
        
        runNewtonFromAllPartialSolutions();
        
        int missing = n_initial_points-allPartialSolutions.size();
        for (int i=0; i<missing; ++i)
            allPartialSolutions.add(Factory.createInvalidSolution());
        
        extendedPartialSolutions = new ArrayList<>();
    }
    
    private void run_lmbopt_newton() {
        //int n_initial_points = allPartialSolutions.size();
        selector.setFullProblem(); // all constraints selected, nothing fixed, all variables guessed
        solver = Factory.createSolver("lmbopt");
        for (Solution s : allPartialSolutions) {
            assert !Double.isNaN(s.getCumulativeInfeasibility());
            currentPartialSolution = s;
            subProblem = problem.generateSubOptProblem(selector.getNextConstraints(),
                                                       currentPartialSolution.getValues(selector.getFixedVars()),
                                                       selector.getVariablesToGuess(),
                                                       selector.dbgFullProbSparsity());
            solution = solver.solveSubNLPProblem(subProblem, startingPointFromSolution(currentPartialSolution));
            // extended at this point contains unsuccessful solutions
            extendedPartialSolutions.add(solution);
        }
        allPartialSolutions = extendedPartialSolutions;
        extendedPartialSolutions = new ArrayList<>();  
        
        runNewtonFromAllPartialSolutions();
        
        /*int missing = n_initial_points-allPartialSolutions.size();
        for (int i=0; i<missing; ++i)
            allPartialSolutions.add(Factory.createInvalidSolution());
        
        extendedPartialSolutions = new ArrayList<>(); */
    }
 
    private void structureUnawareMultistartOnFullProblem() {
        
        selector.setFullProblem(); // all constraints selected, nothing fixed, all variables guessed
        
        solver = createFullproblemSolver();
        
        /*// Applying small perturbations to stewgou40; run it with 40 points
        ArrayList<TreeMap<String,Double>> refSols = problem.getReferenceSolutions();
        Iterator<TreeMap<String,Double>> itr = refSols.iterator();
        Random rnd = new Random(31L);
        int kk = 1;*/
        
        for (Solution s : allPartialSolutions) {
            
            currentPartialSolution = s;

            subProblem = problem.generateSubOptProblem(selector.getNextConstraints(),
                                                       currentPartialSolution.getValues(selector.getFixedVars()),
                                                       selector.getVariablesToGuess(),
                                                       selector.dbgFullProbSparsity());
            
            startingpointGenerator.setVarBounds(subProblem.getLowerBoundsOfGuessedVars(),
                                                subProblem.getUpperBoundsOfGuessedVars());
 
            solution = solver.solveSubNLPProblem(subProblem, startingpointGenerator.getNextGuess());
            /*TreeMap<String,Double> sol = new TreeMap<String,Double>();
            double[] perturb = new double[9];
            int ii = 0;
            for (Map.Entry<String,Double> e : itr.next().entrySet()) {
                double d0 = e.getValue();
                double eps = ((rnd.nextDouble()-0.5)/0.5)*0.0035;
                double d = d0 + eps;
                d = Math.max(-1, d);
                d = Math.min( 1, d);
                //System.out.println(d);
                sol.put(e.getKey(), d);
                perturb[ii++] = d-d0;
            }
            System.out.println("Solution" + kk++);
            System.out.println(Util.infinityNorm(perturb));
            solution = solver.solveSubNLPProblem(subProblem, Util.extractValues(sol, sol.navigableKeySet()));
            */
            if (solution.isSuccessful()) {

                extendedPartialSolutions.add(solution);

                //dbgRerunFromSolution();
            }
        }

        allPartialSolutions = extendedPartialSolutions;
        
        extendedPartialSolutions = new ArrayList<>();         
    }
    
    private void dumpThoseNearestToReferenceSols(ArrayList<Solution> startingPoints) {
        
        ArrayList<TreeMap<String,Double>> refSols = problem.getReferenceSolutions();
        
        double[] colScales = problem.columnScales(refSols.get(0).navigableKeySet());
        
        ArrayList<double[]> startPoints = new ArrayList<>(startingPoints.size());
        
        for (Solution s : startingPoints) {
            
            startPoints.add(Util.values(s.getAllVars()));
        }
        
        ArrayList<double[]> sols = new ArrayList<>(allPartialSolutions.size());
        
        for (Solution s : allPartialSolutions) {
            
            sols.add(s.isSuccessful() ? Util.values(s.getAllVars()) : null);
        }
        
        //--------------------------------------------------------------------------
        
        ArrayList<TreeMap<String,Double>> nearestSolutions = new ArrayList<>();
        
        for (TreeMap<String,Double> refSol : refSols) {
            
            TreeMap<String,Double> nearest = findNearest(refSol, startPoints, sols, colScales);
            
            if (nearest!=null) {
            
                nearestSolutions.add( nearest );
            }
        }
        
        plotter.writeNearestSolutions(nearestSolutions);
    }
    
    private TreeMap<String,Double> findNearest(TreeMap<String,Double> refSol, 
                                               ArrayList<double[]> startPoints,
                                               ArrayList<double[]> sols,
                                               double[] colScales              ) 
    {
        assert sols.size()==startPoints.size();
        
        double[] xRef = Util.values(refSol);

        double   minDist = Double.MAX_VALUE;
        double[] nearestToRef = null;

        for (int k=0, n=allPartialSolutions.size(); k<n; ++k) {
            
            if (allPartialSolutions.get(k).isSuccessful() && 
                Util.scaledChebyshevDistance(xRef, sols.get(k), colScales) < 0.001) // magic number 
            { 
                // we have found this reference solution from the k-th starting point
                double[] start = startPoints.get(k);
                double dist = Util.scaledEuclideanDistance(xRef, start, colScales);
                
                if (dist < minDist) {
                    minDist = dist;
                    nearestToRef = start;
                } 
            }
        }
        
        // null means that this reference solution was not found
        return nearestToRef == null ? null : Util.createMap(allPartialSolutions.get(0).getVarNames(), nearestToRef);
    }

    private void writeSolutions() {

        int nSols = 0;

        for (Solution s : allPartialSolutions) {
            
            //log.info(s.toString());
            // TODO Clean up how the solutions are dumped
            if (s.getInfeasibilityBestObjective() < INFEASIBILITY_THRESHOLD) {

                ++nSols;
            }
        }
        
        compareWithReferenceSolutions();

        log.info("Acceptable solutions: " + nSols + " of " + allPartialSolutions.size());
    }
    
    private void showConditionNumbers() {
        
        if (!SHOW_CONDITION_NUMBER)
            return;
          
        selector.setFullProblem(); // all constraints selected, nothing fixed, all variables guessed
        
        subProblem = problem.generateSubOptProblem(selector.getNextConstraints(),
                                                   new TreeMap<String,Double>(),
                                                   selector.getVariablesToGuess(),
                                                   selector.dbgFullProbSparsity());

       subProblem.showConditionNumbers(allPartialSolutions);
    }
    
    private void compareWithReferenceSolutions() {

        // TODO Make the magic number a parameter
        int[] count = problem.compareWithReferenceSolutions(allPartialSolutions, INFEASIBILITY_THRESHOLD, 1.0e-3);
        
        for (int i=0; i<count.length; ++i) {
            
            log.info("Solution "+(i+1)+" found "+count[i]+" times");         
        }
               
        int countSum = 0;
        
        for (int i : count) 
            countSum += i;
        
        log.info("Found references solutions "+countSum+" times");
    }

    @SuppressWarnings("unused")
    private void dbgRerunFromSolution() {

        if (log.isTraceEnabled()) {

            log.trace("==============================================================");
            log.trace("Rerunning solver from the solution; it should not take long and should succeed");

            solver.solveSubNLPProblem(subProblem, startingPointFromSolution(solution));
            // ignore costs

            log.trace("==============================================================");
        }
    }

    /** The returned values are in alphabetical order */
    private double[] startingPointFromSolution(Solution s) {

        TreeMap<String, Double> sol = s.getValues(selector.getVariablesToGuess());

        double[] values = new double[sol.size()];

        int k = 0;

        for (double d : sol.values()) {

            values[k++] = d;
        }

        return values;
    }
}
