//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms;

import java.util.ArrayList;
import java.util.Random;

public final class LatinHypercubeSampling {

    private static int[][] generateIndices(int nVars, int m, Random rnd) {
        
        int[][] indices = new int[nVars][m];
        
        for (int i=0; i<nVars; ++i) {
            
            for (int j=0; j<m; ++j) {
                
                indices[i][j] = j;
            }
        }

        for (int i=1; i<nVars; ++i) {
            
            Util.shuffle(indices[i], rnd);
        }
        
        return indices;
    }
    
    private static double[] getSample(int[][] indices, int k, double[] lb, double[] ub) {

        int m = indices[0].length; // m is the number of all samples, we are looking at the k-th sample
        
        int nVars = lb.length;
        
        double[] x = new double[nVars];
        
        double oneStep = 1.0D/(m-1);
        
        for (int i=0; i<nVars; ++i) {
            
            double stepSize = oneStep*(ub[i]-lb[i]);            
            
            int index = indices[i][k];
            
            double x_i = lb[i] +  index*stepSize;
            
            assert lb[i] <= x_i;
            assert ub[i] + Math.abs(1.0e-6*ub[i]) >= x_i;
            
            x_i = Math.min(x_i, ub[i]);
 
            x[i] = x_i;
        }

        return x;        
    }
    
    public static ArrayList<double[]> generateLHSamples(Random rnd, int m, double[] lb, double[] ub) {
        
        assert lb.length==ub.length;
        
        int nVars = lb.length;
        
        assert nVars >= 1 : "There should be at least on variable to be sampled! (nVars="+nVars+")";
        
        assert m > 2 : "There should be at least 3 variables per coordinate in a Latin hypercube sampling! (gave: "+m+")";
        
        int[][] indices = generateIndices(nVars, m, rnd);
        
        ArrayList<double[]> samples = new ArrayList<>();
        
        for (int k=0; k<m; ++k) {
            
            samples.add(getSample(indices, k, lb, ub));
        }

        return samples;
    }
    
    public static void main(String[] args) {
        
        double[] lb = new double[]{ 2, 2, 2 };
        double[] ub = new double[]{ 5, 5, 5 };
        
        Random rnd = new Random(31L);
        
        ArrayList<double[]> samples = generateLHSamples(rnd, 21, lb, ub);
        
        for (double[] samp : samples) {
            
            System.out.println( Util.arrayToString(samp, "%4.2f") );
        }
        
        return;
    }
}
