//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Map;
import java.util.TreeMap;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

//TODO Sanity check should be implemented

enum IntOptKey {        
    POINTS_TO_PROPAGATE,
    POINTS_PER_FRONT,
    EXPECTED_OUTPUT,
    EXTENSION_ATTEMPTS,
    FIXED_LAG;
}

enum DblOptKey {
    BOUND_RELAX_FACTOR,
    INFEAS_REG,
    INFEASIBILITY_THRESHOLD;
}

enum StrOptKey {
    FULL_PROBLEM_ONLY,
    SUBPROBLEM_SOLVER,
    FULLPROBLEM_SOLVER,
    SCALING,
    PLOT,
    SUBPROB_LOG,
    FULLPROB_LOG,
    SHOW_CONDITION_NUMBER;    
}

public final class Options {

    private static final Logger log = LogManager.getLogger();

    private final EnumMap<IntOptKey,Integer> intOptions    = new EnumMap<>(IntOptKey.class);
    private final EnumMap<DblOptKey,Double>  doubleOptions = new EnumMap<>(DblOptKey.class);
    private final EnumMap<StrOptKey,String>  stringOptions = new EnumMap<>(StrOptKey.class);

    public Options(String optionsFile) {

        for (IntOptKey key : EnumSet.allOf(IntOptKey.class))
            intOptions.put(key, null);

        for (DblOptKey key : EnumSet.allOf(DblOptKey.class))
            doubleOptions.put(key, null);

        for (StrOptKey key : EnumSet.allOf(StrOptKey.class))
            stringOptions.put(key, null);

        TreeMap<String,String> options = loadOptions(optionsFile);

        for (Map.Entry<String,String> e : options.entrySet())
            add(e.getKey(), e.getValue());

        checkForMissingValues();

        showAllOptions();
    }

    private void checkForMissingValues() {

        boolean missing = false;

        for (Map.Entry<IntOptKey,Integer> e : intOptions.entrySet()) {

            if (e.getValue()==null) {                
                log.fatal("The key " + e.getKey() + " misses its value");
                missing = true;                
            }
        }

        for (Map.Entry<DblOptKey,Double> e : doubleOptions.entrySet()) {

            if (e.getValue()==null) {                
                log.fatal("The key " + e.getKey() + " misses its value");
                missing = true;                
            }
        }

        for (Map.Entry<StrOptKey,String> e : stringOptions.entrySet()) {

            if (e.getValue()==null) {                
                log.fatal("The key " + e.getKey() + " misses its value");
                missing = true;                
            }
        }

        if (missing) {

            throw new ProgramTerminationRequested();
        }
    }

    private void showAllOptions() {

        log.info("Running with the following options:");

        for (Map.Entry<IntOptKey,Integer> e : intOptions.entrySet()) {

            log.info(e.getKey() + " = " + e.getValue());
        }

        for (Map.Entry<DblOptKey,Double> e : doubleOptions.entrySet()) {

            log.info(e.getKey() + " = " + e.getValue());
        }

        for (Map.Entry<StrOptKey,String> e : stringOptions.entrySet()) {

            log.info(e.getKey() + " = " + e.getValue());
        }
    }

    private void add(String key, String value) {

        int addedInt = tryAddingAsInteger(key, value) ? 1 : 0;

        int addedDouble = tryAddingAsDouble(key, value) ? 1 : 0;

        int addedStr = tryAddingAsString(key, value) ? 1 : 0;

        int keyInsertedThisManyTimes = addedInt+addedDouble+addedStr; 

        assert keyInsertedThisManyTimes <= 1 : "Key found multiple times: "+key+" = "+value;

        if (keyInsertedThisManyTimes < 1) {

            log.fatal("Key \""+key+"\" not found!");

            throw new ProgramTerminationRequested();
        }
        // Success has already been logged
    }

    private interface Callback {

        /** Returns old value in the map or null if none */
        Object push(String key, String value);
    }

    private boolean tryAddingAsInteger(String key, String value) {

        Callback integerCallback =  new Callback() {

            @Override
            public Object push(String key, String value) {

                IntOptKey enumKey = Enum.valueOf(IntOptKey.class, key);

                return intOptions.put(enumKey, Integer.valueOf(value));
            }
        };

        return addOption(key, value, integerCallback);
    }

    private boolean tryAddingAsDouble(String key, String value) {

        Callback doubleCallback =  new Callback() {

            @Override
            public Object push(String key, String value) {

                DblOptKey enumKey = Enum.valueOf(DblOptKey.class, key);

                return doubleOptions.put(enumKey, Double.valueOf(value));
            }
        };

        return addOption(key, value, doubleCallback);
    }

    private boolean tryAddingAsString(String key, String value) {

        Callback stringCallback =  new Callback() {

            @Override
            public Object push(String key, String value) {

                StrOptKey enumKey = Enum.valueOf(StrOptKey.class, key);

                return stringOptions.put(enumKey, value);
            }
        };

        return addOption(key, value, stringCallback);
    }

    private static boolean addOption(String key, String value, Callback c) {

        try {

            Object oldValue = c.push(key, value);

            if (oldValue == null) {
                log.debug("Added "+key+" = "+value);
                return true;
            }
            else {
                log.fatal("Key "+key+" has already been added with value "+oldValue);
                throw new ProgramTerminationRequested();
            }
        }
        catch (NumberFormatException e) {
            // Thrown in Callback by *.valueOf
            log.fatal("Failed to parse value \""+value+"\" of key "+key);
            throw new ProgramTerminationRequested();
        }
        catch (IllegalArgumentException e) { 
            // Thrown in Callback by Enum.valueOf,
            // that is probably OK, the key might belong to some other type
            return false;
        }
    }

    public int get(IntOptKey key) {

        Integer value = intOptions.get(key);

        assert value != null : "Key: "+key;

        return value;
    }

    public double get(DblOptKey key) {

        Double value = doubleOptions.get(key);

        assert value != null : "Key: "+key;

        return value;
    }

    public String get(StrOptKey key) {

        String value = stringOptions.get(key);

        assert value != null : "Key: "+key;

        return value;
    }

    //-----------------------------------------------------------------------------

    static private TreeMap<String, String> loadOptions(String fileName) {

        ArrayList<String> fileContent = Util.stringVectorFromFile(fileName);

        TreeMap<String,String> result = new TreeMap<>();

        int k=0;

        for (String line : fileContent) {

            addLine(line, result, ++k);
        }

        return result;
    }

    static private void addLine(String line, TreeMap<String,String> keyValuePairs, int lineNumber) {

        String[] arr = lineToValueKeyPair(line);

        if (arr.length==1 && arr[0].equals("")) {
            log.debug("Line "+lineNumber+" \""+line+"\" is discarded");
            return;
        }

        if (arr.length!=2) {
            log.fatal("Line "+lineNumber+" \""+line+"\" (parsed as "+Arrays.toString(arr)+", length "+arr.length+")");
            throw new ProgramTerminationRequested();
        }

        String oldValue = keyValuePairs.put(arr[0], arr[1]);

        if (oldValue==null) {
            log.debug("Line "+lineNumber+" \""+line+"\" is added as "+Arrays.toString(arr));
        }
        else {
            log.fatal("Line "+lineNumber+": key "+arr[0]+" was already in the map with "+oldValue+", now found with "+arr[1]);
            throw new ProgramTerminationRequested();
        }
    }

    static private String[] lineToValueKeyPair(String line) {

        int commentBegin = line.indexOf('#');

        String lineWithoutComment = (commentBegin!=-1) ? line.substring(0, commentBegin) : line;

        return lineWithoutComment.trim().split("\\s+");        
    }

    public static void writeAllOptionsToFile(String fileName) {

        ArrayList<String> content = new ArrayList<>();

        content.add("# Integer options");
        for (IntOptKey k : EnumSet.allOf(IntOptKey.class))
            content.add(k+"\t");


        content.add("# Double options");
        for (DblOptKey k : EnumSet.allOf(DblOptKey.class))
            content.add(k+"\t");


        content.add("# String options");
        for (StrOptKey k : EnumSet.allOf(StrOptKey.class))
            content.add(k+"\t");

        Util.writeToFile(fileName, content);
    }
}
