//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

final class IndexDistanceSquared  {
    
    final int    index;
    final double distanceSquared;
    
    public IndexDistanceSquared(int index, double distanceSquared) {
        this.index = index;
        this.distanceSquared = distanceSquared;
        assert 0.0 <= distanceSquared;
    }
    
    final static class CompareByDistance implements Comparator<IndexDistanceSquared>  {
        @Override
        public int compare(IndexDistanceSquared a, IndexDistanceSquared b) {
            return Double.compare(a.distanceSquared, b.distanceSquared);
        }     
    }
}

//=================================================================================================

final class Reparameterization {

    static ArrayList<Solution> forwardProjection(SubNLPProblem subProblem,
                                                 Solver solver,
                                                 ArrayList<Solution> allPartialSolutions,
                                                 SortedSet<String> xVars,
                                                 SortedSet<String> yVars,
                                                 TreeMap<String,Double> rowScales,
                                                 TreeMap<String,Double> columnScales,
                                                 int POINTS_TO_PROPAGATE,
                                                 int EXTENSION_ATTEMPTS,
                                                 int EXPECTED_OUTPUT,
                                                 double BOUND_RELAX_FACTOR,
                                                 Random rnd,
                                                 int dbgSubproblemIndex,
                                                 boolean isLast)
    {
        ArrayList<Solution> ret;
        
        if (yVars.size() <= subProblem.getNumberOfConstraints()) {
            
            if (yVars.size()!=subProblem.getNumberOfConstraints() && !isLast)
                assert !isLast;
            
            SquareOrOverDetermined reparam = new SquareOrOverDetermined(subProblem, solver, allPartialSolutions, xVars, yVars, rowScales, columnScales, rnd, EXTENSION_ATTEMPTS, EXPECTED_OUTPUT, BOUND_RELAX_FACTOR);
            
            ret = reparam.run();
        }
        else {

            Underdetermined reparam = new Underdetermined(subProblem, solver, allPartialSolutions, xVars, yVars, rowScales, columnScales, rnd, EXTENSION_ATTEMPTS, EXPECTED_OUTPUT, BOUND_RELAX_FACTOR);

            ret = reparam.run();
        }
        
        if (!isLast) {

            int deleteAtMost = (int) (0.80*allPartialSolutions.size()); // magic number

            discardTooBoundInfeasible(dbgSubproblemIndex, ret, deleteAtMost, POINTS_TO_PROPAGATE);
        }
        
        for (Solution s : ret) {

            s.incrementCumulativeInfeasibilityWithBestObjAndBoundViolSquared();
            
            s.markAsFromForwardProjection();
        }

        return ret;
    }
    
    private static void discardTooBoundInfeasible(int dbgSubproblemIndex, ArrayList<Solution> ret, int n, int POINTS_TO_PROPAGATE) {
        
        dbgInfeasibilityDistribution(dbgSubproblemIndex, ret);
        
        ArrayList<IndexDistanceSquared> boundInfeas = new ArrayList<>();
        
        //@SuppressWarnings("unused")
        double delta = 1.0d/POINTS_TO_PROPAGATE;

        for (int k=0; k<ret.size(); ++k) {

            Solution s = ret.get(k);

            //                                   4*delta 0.10
            if (s.getInfNormScaledBoundViol() >= 2.0*delta) { // magic number

                boundInfeas.add(new IndexDistanceSquared(k, s.getScaledBoundconstraintViolationSquared()));
            }
        }    
        
        Collections.sort(boundInfeas, new IndexDistanceSquared.CompareByDistance());
        
        //-----------------------------------------------
        
        int deleteAtMost = Math.min(n, boundInfeas.size());
        
        int[] indexToDelete = new int[deleteAtMost];
        
        int pos = 0;
        
        for (int k=boundInfeas.size()-1; k>=boundInfeas.size()-deleteAtMost; --k) {
            
            indexToDelete[pos++] = boundInfeas.get(k).index;
        }
        
        assert pos==deleteAtMost;
        
        //-----------------------------------------------
        
        Arrays.sort(indexToDelete);
        
        for (int k=indexToDelete.length-1; k>=0; --k) {
            
            ret.remove(indexToDelete[k]);
        }
    }
    
    private static void dbgInfeasibilityDistribution(int dbgSubproblemIndex, ArrayList<Solution> ret) {
        
        if (dbgSubproblemIndex < 0)
            return;
        
        double[] boundViols = new double[ret.size()];
        
        double[] bestObj = new double[ret.size()];
        
        int k=0;
        
        for (Solution s : ret) {
            
            boundViols[k] = Math.sqrt(s.getScaledBoundconstraintViolationSquared());
            
            bestObj[k] = Math.sqrt(s.getInfeasibilityBestObjective());
            
            ++k;
        }
        
        Util.printLocation();
        
        System.out.println("At subproblem "+Math.abs(dbgSubproblemIndex)+" in "+(dbgSubproblemIndex>0?"forward":"backward")+" mode");
        
        System.out.println("Scaled bound violation distribution");
        
        Util.sort_and_show_statistics(boundViols);
        
        System.out.println("Scaled L2 norm of constraint violation distribution");
        
        Util.sort_and_show_statistics(bestObj);
    }
}

//=================================================================================================

final class SquareOrOverDetermined {

    private final SubNLPProblem subProblem; // F(x,y)=0
    private final Solver solver;
    private final ArrayList<Solution> allPartialSolutions;
    private final SortedSet<String> xVars;
    private final SortedSet<String> yVars;
    private final Random rnd;
    
    private final TreeMap<String,Double> lowerBounds;
    private final TreeMap<String,Double> upperBounds;
    
    private final int EXTENSION_ATTEMPTS;
    private final int EXPECTED_OUTPUT;
    private final double BOUND_RELAX_FACTOR;

    private final double delta_min = 0.01; // magic number
    
    private final TreeMap<String,Double> rowScales;
    private final TreeMap<String,Double> columnScales;
    private final double[] yScales;    
    private final double[] yL;    
    private final double[] yU;
    private final ArrayList<Solution> yFwd;

    SquareOrOverDetermined(SubNLPProblem subProblem,
                           Solver solver,
                           ArrayList<Solution> allPartialSolutions,
                           SortedSet<String> xVars,
                           SortedSet<String> yVars,
                           TreeMap<String,Double> rowScales,
                           TreeMap<String,Double> columnScales,
                           Random rnd,
                           int EXTENSION_ATTEMPTS,
                           int EXPECTED_OUTPUT,
                           double BOUND_RELAX_FACTOR )
    {
        // convenience ctor, retrieves the orig var bounds from the subproblem
        this(subProblem, 
             solver, 
             allPartialSolutions, 
             xVars, 
             yVars, 
             rowScales, 
             columnScales, 
             rnd, 
             subProblem.allLowerBounds(), 
             subProblem.allUpperBounds(), 
             EXTENSION_ATTEMPTS,
             EXPECTED_OUTPUT,
             BOUND_RELAX_FACTOR);
    }
    
    SquareOrOverDetermined(SubNLPProblem subProblem,
                           TreeMap<String,Double> allLowerBounds,
                           TreeMap<String,Double> allUpperBounds,
                           Solver solver,
                           ArrayList<Solution> allPartialSolutions,
                           SortedSet<String> xVars,
                           SortedSet<String> yVars,
                           TreeMap<String,Double> rowScales,
                           TreeMap<String,Double> columnScales,
                           Random rnd, 
                           int EXTENSION_ATTEMPTS,
                           int EXPECTED_OUTPUT,
                           double BOUND_RELAX_FACTOR)
    {
        this(subProblem, 
             solver, 
             allPartialSolutions, 
             xVars, 
             yVars, 
             rowScales, 
             columnScales, 
             rnd, 
             allLowerBounds, 
             allUpperBounds,
             EXTENSION_ATTEMPTS,
             EXPECTED_OUTPUT,
             BOUND_RELAX_FACTOR );
    }    

    private SquareOrOverDetermined(SubNLPProblem subProblem,
                           Solver solver,
                           ArrayList<Solution> allPartialSolutions,
                           SortedSet<String> xVars,
                           SortedSet<String> yVars,
                           TreeMap<String,Double> rowScales,
                           TreeMap<String,Double> columnScales,
                           Random rnd,
                           TreeMap<String,Double> allLowerBounds,
                           TreeMap<String,Double> allUpperBounds,
                           int EXTENSION_ATTEMPTS,
                           int EXPECTED_OUTPUT,
                           double BOUND_RELAX_FACTOR )
    {
        assert yVars.size() <= subProblem.getNumberOfConstraints();
        
        subProblem.dbgAllVarsCovered(xVars, yVars);
        
        this.subProblem = subProblem;
        this.solver = solver;
        this.allPartialSolutions = allPartialSolutions;
        this.xVars = xVars;
        this.yVars = yVars;
        this.columnScales = columnScales;
        this.rnd = rnd;
        
        this.EXTENSION_ATTEMPTS = EXTENSION_ATTEMPTS;
        this.EXPECTED_OUTPUT    = EXPECTED_OUTPUT;
        this.BOUND_RELAX_FACTOR = BOUND_RELAX_FACTOR;
        
        lowerBounds = allLowerBounds;
        upperBounds = allUpperBounds;
        
        this.rowScales = rowScales;
        yScales = Util.extractValues(columnScales, yVars);
        yL = Util.extractValues(lowerBounds, yVars);        
        yU = Util.extractValues(upperBounds, yVars);

        int nMax = allPartialSolutions.size()*EXTENSION_ATTEMPTS;
        yFwd = new ArrayList<>(nMax);
    }
    
    ArrayList<Solution> run() {

        projectForward();
        
        return yFwd;
    }

    private void projectForward() {

        subProblem.setVarRoles(xVars, yVars);
        
        //subProblem.setUnbounded(yVars);
        subProblem.relaxBounds(yVars, yL, yU, BOUND_RELAX_FACTOR);
        
        for (Solution s : allPartialSolutions) {
            
            TreeMap<String,Double> x = s.getValues(xVars);

            subProblem.fixVars(x);

            tryExtension(s, x);
        }
        
        // both x and y need to be reset (x was fixed, y was set unbounded)
        
        subProblem.setBounds(lowerBounds, upperBounds);
    }
    
    private void tryExtension(Solution s, TreeMap<String,Double> x) {

        ArrayList<double[]> yAlreadyFound = new ArrayList<>(EXTENSION_ATTEMPTS);

        double[] y = new double[yVars.size()];

        for (int i=0; (yAlreadyFound.size()<EXPECTED_OUTPUT) && i<EXTENSION_ATTEMPTS; ++i) {
            
            double[] yInitial = generateStartingPointWithLHS(x);

            Solution ySol = solver.solveSubNLPProblem(subProblem, yInitial, y);

            // TODO is delta_min a good threshold? Shouldn't it depend on the dimension as well? Square it?
            
            if (ySol.isSuccessful() && isSufficientlyDifferent(y, yAlreadyFound, delta_min) ) {

                saveForwardProjection(s, ySol, y, yAlreadyFound);
            }
        }
    }
    
    private double[] generateStartingPointWithLHS(TreeMap<String,Double> x) {
        
        ArrayList<double[]> samples = LatinHypercubeSampling.generateLHSamples(rnd, 30, yL, yU); // magic number
        
        double minViol = Double.POSITIVE_INFINITY;
        
        double[] best  = null;
        
        for (double[] sample : samples) {
        
            TreeMap<String,Double> z = new TreeMap<>(x);
        
            z.putAll(Util.createMap(yVars, sample));
        
            Violations viol = subProblem.allViolations(z, rowScales, columnScales);
            
            if (viol.LinfConsAndBound < minViol) {
                
                minViol = viol.LinfConsAndBound;
                
                best = sample;
            }
        }
        
        return best;
    }
    
    private void saveForwardProjection(Solution s,
                                       Solution ySol, 
                                       double[] y, 
                                       ArrayList<double[]> yAlreadyFound) 
    {
        Solution tmp =  s.extend(ySol.getAllVars(), ySol.getInfeasibilityBestObjective());
        
        tmp.setScaledBoundconstraintViolation( Util.scaledBoundViolationSquared(y, yL, yU, yScales) );
        
        tmp.setInfNormScaledBoundViol( Util.scaledBoundViolationInfinityNorm(y, yL, yU, yScales) );

        yFwd.add(tmp);
    
        double[] yCopy = new double[y.length];

        System.arraycopy(y, 0, yCopy, 0, y.length);

        yAlreadyFound.add(yCopy);
    }
    
    private boolean isSufficientlyDifferent(double[] y, ArrayList<double[]> yOthers, double threshold) {
        
        for (double[] yOther : yOthers) {
            
            double distanceSquared = Util.scaledDistanceSquared(y, yOther, yScales); 
            
            if (distanceSquared < threshold) {
            
                return false;
            }
        }
        
        return true;
    }
}

//=================================================================================================

final class Underdetermined {
    
    private final SubNLPProblem subProblem; // F(x,y)=0
    private final Solver solver;
    private final ArrayList<Solution> allPartialSolutions;
    private final SortedSet<String> xVars;
    private final SortedSet<String> yVars;
    private final Random rnd;
    
    private final TreeMap<String,Double> lowerBounds;
    private final TreeMap<String,Double> upperBounds;
    
    private final int EXTENSION_ATTEMPTS;
    private final int EXPECTED_OUTPUT; 
    private final double BOUND_RELAX_FACTOR;
    
    private final int samples_to_collect = 5; // magic number
    //private final double delta_min = 0.01; // magic number
    
    private final TreeMap<String,Double> columnScales;
    private final TreeMap<String,Double> rowScales;
//    private final double[] yScales;    
//    private double[] yL;    
//    private double[] yU;
    private final ArrayList<Solution> yFwd;
//    private ArrayList<double[]> yFwdArray; 

    Underdetermined(SubNLPProblem subProblem,
                    Solver solver,
                    ArrayList<Solution> allPartialSolutions,
                    SortedSet<String> xVars,
                    SortedSet<String> yVars,
                    TreeMap<String,Double> rowScales,
                    TreeMap<String,Double> columnScales,
                    Random rnd,
                    int EXTENSION_ATTEMPTS,
                    int EXPECTED_OUTPUT,
                    double BOUND_RELAX_FACTOR )
    {
        assert yVars.size() > subProblem.getNumberOfConstraints();
        
        subProblem.dbgAllVarsCovered(xVars, yVars);
        
        this.subProblem = subProblem;
        this.solver = solver;
        this.allPartialSolutions = allPartialSolutions;
        this.xVars = xVars;
        this.yVars = yVars;
        this.columnScales = columnScales;
        this.rowScales = rowScales;
        this.rnd = rnd;
        
        this.EXTENSION_ATTEMPTS = EXTENSION_ATTEMPTS;
        this.EXPECTED_OUTPUT    = EXPECTED_OUTPUT;
        this.BOUND_RELAX_FACTOR = BOUND_RELAX_FACTOR;
        
        lowerBounds = subProblem.allLowerBounds();
        upperBounds = subProblem.allUpperBounds();
        
//        yScales = Util.extractValues(columnScales, yVars);
        
        yFwd = new ArrayList<>(samples_to_collect*allPartialSolutions.size());
    }

    ArrayList<Solution> run() {

        projectForwardUD();
        
        return yFwd;
    }

    private void projectForwardUD() {

        int toFix = yVars.size() - subProblem.getNumberOfConstraints();

        for (Solution s : allPartialSolutions) {

            TreeMap<String,Double> x = s.getValues(xVars);

            subProblem.fixVars(x); // will be restored by SquareOrOverdetermined

            ArrayList<TreeMap<String,Double>> varsToFixAdditionally 
                = subProblem.sampleUnderDeterminedWithQRP(rnd, yVars, rowScales, columnScales, samples_to_collect, toFix);

            for (TreeMap<String,Double> varsToFix : varsToFixAdditionally) {

                projectForward(s, varsToFix);
            }
        }
    }

    private void projectForward(Solution s, TreeMap<String,Double> varsToFix) {

        ArrayList<Solution> onePartialSol = new ArrayList<>(1);

        onePartialSol.add( s.extend(varsToFix, 0) );

        TreeSet<String> fixed = new TreeSet<String>(xVars);

        fixed.addAll(varsToFix.keySet());

        TreeSet<String> guessed = new TreeSet<String>(yVars);

        guessed.removeAll(varsToFix.keySet());

        SquareOrOverDetermined reparam 
            = new SquareOrOverDetermined(subProblem, lowerBounds, upperBounds, solver, onePartialSol, fixed, guessed, rowScales, columnScales, rnd, EXTENSION_ATTEMPTS, EXPECTED_OUTPUT, BOUND_RELAX_FACTOR);

        // How many extension attempts should be allowed?
        
        yFwd.addAll( reparam.run() );

        // More forward projected points may be generated than in the well-defined case, run candidate selector?

        onePartialSol.clear();
    }
}
