//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.CopyOption;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

public final class Util {
    
    private static final Logger log = LogManager.getLogger();

    public static void testAssertationsEnabled() {
        
        try {
            assert false;
        }
        catch (AssertionError e) {
            return;
        }
        
        System.err.println("Please add the -ea flag to the VM arguments!");
        System.exit(1);
    }
    
    private static ArrayList<String> readStringVectorFromFile(String fileName) throws IOException {
        
        ArrayList<String> result = new ArrayList<>();
        
        try (BufferedReader reader = new BufferedReader(new FileReader(fileName))) {
            
            String line;

            while ((line=reader.readLine()) != null) {
                
                result.add(line);
            }
        } 
        
        return result;
    }
    
    public static ArrayList<String> stringVectorFromFile(String fileName) {
        
        try {
            
            return readStringVectorFromFile(fileName);
        }
        catch (FileNotFoundException e) {
            
            throw logIOError(e, fileName);
        }
        catch (IOException e) {
  
            throw logIOError(e, fileName, "reading");
        }
    }
    
    private static ArrayList<String[]> readSparseStringMatrixFromFile(String fileName) throws IOException {
        
        ArrayList<String[]> result = new ArrayList<>();
         
        try (BufferedReader reader = new BufferedReader(new FileReader(fileName))) {
            
            String line;

            while ((line=reader.readLine()) != null) {
                
                result.add(line.split("\t"));
            }
        } 
        
        return result;
    }
    
    public static ArrayList<String[]> sparseStringMatrixFromFile(String fileName) {
        
        try {
            
            return readSparseStringMatrixFromFile(fileName);
        }
        catch (FileNotFoundException e) {
            
            throw logIOError(e, fileName);
        }
        catch (IOException e) {
  
            throw logIOError(e, fileName, "reading");
        }
    }
    
    public static String filter(String name) {
        
        return name.replaceAll("[\\[\\]\\,]", "_");
    }
    
    public static ArrayList<String> filterNameVector(List<String> names) {
        
        ArrayList<String> result = new ArrayList<>();

        for (String name : names) {

            result.add(filter(name));
        }

        return result;
    }
    
    public static ArrayList<String> filterNameArray(String[] names) {
        
        return filterNameVector(Arrays.asList(names));
    }

    public static ArrayList<String[]> filterNameMatrix(ArrayList<String[]> nameMatrix) {
        
        ArrayList<String[]> result = new ArrayList<>();

        for (String[] names : nameMatrix) {
            
            ArrayList<String> filtered = filterNameVector(Arrays.asList(names)); 
            
            result.add(filtered.toArray(new String[filtered.size()]));
        }

        return result;
    }

    public static int powerInteger(int theBase, int theExponent) {

        int base = theBase; // just because the warning cannot be suppressed :( 
        
        int exponent = theExponent;
        
        if (exponent < 1) {
            
            throw new IllegalArgumentException("The exponent should be at least 1 but it is "+exponent);
        }
        
        final double predicted = (int) Math.pow(base, exponent); // TODO Remove Math.pow when done with testing 

        int result = 1;
        
        while (exponent!=0) {
            
            if ((exponent&1)!=0) {
                result *= base;                
            }

            exponent >>= 1;
        
            base *= base;
        }

        assert result == predicted;
        
        return result;
    }
    
    public static int numberOfDigits(int x) {
        assert x > 0 : "The argument should be at least 1 but it is "+x;
        int number = x; 
        int digits = 0;        
        while (number!=0) {
            number /= 10;
            ++digits;
        }
        return digits;
    }
    
    public static <T> double[] values(TreeMap<T,Double> map) {

        double[] values = new double[map.size()];

        int k = 0;

        for (double d : map.values()) {

            values[k++] = d;
        }

        return values;
    }
    
    /** asserts that all keys have non-null mapping */
    public static double[] extractValues(TreeMap<String,Double> map, Collection<String> keys) {
        
        double[] values = new double[keys.size()];

        int k = 0;

        for (String key : keys) {

            Double val = map.get(key);
            
            assert val!=null : "Key \""+key+"\" not found";
            
            values[k++] = val;
        }

        return values;
    }
    
    public static double[] extractScaledValues(TreeMap<String,Double> map, Collection<String> keys, double[] scales) {
        
        assert scales.length==keys.size();
        
        double[] values = new double[keys.size()];

        int k = 0;

        for (String key : keys) {

            Double val = map.get(key);
            
            assert val!=null : "Key \""+key+"\" not found";
            
            values[k] = scales[k]*val;
            
            ++k;
        }

        return values;
    }
    
    public static double[] extractScaledValues(TreeMap<String,Double> x, TreeMap<String,Double> scales) {
        
        double[] values = new double[scales.size()];

        int k = 0;

        for (Map.Entry<String,Double> e : scales.entrySet()) {
            
            String key = e.getKey();

            Double val = x.get(key);
            
            assert val!=null : "Key \""+key+"\" not found";
            
            values[k++] = e.getValue()*val;
        }

        return values;
    }
    
    public static TreeMap<String,Double> extractKeyValues(TreeMap<String,Double> map, Set<String> keys) {
        
        TreeMap<String,Double> res = new TreeMap<>();

        for (String key : keys) {

            Double val = map.get(key);
            
            assert val!=null : "Key \""+key+"\" not found";
            
            res.put(key, val);
        }

        return res;
    }
    
    public static TreeMap<String,Double> createMap(Collection<String> keys, double[] values) {
        
        assert keys.size()==values.length;
        
        TreeMap<String,Double> result = new TreeMap<>();
        
        int k=0;
        
        for (String s : keys) {
            
            Double prev = result.put(s, values[k++]);
            
            assert prev==null;
        }
        
        return result;
    }

    public static TreeMap<String,String> toHexString(TreeMap<String,Double> x) {
        
        TreeMap<String,String> m = new TreeMap<>();
        
        for (Map.Entry<String,Double> e : x.entrySet()) {
            
            m.put(e.getKey(), Double.toHexString(e.getValue()));
        }
        
        return m;
    }
    
    public static double infinityNorm(double[] x) {
        
        double norm = Math.abs(x[0]);
 
        for (int i=1; i<x.length; ++i) {
            
            double abs_xi = Math.abs(x[i]);            
            
            if (norm < abs_xi) {
                norm = abs_xi;
            }
        }
        
        return norm;
    }
    
    public static double L2Norm(double[] x) {
        
        assert x.length > 0;
        
        double sum = 0.0;
        
        for (double d : x) {
            
            sum += (d*d);
        }
        
        return Math.sqrt(sum);
    }
    
    public static double distanceSquared(double[] x, double[] y) {

        int n = x.length;

        assert y.length == n;

        double sum = 0.0;

        for (int k=0; k<n; ++k) {

            double diff = (x[k]-y[k]);

            sum += ( diff*diff );
        }

        return sum;
    }
    
    public static int findNearest(double[] x, ArrayList<double[]> others) {
        
        double minDist = Double.MAX_VALUE;
        
        int index = -1;

        for (int k=0, n=others.size(); k<n; ++k) {
            
            double d = distanceSquared(x, others.get(k));
            
            if (d < minDist) {
                
                minDist = d;
                
                index = k; 
            }
        }
        
        assert index!=-1;
        
        return index;
    }
    
    public static double distSquaredToNN(double[] x, ArrayList<double[]> others) {
        
        double minDist = Double.MAX_VALUE;

        for (double[] y : others) {
            
            double d = distanceSquared(x, y);
            
            if (d < minDist) {
                
                minDist = d;
            }
        }
        
        assert minDist!=Double.MAX_VALUE;
        
        return minDist;
    } 
    
    public static double distSquaredToNN(ArrayList<double[]> xArray, int index) {
        
        double min = Double.MAX_VALUE;
        
        double[] y = xArray.get(index);

        for (int i=0; i<xArray.size(); ++i) {

            if (i!=index) {
             
                double d = distanceSquared(xArray.get(i), y);
                
                if (d < min) {
                    
                    min = d;
                }                    
            }
        }
        
        assert min!=Double.MAX_VALUE;
        
        return min;
    }
    
    public static double[] distSquaredDistribution(ArrayList<double[]> x) {
        
        int n = x.size();
        
        assert n >= 2 : "n = "+n;
        
        double[] dNN = new double[n];
        
        for (int i=0; i<n; ++i) {
            
            dNN[i] = distSquaredToNN(x, i);           
        }      
        
        Arrays.sort(dNN);
        
        return dNN;
    }
    
    public static boolean areAllZeros(double[] x) {

        for (double e : x) {
            if (e!=0.0)
                return false;
        }
        return true;
    }

    public static double[] add(double[] x, double[] y) {

        int n = x.length;

        assert y.length == n;

        double[] z = new double[n];

        for (int k=0; k<n; ++k) {

            z[k] = x[k]+y[k];
        }

        return z;
    }
    
    public static double[] substract(double[] x, double[] y) {

        int n = x.length;

        assert y.length == n;

        double[] z = new double[n];

        for (int k=0; k<n; ++k) {

            z[k] = x[k]-y[k];
        }

        return z;
    }
    
    public static double[] mult(double c, double[] x) {

        int n = x.length;

        double[] z = new double[n];

        for (int k=0; k<n; ++k) {

            z[k] = c*x[k];
        }

        return z;
    }
    
    public static double[] inplaceMult(double c, double[] x) {
        for (int k=0; k<x.length; ++k)
            x[k]*= c;
        return x;
    }
    
    public static void inplaceIncrement(double[] sum, double[] x) {
        assert sum.length==x.length;
        for (int k=0; k<x.length; ++k)
            sum[k] += x[k];
    }
    
    public static double[] average(ArrayList<double[]> x) {
        double[] average = new double[x.get(0).length];
        for (double[] e : x)
            inplaceIncrement(average, e);
        return inplaceMult(1.0/x.size(), average);
    }
    
    public static double[][] mult(double[][] A, double[][] B) {
        final int n = A.length;
        final int m = A[0].length;
        final int p = B[0].length;
        assert m==B.length : "Invalid sizes: ("+n+","+m+")*("+B.length+","+p+")";
        double[][] C = new double[n][p];
        for (int i=0; i<n; ++i) {
            double[] Ai = A[i]; 
            double[] Ci = C[i];
            for (int j=0; j<p; ++j) {
                double sum = 0.0;
                for (int k=0; k<m; ++k)
                    sum += Ai[k]*B[k][j]; 
                Ci[j] = sum;
            }
        }
        return C;
    }
    
    public static double[] mult(double[][] A, double[] x) {
        final int n = A.length;
        final int m = A[0].length;
        assert m==x.length : "Invalid sizes: "+m+" and "+x.length;
        double[] c = new double[n];
        for (int i=0; i<n; ++i) {
            double[] Ai = A[i]; 
            double sum = 0.0;
            for (int k=0; k<m; ++k)
                sum += Ai[k]*x[k]; 
            c[i] = sum;
        }
        return c;       
    }
    
    public static double[] scale(double[] scales, double[] x) {
        
        int n = x.length;
        
        assert n==scales.length;
        
        double[] result = new double[n];
        
        for (int i=0; i<n; ++i) {
        
            result[i] = scales[i]*x[i];
        }
        
        return result;
    }
    
    public static double[] unscale(double[] scales, double[] x) {
        
        int n = x.length;
        
        assert n==scales.length;
        
        double[] result = new double[n];
        
        for (int i=0; i<n; ++i) {
        
            result[i] = x[i]/scales[i];
        }
        
        return result;
    }
    
    public static double scaledDistanceSquared(double[] x, double[] y, double[] scales) {

        int n = x.length;

        assert y.length == n;
        assert scales.length == n;

        double sum = 0.0;

        for (int k=0; k<n; ++k) {

            double diff = scales[k]*(x[k]-y[k]);

            sum += ( diff*diff );
        }

        return sum;
    }
    
    public static double scaledEuclideanDistance(double[] x, double[] y, double[] scales) {

        return Math.sqrt( scaledDistanceSquared(x, y, scales) );
    }
    
    public static double scaledChebyshevDistance(double[] x, double[] y, double[] scales) {

        int n = x.length;

        assert y.length == n;
        assert scales.length == n;

        double maxDiff = 0.0;

        for (int k=0; k<n; ++k) {

            double diff = Math.abs( scales[k]*(x[k]-y[k]) );

            if (diff > maxDiff) {
                
                maxDiff = diff;
            }
        }

        return maxDiff;
    }
    
    public static boolean isBetweenBounds(double[] x, double[] lb, double[] ub) {
     
        int n = x.length;
        
        assert n==lb.length;
        assert n==ub.length;
        
        for (int i=0; i<n; ++i) {
            
            if (x[i]<lb[i] || x[i]>ub[i])
                return false;
        }
        
        return true;
    }
    
    public static double[] projectToBox(double[] x, double[] lb, double[] ub) {
        
        int n = x.length;
        
        assert n==lb.length;
        assert n==ub.length;
        
        double[] x_projected = new double[n];
        
        for (int i=0; i<n; ++i) {
            
            if (x[i]<lb[i])    {
                x_projected[i] = lb[i];   
            }
            else if (x[i]>ub[i]) {
                x_projected[i] = ub[i];                
            }
            else {
                x_projected[i] =  x[i];
            }
        }
        
        return x_projected;        
    }
    
    public static double scaledBoundViolationSquared(double[] x, double[] xL, double[] xU, double[] xScales) {

        final int n = x.length;

        assert n==xL.length;
        assert n==xU.length;
        assert n==xScales.length;

        double boundViol = 0.0;

        for (int i=0; i<n; ++i) {

            double viol = xScales[i] * (Math.max(xL[i] - x[i], 0.0) + Math.max(x[i] - xU[i], 0.0));

            boundViol += viol*viol;
        }

        return boundViol;
    }

    public static double scaledBoundViolationInfinityNorm(double[] x, double[] xL, double[] xU, double[] xScales) {

        final int n = x.length;

        assert n==xL.length;
        assert n==xU.length;
        assert n==xScales.length;

        double maxViol = 0.0;

        for (int i=0; i<n; ++i) {

            double viol = xScales[i] * (Math.max(xL[i] - x[i], 0.0) + Math.max(x[i] - xU[i], 0.0));

            if (viol > maxViol) {

                maxViol = viol;
            }
        }

        return maxViol;
    }
    
    public static int lowerPercentile(double q, int size) {
        
        assert ((q>=0.0) && (q<=1.0)) : "q = "+q;
        
        assert size >= 1 : "size = "+size;
        
        int ret = (int) Math.floor(q*(size-1));
        
        assert ret < size;
        
        return ret;        
    }
    
    public static int upperPercentile(double q, int size) {
        
        assert ((q>=0.0) && (q<=1.0)) : "q = "+q;
        
        assert size >= 1 : "size = "+size;
        
        int ret = (int) Math.ceil(q*(size-1));
        
        assert ret < size;
        
        return ret;        
    }

    public static int lowerMedian(int size) {
        
        return lowerPercentile(0.5, size);
    }    
    
    public static int upperMedian(int size) {
        
        return upperPercentile(0.5, size);
    }
    
    public static void sort_and_show_statistics(double[] arr) {
        
        Arrays.sort(arr);
        
        int q1 = lowerPercentile(0.25, arr.length);
        
        int med = upperMedian(arr.length);
        
        int q3 = upperPercentile(0.75, arr.length); 
        
        System.out.printf("size=%d  0/25/50/75/100%% : %5.3f  %5.3f  %5.3f  %5.3f  %5.3f\n",arr.length,arr[0],arr[q1],arr[med],arr[q3],arr[arr.length-1]);
    }

    public static double max(double... args) {
        
        assert args.length!=0;
        
        double max = args[0];
        
        for (int i=1; i<args.length; ++i) {
            
            double val = args[i];
            
            if (val > max) {
                
                max = val;
            }
        }
        
        return max;
    }

    public static String arrayToString(double[] arr) {
        
        return arrayToString(arr, "%g");
    }
    
    /** Conversion character %g for example; see java.util.Formatter for more */
    public static String arrayToString(double[] arr, String format) {
        
        StringBuilder builder = new StringBuilder();
        
        builder.append("[");
        
        for (int i=0; i<arr.length; ++i) {
            
            builder.append(String.format(format, arr[i]));
            
            builder.append((i<arr.length-1) ? "," : "");
        }
        
        builder.append("]");
        
        return builder.toString();
    }
    
    public static String matrixToString(double[][] mat) {
        
        return matrixToString(mat, "%6.3f");
    }
    
    /** Conversion character %5.3f for example; see java.util.Formatter for more */
    public static String matrixToString(double[][] mat, String format) {
        
        StringBuilder builder = new StringBuilder();
        
        for (double[] row : mat) {
            
            builder.append("[");

            for (int j=0; j<row.length; ++j) {

                builder.append(String.format(format, row[j]));

                builder.append((j<row.length-1) ? "," : "");
            }

            builder.append("]\n");
        }

        return builder.toString();
    }
    
    public static String toString(int[] arr) {
        
        StringBuilder builder = new StringBuilder();
        
        builder.append("[");
        
        for (int i=0; i<arr.length; ++i) {
            
            builder.append(arr[i]);
            
            builder.append((i<arr.length-1) ? "," : "");
        }
        
        builder.append("]");
        
        return builder.toString();        
    }
    
    public static double[] toArray(Collection<Double> list) {
        
        double[] array = new double[list.size()];
        
        int k=0;
        
        for (double d : list) {
            
            array[k++] = d;
        }
        
        return array;
    }
    
    public static <T> ArrayList<T> copyHead(ArrayList<T> a, int lengthAtMost) {
        
        assert lengthAtMost > 0;
        
        ArrayList<T> res = new ArrayList<>();
        
        int n = Math.min(lengthAtMost, a.size());
        
        res.ensureCapacity(n);
        
        for (int i=0; i<n; ++i) {
            
            res.add(a.get(i));
        }
        
        return res;
    }
    
    public static double[] copyHead(double[] a, int length) {
        
        assert a.length >= length : "Array length is "+a.length+" and asked to copy "+length+" elements";
    
        double[] res = new double[length];
        
        System.arraycopy(a, 0, res, 0, length);
        
        return res;
    }
    
    public static TreeMap<String,Double> clip(String[] keys, double[] values) {
        
        assert keys.length==values.length;
        
        TreeMap<String,Double> result = new TreeMap<>();

        for (int i=0; i<keys.length; ++i) {
            
            Double prev = result.put(keys[i], values[i]);
            
            assert prev==null;
        }
        
        return result;
    }

    public static void writeToFile(String fileName, ArrayList<String> content) {
        
        try (BufferedWriter writer = new BufferedWriter(new FileWriter(fileName))) {
            
            for (String s : content) {
                
                writer.write(s);
                
                writer.newLine();
            }
        }
        catch (IOException e) {
            
            throw logIOError(e, fileName, "writing");
        }
    }

    public static void executeCommand(String cmd){

        try {
            
            Process p = Runtime.getRuntime().exec(cmd);

            p.waitFor();
        }
        catch(IOException | InterruptedException e) {

            e.printStackTrace();
            
            log.fatal("Error executing \""+cmd+"\"");
            
            throw new ProgramTerminationRequested();
        }
    }

    private static ProgramTerminationRequested logIOError(FileNotFoundException e, String fileName) {
        
        e.printStackTrace();
        
        log.fatal("File \""+fileName+"\" not found");
        
        return new ProgramTerminationRequested();
    }
    
    public static ProgramTerminationRequested logIOError(IOException e, String fileName, String readingOrWriting) {
        
        e.printStackTrace();
        
        log.fatal("Fatal error "+readingOrWriting+" file "+fileName);
        
        return new ProgramTerminationRequested();       
    }

    public static void testLoggerEnabled() {
        
        if (!log.isInfoEnabled()) {
            System.err.println("You might wish to add -Dlog4j.configurationFile=config/log4j2.xml to the VM arguments");
        }        
    }

    public static void printLocation() {

        try {
            throw new Exception();
        }
        catch (Exception e) {
            StackTraceElement[] ste = e.getStackTrace();
            System.out.println(ste[1]);
        }
    }

    public static String toTimeStamp(long timestamp) {
        
        long millis = timestamp / 1000000;
        
        long second = (millis / 1000) % 60;
        long minute = (millis / (1000 * 60)) % 60;
        long hour = (millis / (1000 * 60 * 60)) % 24;

        return String.format("%02d:%02d:%02d.%03d", hour, minute, second, millis%1000);
    }
    
    public static void copyFile(String src, String dst, CopyOption... options) {
        
        if (src.equals(dst)) {
            
            return;
        }

        try {
            
            Files.copy(new File(src).toPath(), new File(dst).toPath(), options);
            
        } catch (IOException e) {
            
            e.printStackTrace();
            
            log.fatal("Error copying file \""+src+"\" to \""+dst+"\"");
            
            throw new ProgramTerminationRequested();
        }
    }
    
    public static void executeCmd(String cmd) {
        
        try {
            
            Runtime.getRuntime().exec(cmd);
            
        } catch (IOException e) {

            e.printStackTrace();
            
            log.fatal("Failed to execute command \""+cmd+"\"");
            
            throw new ProgramTerminationRequested();
        }
    }
    
    public static void shuffle(int[] arr, Random rnd) {
        
        for (int i=arr.length-1; i>0; --i) {
            
            int j = rnd.nextInt(i+1);
            
            int tmp = arr[i];
            
            arr[i] = arr[j];
            
            arr[j] = tmp;            
        }        
    }
}
