//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import sdms.CandidateSelector;
import sdms.Solution;
import sdms.Util;

interface NormSelector {
    
    double getNorm(Solution s);
    
    double getDistance(double[] x, double[] y, double[] scales);
}

final class SubproblemNorm implements NormSelector {

    @Override
    public double getNorm(Solution s) { return s.getNormOfInfeasibility(); } // returns the L2 norm

    @Override
    public double getDistance(double[] x, double[] y, double[] scales) {
        
        return Util.scaledEuclideanDistance(x, y, scales);
    }    
}

final class CumulativeNorm implements NormSelector {
    
    @Override
    public double getNorm(Solution s) { return Math.sqrt(s.getCumulativeInfeasibility()); }

    @Override
    public double getDistance(double[] x, double[] y, double[] scales) {
        
        return Util.scaledEuclideanDistance(x, y, scales);
    }
}

final class InfinityNorm implements NormSelector {
    
    @Override
    public double getNorm(Solution s) { return s.getInfeasibilityInfinityNorm(); }

    @Override
    public double getDistance(double[] x, double[] y, double[] scales) {
        
        return Util.scaledChebyshevDistance(x, y, scales);
    }
}

//=====================================================================
//
// TODO Minimum distance from other solutions seems to be ignored;
// Regularization term for the infeasibility score seems to be unused, 
//   see the callers;
// There is a mess with the Apache logger and sysout, keep only one
//
//=====================================================================

final class CandidateSelectorImpl implements CandidateSelector {

    private static final Logger log = LogManager.getLogger();    
    
    private ArrayList<Solution> solutions;
    
    private int size;
    
    private double[][] values;
    
    private double[] scales;
    
    private ArrayList<double[]> distances;
    
    private int[] chosen;
    
    private int[] notYetChosen;
    
    private int nPoints;
    
    private double maxDistance;
    
    private double scaleInfeas;
    
    private NormSelector norm;
    
    static CandidateSelector createSubproblemEvaluator(double[] scales) {
        CandidateSelectorImpl cs = new CandidateSelectorImpl();
        cs.norm = new SubproblemNorm();
        cs.scales = scales;
        return cs;
    }
    
    static CandidateSelector createCumulativeEvaluator(double[] scales) {
        CandidateSelectorImpl cs = new CandidateSelectorImpl();
        //cs.norm = new CumulativeNorm();
        cs.norm = new InfinityNorm();
        cs.scales = scales;
        return cs;
    }
    
    private CandidateSelectorImpl() {
        
    }    
 
    @Override
    public void selectMostPromisingOnlyByDistance(ArrayList<Solution> partialSolutions, int pointsToPropagate, double INFEAS_REG) {
        
        // Infeasibility is computed only for monitoring the infeasibility distribution; it is ignored otherwise 
        
        init(partialSolutions, pointsToPropagate);
        
        setScaleInfeasBasedOnInfeasDistribution(INFEAS_REG);
        
        double minimumDistanceFromOtherSolutions = -1.0; // Here we ignore delta_min

        // runs from 0 to nPoints-1 but 1 point (the least infeasible) is already chosen
        for (int i=1; i<nPoints; ++i) {
            
            double Winfeas = 0.0;
            
            addMostPromisingOne(minimumDistanceFromOtherSolutions, Winfeas);
        } 
        
        return;
    }
    
    @Override
    public void selectMostPromisingCandidates(ArrayList<Solution> partialSolutions, int pointsToPropagate, double infeasibilityWeight) {
        
        init(partialSolutions, pointsToPropagate);
        
        double minimumDistanceFromOtherSolutions = -1.0; // Here we ignore delta_min

        // runs from 0 to nPoints-1 but 1 point (the least infeasible) is already chosen
        for (int i=1; i<nPoints; ++i) {
            
            addMostPromisingOne(minimumDistanceFromOtherSolutions, infeasibilityWeight);
        } 
    }
    
    private void init(ArrayList<Solution> sols, int pointsKeepAtMost) {
        
        solutions = sols;
        
        size = solutions.size();
        
        dbgNames();
        
        initDistances();
        
        traceDistances();
        
        initIndexSets();
        
        findMostAndLeastInfeasible();
        
        nPoints = Math.min(size, pointsKeepAtMost);
        
        assert (maxDistance>=0.0) || (nPoints==1);
        
        maxDistance = Math.max(maxDistance, Double.MIN_VALUE);
    }

    private void dbgNames() {

        if (!log.isInfoEnabled())
            return;

        for (int i=0; i<size; ++i) {

            for (int j=i+1; j<size; ++j) {

                Iterator<String> a = solutions.get(i).getVarNames().iterator();
                Iterator<String> b = solutions.get(j).getVarNames().iterator();
                
                while (a.hasNext()) {
                    String x = a.next();
                    String y = b.next();
                    assert x.equals(y) : x + " != " + y;
                }
                
                assert !b.hasNext();
            }
        }
    }

    private void initDistances() {
        
        if (log.isDebugEnabled()) log.debug("size = " + size);
        
        initValues();
        
        distances = new ArrayList<double[]>();

        for (int i=0; i<size; ++i) {
            
            distances.add(new double[i]);            
        }
        
        maxDistance = -1;
         
        for (int i=0; i<size; ++i) {
           
            for (int j=i+1; j<size; ++j) {

                double d = computeDistance(i, j);
                
                distances.get(j)[i] = d;
                
                if (d > maxDistance) {
                    
                    maxDistance = d;
                }
            }
        }
       
        if (log.isDebugEnabled()) log.debug("max distance " + maxDistance);
    }
    
    private void initValues() {
        
        values = new double[size][solutions.get(0).dimension()];
        
        for (int i=0; i<size; ++i) {
            
            double[] val = values[i];
        
            int k = 0;
            
            Collection<Double> solutionValues = solutions.get(i).getValues(); 

            for (double d : solutionValues) {

                val[k++] = d;
            }
            
            assert k==val.length;
        }
    }

    private double computeDistance(int i, int j) {
        
        return norm.getDistance(values[i], values[j], scales);
    }
    
    private void traceDistances() {

        if (!log.isTraceEnabled())
            return;

        for (int i=0; i<size; ++i) {

            log.info(i+"   "+Util.arrayToString(distances.get(i)));
        }

        ArrayList<Double> dists = new ArrayList<>();

        for (int i=0; i<size; ++i) {

            for (double d : distances.get(i)) {
                dists.add(d);
            }
        }

        Collections.sort(dists);

        int n = dists.size();

        log.info("Min/0.1/Median/0.9/Max "+dists.get(               0 )+" / "+
                  dists.get((int)(  n/10.0))+" / "+
                  dists.get((int)(  n/2.0 ))+" / "+
                  dists.get((int)(9*n/10.0))+" / "+
                  dists.get(        n-1   )  );

        return;
    }
 
    private void initIndexSets() {
        
        chosen = new int[0];
        
        notYetChosen = new int[size];
        
        for (int i=0; i<size; ++i)
            notYetChosen[i] = i;        
    }
    
    private void addChosenIndex(int index) {
        
        // add index to the chosen ones
        int[] newChosen = new int[chosen.length+1];
        
        System.arraycopy(chosen, 0, newChosen, 0, chosen.length);
        
        newChosen[newChosen.length-1] = index;
        
        chosen = newChosen;
        
        // remove index from the not yet chosen ones
        int pos = Arrays.binarySearch(notYetChosen, index);
        
        assert pos >= 0 : "Index "+index+" not found";
        
        int[] newNotYetChosen = new int[notYetChosen.length-1];
        
        System.arraycopy(notYetChosen, 0, newNotYetChosen, 0, pos);
        
        System.arraycopy(notYetChosen, pos+1, newNotYetChosen, pos, newNotYetChosen.length-pos);
        
        notYetChosen = newNotYetChosen;
/*       
        TreeSet<Integer> indices = new TreeSet<>();
        
        for (int i : chosen)
            indices.add(i);
        
        for (int i : notYetChosen)
            indices.add(i);
        
        assert indices.size() == size : indices.toString();
*/        
    }
    
    private void findMostAndLeastInfeasible() {
        
        int minIndex = 0;
//        int maxIndex = 0;
        
        double min = norm.getNorm(solutions.get(0));
        double max = min;
        
        for (int i=1; i<size; ++i) {
            
            double infeas = norm.getNorm(solutions.get(i));
            
            if (infeas > max) {
                max = infeas;
//                maxIndex = i;
            }
            
            if (infeas < min) {
                min = infeas;
                minIndex = i;
            }
        }
 /*       
        if (log.isTraceEnabled()) {
            Util.printLocation();
            Solution sMin = solutions.get(minIndex);
            System.out.println("Min infeas at "+minIndex+" = "+norm.getNorm(sMin)+"  "+sMin.getAllVars());
            Solution sMax = solutions.get(maxIndex);
            System.out.println("Max infeas at "+maxIndex+" = "+norm.getNorm(sMax)+"  "+sMax.getAllVars());
        }
*/        
        addChosenIndex(minIndex);
        
        scaleInfeas = 1.0/(max + 0.01); // TODO Remove when done
    }
    
    private void setScaleInfeasBasedOnInfeasDistribution(double INFEAS_REG) {
        
        double[] residualDistribution = new double[size];

        for (int i=0; i<size; ++i) {
            
            residualDistribution[i] = norm.getNorm(solutions.get(i));
        }
        
        Arrays.sort(residualDistribution);
        
        assert residualDistribution[size-1]  < 1.0e+3 : "Suspiciously large infeasibility: "+residualDistribution[size-1];
        
        assert residualDistribution[0] >= 0.0 : "Invalid norm: "+residualDistribution[0];
//        
//      System.out.println("Unscaled infeasibility distribution "+Util.arrayToString(residualDistribution));
//        
        double q = 0.5; // magic number upper median in infeasibility scaling
        
        int qUpper = Util.upperPercentile(q, size);
        
        double infeas = residualDistribution[qUpper];
        
        double qEff = (qUpper+1.0) / size; 
        
        scaleInfeas = qEff / (infeas + qEff*INFEAS_REG);
        
        return;
    }
        
    private boolean addMostPromisingOne(double minimumDistanceFromOtherSolutions, double Winfeas) {
        
        double mostDistant = Double.MIN_VALUE;
        
        for (int i : notYetChosen) {
            
            double dMin = Double.MAX_VALUE;
            
            for (int j : chosen) {

                double d = (i<j) ? distances.get(j)[i] : distances.get(i)[j];
                
                if (d < dMin) {
                    
                   dMin = d;
                }
            }
            
            if (dMin > mostDistant) {
                
                mostDistant = dMin;
            }
        }
        
        maxDistance = mostDistant;
        
        assert maxDistance < 1.0e+3 : "maxDistance = "+maxDistance;
       
//
        double[] distTerms   = new double[notYetChosen.length];
        double[] infeasTerms = new double[notYetChosen.length];
        int distribPos = 0;
//           
        //----------------------------------------------------------------------------------            
         
        double objective   = Double.MAX_VALUE;
        double minDistance = Double.MAX_VALUE;
        int    index       = -1;
        
        for (int i : notYetChosen) {
            
            // find the distance to the closest chosen one; manually inlined function
            double distance = Double.MAX_VALUE;
            
            for (int j : chosen) {

                double d = (i<j) ? distances.get(j)[i] : distances.get(i)[j];
                
                if (d < distance) {
                    
                    distance = d;
                }
            }
            
            assert distance != Double.MAX_VALUE;            
            //---------------------------------------------------------------------------
            double infeas = norm.getNorm(solutions.get(i));
           
            assert infeas >= 0.0;
            
            double distancePenalty = (maxDistance-distance)/maxDistance;
            
            double infeasPenalty   = Winfeas*(scaleInfeas*infeas); 
            
            double penalty = distancePenalty + infeasPenalty; 
//            
            distTerms[distribPos]     = distancePenalty;
            infeasTerms[distribPos++] = infeasPenalty;            
//            
            assert !Double.isNaN(penalty) : "distance = "+distance+", infeas = "+infeas;
            assert penalty >= 0;
            
            if (penalty < objective) {
                objective = penalty;
                index = i;
                minDistance = distance;
            }            
        }
/*       
//        
        Arrays.sort(distTerms);
        Arrays.sort(infeasTerms);
        System.out.println("");
        System.out.println("Distance term distribution "+Util.arrayToString(distTerms));
        System.out.println("Infeasibility term distrib "+Util.arrayToString(infeasTerms));        
//
        if (log.isInfoEnabled()) { logResults(index, minDistance, objective, Winfeas); }
*/        
        if (minDistance > minimumDistanceFromOtherSolutions) {
            addChosenIndex(index);
            return true;            
        }
        return false;
    }
    
    @SuppressWarnings("unused")
    private void logResults(int index, double minDistance, double objective, double Winfeas) {
        
        Solution s = solutions.get(index);
        
        Util.printLocation();
        
        System.out.println("Most promising is index "+index+",  "+norm.getNorm(s)+"  "+s.getAllVars());
        System.out.println("Min distance  = "+minDistance);
        System.out.println("Max distance  = "+maxDistance);
        System.out.println("Distance term = "+(maxDistance-minDistance)/maxDistance);
        
        double infeas = norm.getNorm(solutions.get(index));
        
        System.out.println("Infeasiblity  = "+infeas);
        System.out.println("scaleInfeas   = "+scaleInfeas);
        System.out.println("Infeas. term  = "+Winfeas*(scaleInfeas*infeas));
        System.out.println("Penalty      = "+objective);        
    }
    
    @Override
    public ArrayList<Solution> getSelectedOnes() {

        ArrayList<Solution> kept = new ArrayList<>();
        
        for (int i : chosen) {
         
            kept.add(solutions.get(i));
        }   
        
        return kept;
    }
}
