//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Map;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.google.common.collect.Ordering;

import sdms.CandidateSelector;
import sdms.LatinHypercubeSampling;
import sdms.Solution;
import sdms.StartingpointGenerator;
import sdms.Util;
import sdms.Violations;
import tools.math.IndexList;
import tools.math.RealVector;
import dag.Constraint;
import dag.Dag;
import dag.Objective;
import dag.eval.interval.ConstraintIntervalEvalInfo;
import dag.eval.interval.DagIntervalEvalResult;
import dag.nodes.input.ParamNode;
import tools.math.interval.real.RealInterval;
import tools.math.interval.real.RealIntervalVector;
import no.uib.cipr.matrix.DenseLU;
import no.uib.cipr.matrix.DenseMatrix;
import no.uib.cipr.matrix.DenseVector;
import no.uib.cipr.matrix.Matrix.Norm;
import no.uib.cipr.matrix.QRP;
import no.uib.cipr.matrix.SVD;

final class VarBoundsInAlphabeticalOrder {
    VarBoundsInAlphabeticalOrder(double[] lb, double[] ub) {
        this.lb = lb;
        this.ub = ub;
    }
    double[] lb;
    double[] ub;
}

/** Row of the Jacobian at x; just a chunk of data to be passed around, no abstraction */
final class JacRow {

    JacRow(TreeMap<String,Double> x, double residual, TreeMap<String,Double> gradients, double rhs) {
        this.x         = x;
        this.residual  = residual;
        this.gradients = gradients;
        this.rhs       = rhs;
    }
    
    TreeMap<String,Double> x;
    double                 residual;
    TreeMap<String,Double> gradients;
    double                 rhs;
    
    @Override
    public String toString() {
        return "rhs = "+rhs+", res = "+residual+", grad  "+gradients;
    }

}

final class GradEvalRes {

    TreeMap<String,Double> gradValue;
    double[] xValue;

    GradEvalRes(TreeMap<String,Double> gradValue, double[] xValue) {
        this.gradValue = gradValue; 
        this.xValue = xValue;
    }
}

final class Evaluator {
    
    private static final Logger log = LogManager.getLogger();
    
    static private final double MIN_ROW_SCALE = 1.0e-6;
    static private final double MAX_ROW_SCALE = 1.0e+6;
    static private final double MIN_COL_SCALE = 1.0e-6;
    static private final double MAX_COL_SCALE = 1.0e+6;
    
    private Evaluator() { throw new AssertionError("This stateless class is just a collection of functions"); }
    
    static private RealVector toDagsInternalPermutation(Dag dag, TreeMap<String, Double> xVarnameValueMap) {

        IndexList indicesOfVarsInDag = dag.getVarIndices(xVarnameValueMap.keySet());

        return toInternalPermutation(indicesOfVarsInDag, xVarnameValueMap);
    }
    
    static private RealVector toInternalPermutation(IndexList indicesOfVarsInDag, TreeMap<String, Double> xVarnameValueMap) {

        return toInternalPermutation(indicesOfVarsInDag, Util.values(xVarnameValueMap));
    }
    
    static private RealVector toInternalPermutation(IndexList indicesOfVarsInDag, double[] xInAlphabeticalOrder) {
        
        assert  indicesOfVarsInDag.size()==xInAlphabeticalOrder.length;
        
        RealVector x = new RealVector(xInAlphabeticalOrder.length);

        for (int i = 0; i < xInAlphabeticalOrder.length; ++i) {

            int index = indicesOfVarsInDag.get(i);

            x.set(index, xInAlphabeticalOrder[i]);
        }

        return x;
    }
    
    static private double[] toInternalPermutationAsArray(IndexList indicesOfVarsInDag, double[] xInAlphabeticalOrder) {
        
        assert  indicesOfVarsInDag.size()==xInAlphabeticalOrder.length;
        
        double[] x = new double[xInAlphabeticalOrder.length];

        int k=0;
        
        for (int index : indicesOfVarsInDag) {

            x[index] = xInAlphabeticalOrder[k++];
        }

        return x;
    }
    
    static private void ASSERT_NARROW(RealInterval x) {
        
        Double xL = x.getLower();
        
        Double xU = x.getUpper();

        assert (-Double.MAX_VALUE < xL) && (xU < Double.MAX_VALUE): "xL = "+xL+", xU = "+xU;
        
        assert xL <= xU : "xL = "+xL+", xU = "+xU;
        
        double absWidth = xU - xL;
        
        double relWidth = absWidth / ((1e-300) + Math.max(Math.abs(xL), Math.abs(xU)));
        
        double width = Math.min(relWidth, absWidth);
        
        assert width < 1.0e-2 : 
            "Catastrophic loss of precision: ["+String.format("%g", x.getLower())+", "+String.format("%g", x.getUpper())+"]  ("+width+")";
    }
    
    // FIXME Disabled sanity check ASSERT_NARROW
    static private Double getResidual(ConstraintIntervalEvalInfo cei) {
        
        RealInterval leftHandSide = cei.getRangeHull();
        
        RealInterval rightHandSide = cei.getConstraint().getBounds();

        //ASSERT_NARROW(leftHandSide);
        
        //ASSERT_NARROW(rightHandSide);
        
        return leftHandSide.getMid() - rightHandSide.getMid();
    }
    
    static private Double getScaledResidual(ConstraintIntervalEvalInfo cei, TreeMap<String,Double> rowScales) {
        
        double residual = getResidual(cei);
        
        String con = cei.getConstraint().getLabel();
        
        Double scale = rowScales.get(con);
        
        assert scale!=null;
        
        return scale*residual;
    }
    
    /** Bound constraints igonored */
    static double infinityNormOfConstraintViolation(Dag dagCSP, RealVector x) {
        
        DagIntervalEvalResult evalRes = dagCSP.evalRan(new RealIntervalVector(x), 
                                               dagCSP.computeParamBounds(new RealIntervalVector(x), ParamNode.PARAM_EVAL_TYPE_LOCSOL));

        return infinityNormOfConstraintViolation(evalRes);
    }
    
    static private double infinityNormOfConstraintViolation(DagIntervalEvalResult evalRes) {

        double maxNorm = 0.0;

        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            double norm = Math.abs(getResidual(cei));
            
            if (norm > maxNorm) {
                
                maxNorm = norm;
            }
        }

        return maxNorm;
    }
    
    static private double infinityNormOfConstraintViolation(DagIntervalEvalResult evalRes, TreeMap<String,Double> rowScales) {

        double maxNorm = 0.0;

        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            double norm = Math.abs(getScaledResidual(cei, rowScales));
            
            if (norm > maxNorm) {
                
                maxNorm = norm;
            }
        }

        return maxNorm;
    }
    
    static private double sumOfSquaredConstraintViolation(Dag dagCSP, double[] x, TreeMap<String,Double> rowScales) {
        
        DagIntervalEvalResult evalRes = dagCSP.evalRan(new RealIntervalVector(x), 
                                               dagCSP.computeParamBounds(new RealIntervalVector(x), ParamNode.PARAM_EVAL_TYPE_LOCSOL));

        return sumOfSquaredConstraintViolation(evalRes, rowScales);
    }
    
    static private double sumOfSquaredConstraintViolation(DagIntervalEvalResult evalRes, TreeMap<String,Double> rowScales) {

        double norm = 0.0;

        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            double residual = getScaledResidual(cei,rowScales);

            norm += residual*residual;
        }

        return norm;
    } 
    
    static TreeMap<String, Double> evaluateResidual(Dag dagCSP, TreeMap<String, Double> xVarnameValueMap) {

        RealVector x = toDagsInternalPermutation(dagCSP, xVarnameValueMap);

        DagIntervalEvalResult evalRes = dagCSP.evalRan(new RealIntervalVector(x), 
                                               dagCSP.computeParamBounds(new RealIntervalVector(x), ParamNode.PARAM_EVAL_TYPE_LOCSOL));

        TreeMap<String, Double> residuals = new TreeMap<>();

        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            String constraintName = cei.getConstraint().getLabel();
            
            residuals.put(constraintName, getResidual(cei));
        }

        return residuals;
    }
    
    /** Assumes that Dag is a bound constraint problem, points within the bounds */
    static ArrayList<GradEvalRes> evaluateGradients(Dag dagOpt, SortedSet<String> names, ArrayList<double[]> values) {
        
        assert dagOpt.getObjectiveNumber()==1;

        IndexList idxDag = dagOpt.getVarIndices(names);
        
        TreeMap<Integer,String> indexToName = createReverseMap(idxDag, names);

        ArrayList<GradEvalRes> result = new ArrayList<>();

        for (double[] x : values) {

            result.add(new GradEvalRes(evaluateGradients(dagOpt, toInternalPermutation(idxDag, x), indexToName), x));
        }

        return result;
    }

    static private TreeMap<String,Double> evaluateGradients(Dag dagOpt, RealVector x, TreeMap<Integer,String> indexToName) {
        
        Objective obj = dagOpt.getObjectives().get(0);

        DagIntervalEvalResult info = dagOpt.evalGrad(new RealIntervalVector(x),
                                             dagOpt.computeParamBounds(new RealIntervalVector(x),
                                             ParamNode.PARAM_EVAL_TYPE_LOCSOL), obj);
        
        assert info.isFeasible() : "We are supposed to evaluate the gradients at the most promising points, thus, should be feasible!";
        
        TreeMap<Integer,RealInterval> grad = info.get(obj).getGradientRange().getData();

        return permuteBack(grad, indexToName);
    }

    static private TreeMap<String, JacRow> evaluateSparseJacobian(Dag dagCSP, TreeMap<String, Double> xVarnameValueMap) {
        
        SortedSet<String> varNames = xVarnameValueMap.navigableKeySet();
        
        IndexList idxDag = dagCSP.getVarIndices(xVarnameValueMap.keySet());
        
        TreeMap<Integer,String> indexToName = createReverseMap(idxDag, varNames);

        RealVector x = toInternalPermutation(idxDag, xVarnameValueMap);

        DagIntervalEvalResult evalRes = dagCSP.evalGrad(new RealIntervalVector(x),
                                                dagCSP.computeParamBounds(new RealIntervalVector(x), ParamNode.PARAM_EVAL_TYPE_LOCSOL));

        TreeMap<String, JacRow> jac = new TreeMap<>();

        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            Constraint con = cei.getConstraint();

            String constraintName = con.getLabel();

            TreeMap<Integer,RealInterval> grad = cei.getGradientRange().getData();
            
            double rhs = con.getBounds().getMid();

            JacRow prev = jac.put(constraintName, new JacRow(xVarnameValueMap, getResidual(cei), permuteBack(grad, indexToName), rhs));
            
            assert prev==null;
        }

        return jac;
    }
    
    static private TreeMap<Integer,String> createReverseMap(IndexList idxDag, SortedSet<String> varNames) {
        
        TreeMap<Integer,String> indexToName = new TreeMap<>();
        
        int i=0;
        
        for (String name : varNames) {
            
            String prev = indexToName.put(idxDag.get(i++), name);
            
            assert prev==null : "Index " + idxDag.get(i-1) + " is already in the names with \"" + name + "\"";
        }
        
        return indexToName;
    }

    static private TreeMap<String,Double> permuteBack(TreeMap<Integer,RealInterval> x, TreeMap<Integer,String> indexToName) {
        
        TreeMap<String,Double> result = new TreeMap<>();
        
        for (Map.Entry<Integer,RealInterval> e : x.entrySet()) {
            
            String name = indexToName.get(e.getKey());
            
            assert name!=null;
            
            RealInterval range = e.getValue();
            
            ASSERT_NARROW(range);
            
            Double prev = result.put(name, range.getMid());
            
            assert prev==null;
        }
       
        return result;
    }
    
    static private VarBoundsInAlphabeticalOrder getVarBounds(Dag dag, TreeSet<String> varNames) {

        int n = varNames.size();
        
        double[] lb = new double[n];
        
        double[] ub = new double[n];
        
        int k=0;
        
        for (String var : varNames) {
            
            RealInterval bnds = dag.getVar(var).getBounds(); 
            
            lb[k] = bnds.getLower();
            
            ub[k] = bnds.getUpper();
            
            assert lb[k] <= ub[k];
            
            ++k;
        }
        
        return new VarBoundsInAlphabeticalOrder(lb, ub);
    }
    
    static private StartingpointGenerator setupStartingPointGenerator(Dag dag, TreeSet<String> varNames) {

        VarBoundsInAlphabeticalOrder bnds = getVarBounds(dag, varNames);

        StartingpointGenerator spg = new StartingpointGeneratorImpl();
        
        spg.setVarBounds(bnds.lb, bnds.ub);
        
        return spg;
    }
    
    static private TreeMap<String,Double> newPoint(TreeSet<String> varNames, double[] x) {
        
        assert varNames.size() == x.length;
        
        TreeMap<String,Double> result = new TreeMap<>();
        
        int k=0;
        
        for (String s : varNames) {
            
            result.put(s, x[k++]);
        }
        
        return result;
    }
    
    static private ArrayList<TreeMap<String,JacRow>> generateRandomJacobians(Dag dagCSP, TreeSet<String> varNames) {
        
        StartingpointGenerator randomPointGen = setupStartingPointGenerator(dagCSP, varNames);
        
        ArrayList<TreeMap<String,JacRow>> jacobians = new ArrayList<>();
      
        int N_ATTEMPTS = 20;
        
        for (int i=0; i<N_ATTEMPTS; ++i) {
            
            TreeMap<String,Double> x = newPoint(varNames, randomPointGen.getNextGuess());
            
            jacobians.add(evaluateSparseJacobian(dagCSP, x));
        }
        
        return jacobians;
    }
    
    /** This function should be deleted? */
    static private double getRowScale(ArrayList<JacRow> rowSamples) {
        
        assert rowSamples.size() > 1;

        Collections.sort(rowSamples, new Comparator<JacRow>() {
            @Override
            public int compare(JacRow a, JacRow b) {
                return Double.compare(Math.abs(a.residual), Math.abs(b.residual));
            }
        });
        
        int n = (rowSamples.size()+1)/2;
        
        double maxAij = 0.0;
        
        for (int i=0; i<n; ++i) {
            
            JacRow sample = rowSamples.get(i);
            
            double[] grad = Util.values( sample.gradients );
            
            double abs_rhs = Math.abs( sample.rhs);
            
            double aij = Math.max(Util.infinityNorm(grad), abs_rhs);
            
            //double res = Math.abs(sample.residual);
            
            //double max = Math.max(res, aij);
            double max = aij;
            
            if (maxAij < max) {
                
                maxAij = max;
            }
        }

        return enforceRange( 1/(maxAij+Double.MIN_VALUE), MIN_ROW_SCALE, MAX_ROW_SCALE );
    }

    static private TreeMap<String,Double> computeRowScales(ArrayList<TreeMap<String,JacRow>> jacobians) {
        
        TreeMap<String,Double> rowScales = new TreeMap<>();
        
        for (String constraintName : jacobians.get(0).keySet()) {
            
            ArrayList<JacRow> samplesOfConstraint = new ArrayList<>();
          
            for (int i=0; i<jacobians.size(); ++i) {
                
                samplesOfConstraint.add(jacobians.get(i).get(constraintName));
            }
            
            rowScales.put(constraintName, getRowScale(samplesOfConstraint));
        }
      
        return rowScales;
    }
    
    /** Assumes that Dag is a CSP */
    static private TreeMap<String,Double> computeRowScales(Dag dagCSP, TreeSet<String> varNames) {
     
        return computeRowScales( generateRandomJacobians(dagCSP, varNames) );
    }
 
    static private double scaledBoundViolationSquared(Dag dagCSP, double[] xVec, double[] xScale) {
        
        RealIntervalVector bnds = dagCSP.getVarBounds();
        
        double[] xLb = bnds.getLower().getData();
        
        double[] xUb = bnds.getUpper().getData();
        
        return Util.scaledBoundViolationSquared(xVec, xLb, xUb, xScale);
    }
    
    static private double scaledBoundViolationInfinityNorm(Dag dagCSP, double[] xVec, double[] xScale) {
        
        RealIntervalVector bnds = dagCSP.getVarBounds();
        
        double[] xLb = bnds.getLower().getData();
        
        double[] xUb = bnds.getUpper().getData();
        
        return Util.scaledBoundViolationInfinityNorm(xVec, xLb, xUb, xScale);
    }    
    
    private static void checkScaleConsistency(Dag dagCSP,
                                              TreeMap<String,Double> x,
                                              TreeMap<String,Double> rowScales, 
                                              TreeMap<String,Double> colScales) 
    {
        assert x.keySet().equals(colScales.keySet()) : "x = "+x.keySet()+", scales = "+colScales.keySet();
        
        ArrayList<Constraint> cons = dagCSP.getConstraints();
        
        assert cons.size()==rowScales.size();
        
        for (Constraint c : cons) 
            assert rowScales.get(c.getLabel())!=null : "Constraint \""+c.getLabel()+"\" not found";        
    }

    static Violations allConstraintAndVarboundViolations(Dag dagCSP,
                                                         TreeMap<String,Double> x,
                                                         TreeMap<String,Double> rowScales, 
                                                         TreeMap<String,Double> colScales ) 
    {
        checkScaleConsistency(dagCSP, x, rowScales, colScales);        
        
        IndexList indices = dagCSP.getVarIndices(x.keySet());
        
        double[] xVec = toInternalPermutationAsArray(indices, Util.values(x));
        
        DagIntervalEvalResult evalRes = dagCSP.evalRan(new RealIntervalVector(xVec), 
                                               dagCSP.computeParamBounds(new RealIntervalVector(xVec), ParamNode.PARAM_EVAL_TYPE_LOCSOL));
        
        double consInfNorm   = infinityNormOfConstraintViolation(evalRes, rowScales);
        
        double consL2Squared = sumOfSquaredConstraintViolation(evalRes, rowScales);
        
        double[] xScale = toInternalPermutationAsArray(indices, Util.values(colScales));
        
        double bndViol2   = scaledBoundViolationSquared(dagCSP, xVec, xScale);
        
        double bndViolInf = scaledBoundViolationInfinityNorm(dagCSP, xVec, xScale); 

        return new Violations( consL2Squared+bndViol2, Math.max(consInfNorm, bndViolInf), bndViolInf );
    }
    
    static Scales computeScales(Dag dagCSP, TreeSet<String> varNames) {
        
        TreeMap<String,Double> colScales = computeColScales(dagCSP, varNames);
        
        TreeMap<String,Double> rowScales = computeRowScalesBasedOnTypicalResidual( dagCSP );
        
        return new Scales(rowScales, colScales);       
    }
    
    static private TreeMap<String,Double> computeRowScalesBasedOnTypicalResidual(Dag dagCSP) {
        
        int NUM_OF_SAMPLES = 500; // magic number 
        
        TreeMap<String,ArrayList<Double>> residuals = collectRandomInfinityNormResiduals(dagCSP, NUM_OF_SAMPLES);

        TreeMap<String,Double> scales = extractApproxMedian(residuals);

        for (Map.Entry<String,Double> e : scales.entrySet()) {
            
            Double d = e.getValue();
            
            scales.put(e.getKey(), enforceRange( 1/(d+Double.MIN_VALUE), MIN_ROW_SCALE, MAX_ROW_SCALE ));
        }
        
        return scales;
    }
    
    static private TreeMap<String,ArrayList<Double>> collectRandomInfinityNormResiduals(Dag dagCSP, int NUM_OF_SAMPLES) {

        RealIntervalVector bounds = dagCSP.getVarBounds();
        
        double[] lb = bounds.getLower().getData();
        
        double[] ub = bounds.getUpper().getData();
        
        Random rnd = new Random(31L);
        
        ArrayList<double[]> points = LatinHypercubeSampling.generateLHSamples(rnd, NUM_OF_SAMPLES, lb, ub);
        
        TreeMap<String, ArrayList<Double>> residuals = new TreeMap<>();
        
        for (double[] x : points) {

            DagIntervalEvalResult evalRes = dagCSP.evalRan(new RealIntervalVector(x), 
                                                   dagCSP.computeParamBounds(new RealIntervalVector(x), ParamNode.PARAM_EVAL_TYPE_LOCSOL));

            appendResiduals(evalRes, residuals, NUM_OF_SAMPLES);
        }
        
        return residuals;
    }
    
    static private void appendResiduals(DagIntervalEvalResult evalRes, TreeMap<String,ArrayList<Double>> residuals, int NUM_OF_SAMPLES) {
        
        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            String constraintName = cei.getConstraint().getLabel();
            
            ArrayList<Double> res = residuals.get(constraintName);
            
            if (res==null) { 
                
                res = new ArrayList<>(); 
                
                res.ensureCapacity(NUM_OF_SAMPLES); 
            }
            
            res.add(Math.abs(getResidual(cei)));

            residuals.put(constraintName, res);
        }
    }
    
    static private TreeMap<String,Double> extractApproxMedian(TreeMap<String,ArrayList<Double>> residuals) {
        
        TreeMap<String,Double> approxMedian = new TreeMap<>();
        
        for (Map.Entry<String,ArrayList<Double>> e : residuals.entrySet()) {
            
            double[] res = Util.toArray( e.getValue() );
            
            Arrays.sort(res);
            
            approxMedian.put(e.getKey(), res[(res.length+1)/2]);            
        }
        
        return approxMedian;
    }
    
    static private final class NewtonIterate {
        final TreeMap<String,Double> x;
        final double violInfNorm;
        private NewtonIterate(TreeMap<String,Double> x, double violInfNorm) {
            this.x=x; this.violInfNorm=violInfNorm;
        }
    }
    
    static private NewtonIterate newtonStep(Dag dagCSP, double[] lb, double[] ub, TreeMap<String,Double> sol) {
        
        TreeMap<String,JacRow> jac = evaluateSparseJacobian(dagCSP, sol);
        
        double[] rowScales = new double[sol.size()];
        
        Arrays.fill(rowScales, 1.0);
        
        assert rowScales.length == dagCSP.getConstraintNumber();
        
        double[] colScales = rowScales;
        
        DenseMatrix J = toDenseMatrix(new ArrayList<String>(sol.keySet()), jac, rowScales, colScales);
        
        DenseVector res = extractResiduals(jac);
        
        double violInfNorm = Util.infinityNorm(res.getData());
        
        DenseLU lu = DenseLU.factorize(J);
        
        double rcond = Math.min(lu.rcond(J, Norm.One), lu.rcond(J, Norm.Infinity));
        
        if (rcond < 1.0e-16) {
            log.warn("Jacobian is numerically singular to machine precision");
            return null;
        }
        
        //System.out.printf("Condition number: %.2g\n", 1/rcond);
        
        double[] y = extractFirstColumn( lu.solve(new DenseMatrix(res)) );
        
        double[] x_old = Util.toArray(sol.values());
        
        double[] x_new = Util.add(y, x_old);
        
        //x_new = Util.projectToBox(x_new, lb, ub);
        
        //double alpha = violInfNorm > 0.05 ? 0.5 : 1.0;
        double alpha = 1.0; // full Newton step
        
        double[] x_damped = Util.add( Util.mult(alpha, x_new), Util.mult(1-alpha, x_old));  
        
        TreeMap<String,Double> s_new = Util.createMap(sol.keySet(), x_damped);
        
        double newInfNorm = Util.infinityNorm( Util.values( evaluateResidual(dagCSP, s_new) ) );
        
        //System.out.printf("Violation %.2g --> %.2g\n", violInfNorm, newInfNorm);
        
        return new NewtonIterate(s_new, newInfNorm);
    }
    
    static private Solution denseNewtonIterate(Dag dagCSP, double[] lb, double[] ub, Solution s) {
        
        int iter = 0;
        
        TreeMap<String,Double> sol = s.getAllVars();
        
        do {
            
            NewtonIterate newSol;
            
            try {
            
                newSol = newtonStep(dagCSP, lb, ub, sol);
            }
            catch (Throwable t) {
                log.warn("Function / Jacobian evaluation failed in Newton step");
                return Factory.createInvalidSolution();
            }
            
            if (newSol==null)
                return Factory.createInvalidSolution();
            
            if (newSol.violInfNorm < 1.0e-6) {
                //System.out.println("Convergence reached!");
                return new SolutionImpl(newSol.x, newSol.violInfNorm);
            }
            
            sol = newSol.x;
            
            ++iter;
            
        } while (iter<25); 
        
        log.warn("Iteration limit reached");
        
        return Factory.createInvalidSolution();
    }
    
    static ArrayList<Solution> applyDenseNewton(Dag dagCSP, 
                                                double[] lb, 
                                                double[] ub, 
                                                ArrayList<Solution> solutions) 
    {    
        //sortByInfeasibilityInfNorm(solutions);
        
        ArrayList<Solution> preciseSolutions = new ArrayList<>(solutions.size());
        
        for (Solution s : solutions) {
            
            Solution improvedSol = denseNewtonIterate(dagCSP, lb, ub, s);

            preciseSolutions.add(improvedSol);
            
            //System.out.println("");
        }
        
        return preciseSolutions;
    }
    
    static void showConditionNumbers(Dag dagCSP, ArrayList<Solution> solutions) {
        
        int i=0;
        
        for (Solution s : solutions) {
            
            showConditionNumber(dagCSP, s, ++i);
        }
    }
    
    private static void showConditionNumber(Dag dagCSP, Solution s, int i) {
        
        TreeMap<String,Double> sol = s.getAllVars();
        
        TreeMap<String,JacRow> jac = evaluateSparseJacobian(dagCSP, sol);
        
        double[] rowScales = new double[sol.size()];
        
        Arrays.fill(rowScales, 1.0);
        
        assert rowScales.length == dagCSP.getConstraintNumber();
        
        double[] colScales = rowScales;
        
        DenseMatrix J = toDenseMatrix(new ArrayList<String>(sol.keySet()), jac, rowScales, colScales);       
       
        DenseLU lu = DenseLU.factorize(J);
        
        double rcond = Math.min(lu.rcond(J, Norm.One), lu.rcond(J, Norm.Infinity));
        
        String what = (rcond < 1.0e-12) ? "is ill-conditioned, potentially singular" : "seems to be OK";
        
        System.out.printf("Solution %d, the Jacobian at the solution %s, (1/cond=%.1e)\n", 
                          i, what, rcond);
    }
    
/*    
    private static void sortByInfeasibilityInfNorm(ArrayList<Solution> solutions) {
        
        Collections.sort(solutions, new Comparator<Solution>() {
            @Override
            public int compare(Solution o1, Solution o2) {
                return Double.compare(o1.getInfeasibilityInfinityNorm(), o2.getInfeasibilityInfinityNorm());
            }
        });        
    }
*/    
    private static DenseVector extractResiduals(TreeMap<String,JacRow> jac) {
        
        DenseVector residuals = new DenseVector(jac.size());
        
        int k=0;
        
        for (Map.Entry<String, JacRow> e : jac.entrySet()) {
            
            residuals.set(k++, -e.getValue().residual);
        }
        
        return residuals;
    }
    
    private static double[] extractFirstColumn(DenseMatrix m) {
        
        double[] res = new double[m.numRows()];
        
        for (int i=0; i<res.length; ++i) {
            
            res[i] = m.get(i, 0);
        }
        
        return res;
    }
    
    public static ArrayList<TreeMap<String, Double>> findLeastInfluentialWithQRP(Dag CSPDag, 
                                                                                 TreeMap<String, Double> rowScales,
                                                                                 TreeMap<String, Double> columnScales,
                                                                                 TreeMap<String, Double> xPart, 
                                                                                 ArrayList<String> yVarsSorted, 
                                                                                 ArrayList<double[]> samples,
                                                                                 TreeSet<String> yLast,
                                                                                 int nToFix)
    {
        ArrayList<TreeMap<String,Double>> points = spliceParts(xPart, yVarsSorted, samples);
/*
        //-----
        // HACK
        
        ArrayList<TreeMap<String,Double>> fixed = new ArrayList<>();
        
        String x1 = new ArrayList<>(yLast).get(1);
        String x3 = new ArrayList<>(yLast).get(3);
        
        for (TreeMap<String,Double> p : points) {

            TreeMap<String,Double> tmp = new TreeMap<>();
         
            if (p.get(x3)>0.5)
                tmp.put(x3, p.get(x3));
            else 
                tmp.put(x1, p.get(x1));
            
            fixed.add(tmp);
        }       
        
        return fixed;
        //-----
*/        
        ArrayList<TreeMap<String,JacRow>> jacobians = evalJacobiansAtTreeMapPoints(CSPDag, points);
        
        //-------------------------------------------------------------------------------
        
        checkScaleConsistency(CSPDag, points.get(0), rowScales, columnScales);
        
        double[] arrRowScales = Util.values(rowScales);
        double[] arrColScales = Util.values(columnScales);
        
        ArrayList<String> allVarNames = new ArrayList<>(points.get(0).navigableKeySet()); 
        
        assert arrRowScales.length == jacobians.get(0).size();
        assert arrColScales.length == allVarNames.size();
        
        //------------------------------------------------------------------------------
        
        ArrayList<TreeMap<String,Double>> fixed = new ArrayList<>();

        for (TreeMap<String,JacRow> jac : jacobians) {
            
            DenseMatrix A = toDenseMatrix(allVarNames, jac, arrRowScales, arrColScales);
            
            int[] perm = columnPermutationWithQRP(A);
            
            TreeMap<String,Double> x = jac.firstEntry().getValue().x;
            
            fixed.add( fixTrailing(allVarNames, perm, nToFix, x, yLast) );
        }
       
        return fixed;
    }
    
    static private ArrayList<TreeMap<String,Double>> spliceParts(TreeMap<String, Double> xPart, 
                                                                 ArrayList<String> yVarsSorted, 
                                                                 ArrayList<double[]> samples    ) 
   {    
        ArrayList<TreeMap<String,Double>> points = new ArrayList<>(samples.size());
        
        for (double[] ySample : samples) {
            
            TreeMap<String,Double> point = new TreeMap<>(xPart);
            
            point.putAll( Util.createMap(yVarsSorted, ySample) );
            
            points.add(point);
        }
        
        return points;
    }
    
    static private ArrayList<TreeMap<String,JacRow>> evalJacobiansAtTreeMapPoints(Dag CSPDag, ArrayList<TreeMap<String,Double>> points) {
        
        ArrayList<TreeMap<String,JacRow>> jacobians = new ArrayList<>(points.size());
        
        for (TreeMap<String,Double> point : points) {
         
            try {
            
                jacobians.add( evaluateSparseJacobian(CSPDag, point) );
            }
            catch (Throwable t) {
                t.printStackTrace();
                log.error("Most likely the Jacobian evaluation has failed at this point");
                log.debug(points);
            }
        }
        
        return jacobians;
    }
    
    static private TreeMap<String,Double> fixTrailing(ArrayList<String> allVarNames, 
                                                      int[] perm, 
                                                      int nToFix, 
                                                      TreeMap<String,Double> x, 
                                                      TreeSet<String> yLast) 
    {  
        TreeMap<String,Double> yFixed = new TreeMap<>();

        for (int k = perm.length-1; k>=0 && yFixed.size() < nToFix; --k) {
            
            String name = allVarNames.get(perm[k]);
            
            if (yLast.contains(name)) {
                
                Double value = x.get(name);

                assert value!=null;

                Double old = yFixed.put(name, value);
                
                assert old == null;
            }
        }
        
        assert yFixed.size()==nToFix;
        
        return yFixed;
    }
    
    static ArrayList<TreeMap<String,Double>> sampleUnderDeterminedWithQR(Dag csp, 
                                                                         ArrayList<String> varsSorted, 
                                                                         TreeMap<String,Double> rowScales, 
                                                                         TreeMap<String,Double> colScales, 
                                                                         Random rnd, 
                                                                         int m, 
                                                                         int toFix) 
    {
        assert toFix > 0;
        assert toFix < varsSorted.size();
        
        ArrayList<CSPSample> samples = sample(csp, rowScales, rnd, 6*m); // magic number

        Ordering<CSPSample> ord = Ordering.from(new CSPSample.ByViolation());

        samples = new ArrayList<CSPSample>( ord.leastOf(samples, 2*m) );  // magic number

        ArrayList<Solution> fakeSolutions = toFakeSolutions(csp.getVarNamesArray(), samples);
        
        //Collections.shuffle(fakeSolutions, rnd);
        
        ArrayList<Solution> selected = selectMostPromising(Util.values(colScales), m, fakeSolutions);
        
        ArrayList<TreeMap<String,JacRow>> jacobians = evalJacobians(csp, selected);
        
        return findLeastInfluentialWithQRP(jacobians, varsSorted, Util.values(rowScales), Util.extractValues(colScales, varsSorted), toFix);       
    }
   
    static private ArrayList<CSPSample> sample(Dag csp, TreeMap<String,Double> rowScales, Random rnd, int m) {

        RealIntervalVector bounds = csp.getVarBounds();

        double[] lb = bounds.getLower().getData();

        double[] ub = bounds.getUpper().getData();

        ArrayList<double[]> points = LatinHypercubeSampling.generateLHSamples(rnd, m, lb, ub);

        ArrayList<CSPSample> samples = new ArrayList<>();

        for (double[] xArray : points) {

            double sumViolSquared = sumOfSquaredConstraintViolation(csp, xArray, rowScales);

            if (!Double.isNaN(sumViolSquared)) {

                samples.add(new CSPSample(xArray, sumViolSquared));
            }
        }        

        return samples;
    }

    static private ArrayList<Solution> toFakeSolutions(String[] namesInDagOrder, ArrayList<CSPSample> samples) {
        
        Solution empty = SolutionImpl.createEmptySolution();
        
        ArrayList<Solution> fakeSolutions = new ArrayList<>(); 
        
        for (int i=0, n=samples.size(); i<n; ++i) {
            
            CSPSample s = samples.get(i);
            
            fakeSolutions.add(empty.extend(Util.clip(namesInDagOrder, s.x), s.sumViolSquared)); // permutes the vars from DAG's internal to alphabetic
        }
        
        return fakeSolutions;
    }
    
    static private ArrayList<Solution> selectMostPromising(double[] colScales, int keep, ArrayList<Solution> fakeSolutions) {
        
        CandidateSelector cs = CandidateSelectorImpl.createSubproblemEvaluator(colScales);
        
        cs.selectMostPromisingCandidates(fakeSolutions, keep, 2.0); // magic number infeasWeight
        
        return cs.getSelectedOnes();
    }
    
    static private ArrayList<TreeMap<String,JacRow>> evalJacobians(Dag csp, ArrayList<Solution> selected) {

        ArrayList<TreeMap<String,JacRow>> jacobians = new ArrayList<>();
        
        for (Solution s : selected) {
         
            jacobians.add( evaluateSparseJacobian(csp, s.getAllVars()) );            
        }
        
        return jacobians;
    }

    static private ArrayList<TreeMap<String,Double>> findLeastInfluentialWithQRP(ArrayList<TreeMap<String,JacRow>> jacobians, 
                                                                                 ArrayList<String> varsSorted, 
                                                                                 double[] rowScales,
                                                                                 double[] colScales, 
                                                                                 int toFix)
    {
        return findInfluentialWithQRP(jacobians, varsSorted, rowScales, colScales, toFix, new TrailingFixer()); 
    }
    
    // Package private to make the compiler shut up; unused otherwise
    static ArrayList<TreeMap<String,Double>> findMostInfluentialWithQRP(ArrayList<TreeMap<String,JacRow>> jacobians, 
                                                                                ArrayList<String> varsSorted, 
                                                                                double[] rowScales,
                                                                                double[] colScales, 
                                                                                int toFix)
    {
        return findInfluentialWithQRP(jacobians, varsSorted, rowScales, colScales, toFix, new LeadingFixer()); 
    }      
    
    /** The Fixer decides whether the most or the least influential variables (leading or trailing) variables are fixed */
    static private ArrayList<TreeMap<String,Double>> findInfluentialWithQRP(ArrayList<TreeMap<String,JacRow>> jacobians, 
                                                                            ArrayList<String> varsSorted, 
                                                                            double[] rowScales,
                                                                            double[] colScales, 
                                                                            int toFix,
                                                                            Fixer fixer) 
    {
        assert rowScales.length == jacobians.get(0).size();
        assert colScales.length == varsSorted.size();
        
        ArrayList<TreeMap<String,Double>> fixed = new ArrayList<>();

        for (TreeMap<String,JacRow> jac : jacobians) {
            
            DenseMatrix A = toDenseMatrix(varsSorted, jac, rowScales, colScales);
            
            int[] perm = columnPermutationWithQRP(A);
            
            TreeMap<String,Double> x = jac.firstEntry().getValue().x;
            
            fixed.add( fixer.fix(varsSorted, toFix, perm, x) );
        }

        return fixed;
    }

    static private DenseMatrix toDenseMatrix(ArrayList<String> varsSorted, TreeMap<String,JacRow> jac, double[] rowScales, double[] colScales) {
        
        DenseMatrix denseJac = new DenseMatrix(jac.size(), varsSorted.size());
        
        int rowIndex = 0;
        
        for (Map.Entry<String,JacRow> e : jac.entrySet()) {
            
            for (Map.Entry<String,Double> jacElem : e.getValue().gradients.entrySet()) {
                
                String var = jacElem.getKey();
                
                int colIndex = Collections.binarySearch(varsSorted, var); 
                
                if (colIndex >= 0) {
                    
                    denseJac.set(rowIndex, colIndex, jacElem.getValue()*rowScales[rowIndex]*colScales[colIndex]);
                }
            }
            
            ++rowIndex;
        }

        return denseJac;
    }
    
    static private int[] columnPermutationWithQRP(DenseMatrix A) {
        
        QRP qrp = QRP.factorize(A);
        
//        if (qrp.getRank() != A.numRows()) {
//            
//            log.warn("Jacobian is singular "+A);
//        }
        
        return qrp.getPVector();
    }
   
    interface Fixer {
        
        TreeMap<String,Double> fix(ArrayList<String> varsSorted, int toFix, int[] perm, TreeMap<String,Double> x);
    }

    private static final class TrailingFixer implements Fixer {

        @Override
        public TreeMap<String,Double> fix(ArrayList<String> varsSorted, int toFix, int[] perm, TreeMap<String,Double> x) {
            
            return Evaluator.fix(varsSorted, perm, perm.length-toFix, perm.length, x);
        }
    }

    private static final class LeadingFixer implements Fixer {

        @Override
        public TreeMap<String,Double> fix(ArrayList<String> varsSorted, int toFix, int[] perm, TreeMap<String,Double> x) {  
          
            return Evaluator.fix(varsSorted, perm, 0, toFix, x);
        }
    }
    
    private static TreeMap<String,Double> fix(ArrayList<String> varsSorted, int[] perm, int begin, int end, TreeMap<String,Double> x) {
        
        assert begin >= 0;
        assert end <= perm.length;        
        assert begin < end;
        
        TreeMap<String,Double> fixed = new TreeMap<>();

        for (int i=begin; i<end; ++i) {

            String var = varsSorted.get(perm[i]);

            Double value = x.get(var);

            assert value!=null;

            fixed.put(var, value);
        }

        return fixed;
    }
    
    //---------------------------------------------------------------------------------------
    
    static private double enforceRange(double xVal, double minVal, double maxVal) {
        
        assert minVal < maxVal;
        
        double x = Math.max(xVal,  minVal);
     
        return Math.min(x, maxVal);
    }
    
    static private TreeMap<String,Double> computeColScales(Dag dag, TreeSet<String> varNames) {

        TreeMap<String,Double> scales = new TreeMap<>();
        
        VarBoundsInAlphabeticalOrder bnds = getVarBounds(dag, varNames);
        
        int k = 0;
        
        for (String var : varNames) {
            
            double scl = 1.0/(bnds.ub[k]-bnds.lb[k]+Double.MIN_VALUE);
            
            scales.put(var, enforceRange(scl, MIN_COL_SCALE, MAX_COL_SCALE));
            
            ++k;
        }
        
        return scales;
    }
    
    static {

        ClassLoader.getSystemClassLoader().setDefaultAssertionStatus(true);
    }
    
    static private void SimpleEvaluatorTest() throws Throwable {
        
        String file = "data/EvaluatorTest/EvaluatorTest.mod";
        
        Sdms.setProblemCheck(false, false, false, 1e-6);
   
        Dag dag = Sdms.readProblem(file, true).getModel();
        
        String[] names  = { "x1", "x2", "x3", "x4" };
        Double[] values = {  2.0,  2.0,  2.0,  2.0 };
        
        TreeMap<String,Double> xVarnameValueMap = new TreeMap<>();          
        
        for (int i=0; i<names.length; ++i) {
            
            xVarnameValueMap.put(names[i], values[i]);
        }
        
        TreeSet<String> varNames = new TreeSet<String>(Arrays.asList(names));
        
        TreeMap<String,Double> res = Evaluator.evaluateResidual(dag, xVarnameValueMap);
        
        System.out.println("Residuals "+res);        
        
        TreeMap<String,JacRow> jac = Evaluator.evaluateSparseJacobian(dag, xVarnameValueMap);
        
        System.out.println("Jacobian");
        
        for (Map.Entry<String, JacRow> row: jac.entrySet()) {
            
            System.out.println(row.getKey()+"  "+row.getValue());
        }
        
        TreeMap<String,Double> colScales = computeColScales(dag, varNames);
        
        System.out.println("col scales");
        
        for (Map.Entry<String,Double> scl : colScales.entrySet()) {
            
            System.out.println(scl.getKey() + "  " + scl.getValue());
        }
        
        TreeMap<String,Double> rowScales = computeRowScales(dag, varNames);
        
        System.out.println("row scales");
        
        for (Map.Entry<String,Double> scl : rowScales.entrySet()) {
            
            System.out.println(scl.getKey() + "  " + scl.getValue());
        }         
    }
    
    static private void QRFactorizationWithPivoting() {
        
        DenseMatrix A = new DenseMatrix( new double[][] {{-1, 0, 2 }, { 0, 3, 0} });
        
        QRP qrp = QRP.factorize(A);

        System.out.println("column permutation: "+Util.toString(qrp.getPVector()));
        
        System.out.println("rank = "+qrp.getRank());

        return;
    }
    
    public static void mainQRP() {
        
        QRFactorizationWithPivoting();
    }
    
    public static void mainTest(String[] args) throws Throwable {
        
        SimpleEvaluatorTest();
        
        System.out.println("args[] = "+Arrays.asList(args));
        
        System.out.println("------------------------------------------------------");
        
        //String file = "data/JacobsenTornNatruralOrdering/JacobsenTorn.mod";
        String file = "data/mss20TornBlk/mss20Torn.mod";
        
        System.out.println("Scales for problem "+file);
        
        Sdms.setProblemCheck(false, false, false, 1e-6);
   
        Dag dag = Sdms.readProblem(file, true).getModel();
       
        TreeSet<String> varNames = new TreeSet<String>(dag.getVarNames());
       
        TreeMap<String,Double> colScales = computeColScales(dag, varNames);
        
        System.out.println("col scales");
        
        for (Map.Entry<String,Double> scl : colScales.entrySet()) {
            
            System.out.println(scl.getKey() + "  " + scl.getValue());
        }
        
        //TreeMap<String,Double> rowScales = computeRowScales(dag, varNames);
        TreeMap<String,Double> rowScales = computeRowScalesBasedOnTypicalResidual(dag);
        
        System.out.println("row scales");
        
        for (Map.Entry<String,Double> scl : rowScales.entrySet()) {
            
            System.out.println(scl.getKey() + "  " + scl.getValue());
        }   
        
        return;
    }
    
    private static void printDenseMatrix(DenseMatrix A) {
        
        for (int i=0, n=A.numRows(); i<n; ++i) {
            
            System.out.printf("[");
            
            for (int j=0, m=A.numColumns(); j<m; ++j) {
                
                if (j>0) System.out.printf(",");
                
                System.out.printf("%6.3f", A.get(i,j));
                
            }
            
            System.out.printf("]\n");                
        }
    }
    
    private static void svdTest() {

        //DenseMatrix A = new DenseMatrix( new double[][] {{-1, 0, 2 }, { 0, 3, 0} });
        //DenseMatrix A = new DenseMatrix( new double[][] {{0,2}, {0,1} }); // data row-wise
        //DenseMatrix A = new DenseMatrix( new double[][] {{0,2,4}, {0,1,2}}); // data row-wise
        //DenseMatrix A = new DenseMatrix( new double[][] {{0,5}, {0,2}, {0,1}}); // data row-wise
        DenseMatrix A = new DenseMatrix( new double[][] {{0,5,1.25}, {0,2,1}, {0,1,0.5}}); // data row-wise
        
        System.out.println("A = ");
        printDenseMatrix(A);

        try {

            SVD svd = SVD.factorize(A);
            
            DenseMatrix U = svd.getU();

            System.out.println("U = ");
            printDenseMatrix(U);

            double[] S = svd.getS();
            
            System.out.println("\nsigmas = "+Util.arrayToString(S, "%6.3f"));
        }
        catch (Exception pokemon) {

        }
        
        return;
    }
    
    public static void mainSVDTest(@SuppressWarnings("unused") String args[]) {

        svdTest();   
    }
}

final class CSPSample {

    double[] x;
    double   sumViolSquared;

    CSPSample(double[] x, double sumViolSquared) { 
        assert sumViolSquared >= 0;
        this.x=x; 
        this.sumViolSquared=sumViolSquared; 
    }

    static class ByViolation implements Comparator<CSPSample> {
        @Override
        public int compare(CSPSample a, CSPSample b) {
            return Double.compare(a.sumViolSquared, b.sumViolSquared);
        }
    }
}

