//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.coinor.Ipopt;

import problem.Problem;
import sdms.*;
import lmbopt.model.Lmbopt;
import tools.MyUtils;
import tools.console.DefaultConsole;
import tools.math.function.FunctionIpoptSolver;
import tools.math.function.FunctionSolver;
import tools.math.function.FunctionSolverOptions;
import tools.opt.OptException;
import tools.opt.OptIpoptSolver;
import tools.options.IllegalValueException;

public final class Factory {

    private Factory() { }
    
    private static final Logger log = LogManager.getLogger();

    public static NLPProblem createNLPProblem(String file, String scale) {

        try {

            Sdms.setProblemCheck(false, false, true, 1e-6);
            
            Problem p = Sdms.readProblem(file, false);
            
            Scales s = computeScales(file, p, scale);

            return new NLPProblemImpl(p, s);            
        } 
        catch (IOException e) {

            log.fatal(e.getMessage());

            throw new ProgramTerminationRequested();            
        } 
        catch (Exception e) {

            e.printStackTrace();

            log.fatal("failed to read the problem");

            throw new ProgramTerminationRequested();
        }
    }
    
    private static Scales computeScales(String file, Problem p, String scale) {
        
        int posExt = file.lastIndexOf(".mod");

        String baseName = (posExt != -1) ? file.substring(0, posExt) : file;
        
        ArrayList<String> permuted_col_names = Util.stringVectorFromFile(baseName + ".pcol");
        
        ArrayList<String> pcol = Util.filterNameVector(permuted_col_names);        
        
        TreeSet<String> varNames = new TreeSet<>(pcol);
        
        switch (scale) {
        
        case "disabled":
            
            return createUnitScales(varNames, baseName);
        
        case "enabled":
            
            return Evaluator.computeScales(p.getModel(), varNames);
  
        default:
            
            log.error("Unrecognized scaling option: \""+scale+"\"");
            
            throw new ProgramTerminationRequested();
        }
    }
    
    private static Scales createUnitScales(TreeSet<String> varNames, String baseName) {
        
        ArrayList<String> permuted_row_names = Util.stringVectorFromFile(baseName + ".prow");
        
        TreeSet<String> rowNames = new TreeSet<>(Util.filterNameVector(permuted_row_names));
        
        TreeMap<String,Double> rowScales = new TreeMap<>();
        
        for (String row : rowNames) {
            
            rowScales.put(row, 1.0);
        }
        
        TreeMap<String,Double> colScales = new TreeMap<>();
        
        for (String col : varNames) {
            
            colScales.put(col, 1.0);
        }
        
        return new Scales(rowScales, colScales);
    }

    public static Solver createSolver(String solver) {

        if (solver.equals("ipopt")) {
            
            log.info("Running with IPOPT as a CSP solver");
            
            return createIpoptAsCspSolver();
        }
        else {

            return createBoundConstraintSolver(solver);
        }
    }
    
    private static Solver createIpoptAsCspSolver() {
        
        return new SolverCSP( OptIpoptSolver.getInstance(MyUtils.getWorkingDir() 
                + MyUtils.getFileSep() + "lib") ); // Apparently options cannot be passed
    }
    
    private static Solver createBoundConstraintSolver(String solver) {
        
        boolean lmbopt = solver.equals("lmbopt");
        
        log.info("Running with "+(lmbopt?"LMBOPT":"IPOPT")+" as a bound constraint solver");

        FunctionSolver sol = lmbopt ? createLMBOPT() : createIPOPT();
        
        try {
          FunctionSolverOptions opts = new  FunctionSolverOptions(0, 1e-4, 500);
          return new SolverBoundConstraint(sol, opts);     
        } catch (OptException e) {
          log.fatal(e.getMessage());
          throw new ProgramTerminationRequested();
        }   
    }

    private static Lmbopt createLMBOPT() {

        try {
            
            tools.options.Options options = Lmbopt.getGlobalOptions();
            
            options.setValue(Lmbopt.OPTION_WRITE_OUTPUT_FILE, false);
            options.setValue(Lmbopt.OPTION_SOLUTION_WRITE_TESTENV_BOOLEAN, false);
            options.setValue(Lmbopt.OPTION_MEASURE_TIME_BOOLEAN, false);
            
            Lmbopt.setGlobalOptions(options);
        } 
        catch (IllegalValueException e) {
            log.fatal(e.getMessage());
            throw new ProgramTerminationRequested();
        }
        
        Lmbopt lmbopt=Lmbopt.getInstance();
        lmbopt.setConsole(new DefaultConsole());
        return lmbopt;
    }

    private static FunctionIpoptSolver createIPOPT() {
        
        try {
            
            tools.options.Options options = FunctionIpoptSolver.getGlobalOptions();
            options.setValue(Ipopt.KEY_PRINT_LEVEL, 3);
            options.setValue(Ipopt.KEY_HESSIAN_APPROXIMATION, "limited-memory");
        } 
        catch (IllegalValueException e) {
            log.fatal(e.getMessage());
            throw new ProgramTerminationRequested();
        }

        return FunctionIpoptSolver.getInstance();
    }

    public static StartingpointGenerator createStartingpointGenerator() {

        return new StartingpointGeneratorImpl();
    }
    
    public static StartingpointGenerator createStartingpointGenerator(Random rnd, double lb[], double ub[]) {

        StartingpointGenerator spg = new StartingpointGeneratorImpl(rnd);
        
        spg.setVarBounds(lb, ub);
        
        return spg;
    }

    public static SubproblemSelector createSubproblemSelector(String file) {

        int posExt = file.lastIndexOf(".mod");

        String baseName = (posExt != -1) ? file.substring(0, posExt) : file;

        ArrayList<String> permuted_col_names = Util.stringVectorFromFile(baseName + ".pcol");
        ArrayList<String> permuted_row_names = Util.stringVectorFromFile(baseName + ".prow");
        ArrayList<String[]> col_of_subproblems = Util.sparseStringMatrixFromFile(baseName + ".csp");
        ArrayList<String[]> row_of_subproblems = Util.sparseStringMatrixFromFile(baseName + ".rsp");
        ArrayList<String[]> permuted_full_sparsity = Util.sparseStringMatrixFromFile(baseName + ".pfs");

        int nRow1 = permuted_full_sparsity.size();
        int nRow2 = permuted_row_names.size();

        assert nRow1 == nRow2 :
            "size mismatch in sparsity pattern, cons in sparsity pattern " + nRow1 + ", number of constraints " + nRow2;
        
        
        ArrayList<String>   pcol = Util.filterNameVector(permuted_col_names);
        ArrayList<String>   prow = Util.filterNameVector(permuted_row_names);
        ArrayList<String[]> csp  = Util.filterNameMatrix(col_of_subproblems);
        ArrayList<String[]> rsp  = Util.filterNameMatrix(row_of_subproblems);
        ArrayList<String[]> pfs  = Util.filterNameMatrix(permuted_full_sparsity);
            
        return new SubproblemSelectorImpl(pcol, csp, prow, rsp, pfs);
    }

    public static ArrayList<Solution> createEmptyPartialSolutions(int n) {

        ArrayList<Solution> emptySols = new ArrayList<>();

        for (int i = 0; i < n; ++i) {

            emptySols.add(SolutionImpl.createEmptySolution());
        }

        return emptySols;
    }
    
    public static Solution createEmptySolution() {
        
        return SolutionImpl.createEmptySolution();
    }
    
    public static Solution createInvalidSolution() {
        
        return new SolutionImpl();
    }
    
    public static CandidateSelector createSubproblemEvaluator(double[] colScales) {
        
        return CandidateSelectorImpl.createSubproblemEvaluator(colScales);
    }
    
    public static CandidateSelector createCumulativeEvaluator(double[] colScales) {
        
        return CandidateSelectorImpl.createCumulativeEvaluator(colScales);
    }
    
    public static Plotter createPlotter(String file, String option) {
        
        switch (option) {
        case "no":
            return new PlotterDoesNothing();
        case "slice":
            return new PlotterImpl(file);  
        case "triangle":
            return new PlotterTriangle();
        default:
            log.fatal("Unrecognized plotting option: \""+option+"\"");
            throw new ProgramTerminationRequested();        
        }
    }
}

