//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import dag.Dag;
import dag.DagFactory;
import problem.Problem;
import problem.solution.ReferenceSolution;
import sdms.Util;
import tools.MyUtils;
import tools.SanityCheckException;
import tools.console.DefaultConsole;
import tools.math.RealVector;
import tools.opt.TestEnvChkFile;
import tools.shell.ShellCommand;
import tools.shell.ShellCommandResult;
import tools.shell.ShellException;

// TODO Remove the ugly hack that FlatteningTest and 
// TestEnvConverter use the same code for the most part

final class FlatteningTest {
    
    private static Logger log;

    Problem p;
    boolean hasPermutation;
    ArrayList<String> allCons;
    ArrayList<String> allVars;
    ArrayList<String[]> permuted_full_sparsity;
    TreeMap<String,String[]> fullProblemSparsity;
    Dag subCSP;
    
    private static ArrayList<String> readTestSet() {
        
        ArrayList<String> unFilteredTestSet = Util.stringVectorFromFile("data/test_set.txt");
        
        ArrayList<String> testSet = new ArrayList<>();
        
        for (String s : unFilteredTestSet) {
            
            testSet.add(s.replace("/original", ""));
        }
        
        return testSet;
    }
    
    private void readFlattenedAmpl(String file) throws Throwable {

        Sdms.setProblemCheck(false, false, true, 1e-6);

        p = Sdms.readProblem(file, false);
    }
    
    // TODO Check both, permutation only if available
    private void readSparsityPattern(String file) {
        
        String baseName = file.substring(0, file.length()-4);
        
        hasPermutation = new File(baseName+".pfs").isFile(); 
        
        if (!hasPermutation) {
            log.debug("Permutation not found, using the original sparsity pattern");
        }
        
        String ROW_EXT = hasPermutation ? ".prow" : ".row";
        String COL_EXT = hasPermutation ? ".pcol" : ".col"; 
        String SPP_EXT = hasPermutation ? ".pfs"  : ".osp";
        
        allCons = Util.filterNameVector(Util.stringVectorFromFile(baseName+ROW_EXT));
        
        allVars = Util.filterNameVector(Util.stringVectorFromFile(baseName+COL_EXT));
        
        permuted_full_sparsity = Util.filterNameMatrix(Util.sparseStringMatrixFromFile(baseName+SPP_EXT));

        fullProblemSparsity = new TreeMap<>();
        
        for (int i=0; i<allCons.size(); ++i) {
            
            fullProblemSparsity.put(allCons.get(i), permuted_full_sparsity.get(i));
        }        
    }
    
    private void buildSubCSP() throws Throwable {

        boolean debug = false;

        // all constraints selected, nothing is fixed
        subCSP = DagFactory.generateSubCSP(p.getModel(), false,
                allCons,
                new TreeMap<String,Double>(), 
                new DefaultConsole(),
                debug);
    }
    
    private void checkSparsityPattern() {

        TreeSet<String> constraintDependencies = new TreeSet<>();

        for (String con : allCons) {

            String[] varsAmpl = fullProblemSparsity.get(con);

            log.trace("Checking "+con+" sparsity");

            TreeSet<String> varsInDag = subCSP.getConstraintDependingOn(subCSP.getConstraint(con));

            assert varsInDag.containsAll(Arrays.asList(varsAmpl)) : 
                "Constraint "+con+" depends on vars in dag: "+varsInDag+", in AMPL "+Arrays.toString(varsAmpl);

            assert varsInDag.size()==varsAmpl.length;

            constraintDependencies.addAll(varsInDag);           
        }

        assert constraintDependencies.containsAll(allVars);
        
        log.debug("Sparsity pattern seems to be OK!");
    }

    private void checkReferenceSolutions() {

        p.checkReferenceSolutions();
        for (ReferenceSolution rs : p.getReferenceSolutions()) {
            double viol = rs.getMaxConstrViol();
            if (log.isTraceEnabled()) {
                log.trace("Reference solution=" + rs);
                log.trace("Reference solution max violation=" + viol
                        + " at " + rs.getMaxViolatedConstraint());
                log.trace("Reference solution sum violation=" + rs.getSumConstrViol());
            }
            if (viol > 1.0e-6) {
                throw new SanityCheckException("Substituted problem constraint "
                        + "max violation is too big, " + viol + " at '"
                        + rs.getMaxViolatedConstraint().getLabel() + "'");
            }
        }
        log.debug("All "+p.getReferenceSolutions().size()+" reference solutions seem to be OK!");
    }

    private void check(String file) throws Throwable {

        readFlattenedAmpl(file);
        
//        new TestEnvConverter(p, file).run(); // HACK Just to generate the files for the test environment

        readSparsityPattern(file);
        
        buildSubCSP();

        checkSparsityPattern();
        
        checkReferenceSolutions();
    }    
    
    private static void printStatistics(int allTests, int passed, ArrayList<String> failed) {
        
        log.info(allTests+" tests, "+passed+" passed, "+failed.size()+" failed");
        
        if (!failed.isEmpty()) {
            
            log.info("Failed tests:");
            
            for (String file : failed) {
                
                log.info(file);
            }
            
            log.error("TEST FAILED!");
        }
        else {
            
            log.info("PASSED!");
        }        
    }

    private void run() {        

        ArrayList<String> allFiles = readTestSet();
        
        int passed = 0;
        
        ArrayList<String> failed = new ArrayList<>();
        
        for (String file : allFiles) {
            
            try {
                
                log.debug("Checking: "+file);
                
                check(file);
                
                ++passed;
                
                log.debug("Passed: "+file);
            }
            catch (Throwable t) {
                
                log.error("Failed: "+file);
                                                
                log.catching(Level.ERROR, t);
                
                failed.add(file);
            }            
        }
        
        printStatistics(allFiles.size(), passed, failed);
    }
    
    static {
        
        String propertyName = "log4j.configurationFile";
        
        if (System.getProperty(propertyName)==null) {
            
            System.setProperty(propertyName, "config/log4j2.xml");
        }
        
        log = LogManager.getLogger();
        
        ClassLoader.getSystemClassLoader().setDefaultAssertionStatus(true);
        
    }

    public static void main(String[] args) {
        
        Util.testAssertationsEnabled();
        
        new FlatteningTest().run();
    }
}

final class TestEnvConverter {

     String PERL_WIN="perl", 
            AMPL_WIN="D:\\Projects\\TestEnvironment\\external\\win\\amplcml\\ampl.exe",
            AMPL2DAG_WIN="D:\\Projects\\TestEnvironment\\external\\win\\ampl2dag.pl",
            DAGSTRUCT_WIN="",
            //DAGSTRUCT_WIN="D:\\Projects\\TestEnvironment\\external\\win\\dag_struct.exe",
            COCOSUBSET_WIN="D:\\Projects\\TestEnvironment\\external\\win",
            SOLCHECK_WIN="D:\\Projects\\TestEnvironment\\external\\win\\solcheck.exe";
     
     String PERL_LINUX="perl", 
            AMPL_LINUX="/home/ali/ampl_old/ampl",
            COCOSUBSET_LINUX= "/home/ali/ws-java-se/Sdms/coverters/linux/converter4",
            AMPL2DAG_LINUX  = COCOSUBSET_LINUX + "/" + "ampl2dag.pl",
            DAGSTRUCT_LINUX = COCOSUBSET_LINUX + "/" + "dag_struct",
            //SOLCHECK_LINUX  = COCOSUBSET_LINUX + "/" + "solcheck"; 
            SOLCHECK_LINUX  = "";
     
    private final Problem p;

    private final String path;

    TestEnvConverter(Problem p, String path) {

        this.p=p;

        this.path=path;
    }

    private static String MODOUT=MyUtils.getWorkingDir()+MyUtils.getFileSep()+
            "StructCSP"+MyUtils.getFileSep();

    private static String RESOUT=MyUtils.getWorkingDir()+MyUtils.getFileSep()+
            "StructCSP_sol"+MyUtils.getFileSep();    

    void run() {

        System.out.println("***************************************************");
        System.out.println(path);

        String outFileName = path.substring(path.indexOf("data/")+5, path.lastIndexOf("/"));

        p.checkReferenceSolutions();

        if(!new File(MODOUT).exists()){
            new File(MODOUT).mkdirs();
        }

        if(!new File(RESOUT).exists()){
            new File(RESOUT).mkdirs();
        }

        char k=65;

        for (ReferenceSolution rs : p.getReferenceSolutions()) {
          try {
            dump(outFileName, rs, k++);
          } catch (ShellException ex) {
            ex.printStackTrace();
          }
        }

        return;
    }

    
    private void dump(String outFileName, ReferenceSolution rs, char k) throws ShellException {

        String fileBaseName = outFileName+"_"+ k;

        String modFile = MODOUT+fileBaseName + ".mod";

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(modFile))) {

            bw.write(p.getModel().toAmplString());
            System.out.println(modFile +" created.");
        }
        catch (Exception e) {

            System.err.println("Writing file \""+modFile+"\" failed");
            throw new AssertionError();
        }
        
        checkIfCanReadBackDumped(modFile);
        
        //---------------------------------------------------
        String dagFile = MODOUT+fileBaseName + ".dag";
        ShellCommand comm=null;
        
        if(MyUtils.isWin() && !AMPL2DAG_WIN.isEmpty()){
          comm=new ShellCommand(PERL_WIN, new String[]{AMPL2DAG_WIN, 
              modFile, COCOSUBSET_WIN, AMPL_WIN});
        }else if(MyUtils.isUnix() && !AMPL2DAG_LINUX.isEmpty()){
          comm=new ShellCommand(PERL_LINUX, new String[]{AMPL2DAG_LINUX, 
              modFile, COCOSUBSET_LINUX, AMPL_LINUX}, 
                  new String[]{"LD_LIBRARY_PATH="+COCOSUBSET_LINUX});
          //System.out.println(comm);
        }
        
        if(comm!=null){
        try {
          /*ShellCommandResult scr =*/ comm.execute();
          //System.out.println(scr.getOutputString());
          //System.err.println(scr.getErrorString());          
          //System.out.println(dagFile +" created.");
        } catch (Exception ex) {
          ex.printStackTrace();
          throw new AssertionError();
        }
        }
        //---------------------------------------------------
        String diFile = MODOUT+fileBaseName + ".di";
        comm=null;
        
        if(MyUtils.isWin() && !DAGSTRUCT_WIN.isEmpty()){
          comm=new ShellCommand(DAGSTRUCT_WIN, new String[]{"-i", 
              dagFile});
        }else if(MyUtils.isUnix() && !DAGSTRUCT_LINUX.isEmpty()){
          comm=new ShellCommand(DAGSTRUCT_LINUX, new String[]{"-i", 
              dagFile}, new String[]{"LD_LIBRARY_PATH="+COCOSUBSET_LINUX});
        }
        
        if(comm!=null){
          try {
            ShellCommandResult res=comm.execute();
            MyUtils.writeFile(diFile, res.getOutputString());
            System.out.println(diFile +" created.");
         } catch (Exception ex) {
            ex.printStackTrace();
           throw new AssertionError();
         }
        }
        //---------------------------------------------------
        String resFile = RESOUT + fileBaseName + ".res";

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(resFile))) {
            
            bw.write("modelstatus = 0\n");

            RealVector x = rs.getXMid();

            ArrayList<String> varNames = rs.getVarNames();
            
            // It is a horrible hack because it assumes that the .dag file already exists... 
            ArrayList<String> namesInCoconut = getVarNamesInCoconutOrder(MODOUT+fileBaseName+".dag", varNames);
            
            if(!namesInCoconut.equals(varNames)){
              System.out.println("Permuted variables detected!");
            }

            int n=x.getDimension();
            
            String[] vals=new String[n];
            
            for (int i = 0; i < n; i++) {
                String vn=varNames.get(i);
                int vind = p.getModel().getVar(vn).getIndex();
                double val=x.get(vind);
                int outind=namesInCoconut.indexOf(vn);
                vals[outind] = "x(" + (outind+1) + ") = " + val + "\n"; 
            }
            
            for (int i = 0; i < n; i++){
                bw.write(vals[i]);
            }
            
            bw.write("obj = 0.0\n");
            bw.write("infeas = 0.00\n");
            bw.write("nonopt = 0.00\n");
            System.out.println(resFile +" created.");
        }
        catch (IOException e) {

            System.err.println(e.getMessage());
            System.err.println("Writing file \""+resFile+"\" failed");
            throw new AssertionError();
        }
        
        //---------------------------------------------------
        String chkFile = RESOUT+fileBaseName + ".chk";
        comm=null;
        
        /*
         "D:\Projects\TestEnvironment\external\win\solcheck.exe" 
         * -u -N -e 1.0E-6 -t 1.0 -n 1 
         * "C:\Users\dferi\AppData\Roaming\TestEnvironment\libs\StructCSP\bratu_020_A.dag" 
         * "C:\Users\dferi\AppData\Roaming\TestEnvironment\libs\StructCSP_sol\bratu_020_A.res" 
         * "C:\Users\dferi\AppData\Roaming\TestEnvironment\libs\StructCSP_sol\bratu_020_A">
         */
        
        if(MyUtils.isWin() && !SOLCHECK_WIN.isEmpty()){
          comm=new ShellCommand(SOLCHECK_WIN, new String[]{
            "-u", "-N", "-e 1.0E-6", "-t 1.0", "-n 1",dagFile, resFile, 
            RESOUT+fileBaseName});
        }else if(MyUtils.isUnix() && !SOLCHECK_LINUX.isEmpty()){
          comm=new ShellCommand(SOLCHECK_LINUX, new String[]{
            "-u", "-N", "-e 1.0E-6", "-t 1.0", "-n 1",dagFile, resFile, 
            RESOUT+fileBaseName},
                  new String[]{"LD_LIBRARY_PATH="+COCOSUBSET_LINUX});
          //System.out.println(comm);
        }
        
        if(comm!=null){
          try {
            /*ShellCommandResult scr = */comm.execute();
            //System.out.println(scr.getOutputString());
            //System.err.println(scr.getErrorString()); 
            //System.out.println(chkFile +" created.");
            TestEnvChkFile techk=new TestEnvChkFile(chkFile);
            if(techk.getMaxviolation()>0){
              System.out.println("SOLUTION CHECK FAILED, violation of "+
                      techk.getMaxviolation()+" at the constraint '"+
                      techk.getMaxviolatedconstr()+"'!");
              throw new AssertionError();
            }else{
              System.out.println("SOLUTION CHECK PASSED.");
            }
         } catch (Exception ex) {
            ex.printStackTrace();
           throw new AssertionError();
         }
        }
        //---------------------------------------------------

        return;
    }

    private static void checkIfCanReadBackDumped(String modFile) {

        Problem p=new Problem(modFile);
        try{
            p.read();
        }catch(Exception e){
            e.printStackTrace();
            throw new AssertionError();
        }
    }

    private static ArrayList<String> getVarNamesInCoconutOrder(String dagFile, ArrayList<String> varNames) {

        try (BufferedReader br = new BufferedReader(new FileReader(dagFile))) {

            return readLineByLine(br, new TreeSet<>(varNames));
        }
        catch (Exception e) {
            System.err.println(e.getMessage());
            System.err.println("Reading file \""+dagFile+"\" failed");
            throw new AssertionError();
        }
    }
    
    private static ArrayList<String> readLineByLine(BufferedReader br, TreeSet<String> varNames) throws IOException {
        
        ArrayList<String> namesInCoconut = new ArrayList<>(varNames.size());
        
        String line;

        while ((line=br.readLine())!=null) {

            if (line.startsWith("<N> N ")) { // A line containing a name

                String name = line.substring(line.indexOf('\'')+1, line.lastIndexOf('\'')); // Names are enclosed: 'name'

                if (varNames.contains(name)) {

                    namesInCoconut.add(name);
                }
            }
        }
        
        assert namesInCoconut.size()==varNames.size() : "Names found in Coconut: "+namesInCoconut+", names in JGloptlab: "+varNames.size();
        
        return namesInCoconut;
    }
}
