//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import dag.Dag;
import dag.DagFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import lmbopt.fun.DagProblem;
import problem.Problem;
import problem.solution.ReferenceSolution;
import sdms.NLPProblem;
import sdms.Solution;
import sdms.SubNLPProblem;
import sdms.Util;
import tools.console.Console;
import tools.console.DefaultConsole;
import tools.math.interval.real.RealInterval;
import tools.math.interval.real.RealIntervalVector;

final class NLPProblemImpl implements NLPProblem {

    private static final Logger log = LogManager.getLogger();
    
    private Problem prob;
    
    private ArrayList<TreeMap<String,Double>> refSols;
    
    private TreeMap<String,Double> rowScales;
    
    private TreeMap<String,Double> colScales;

    NLPProblemImpl(Problem p, Scales scales) {
        
        prob = p;
        
        refSols = new ArrayList<>();
        
        for (ReferenceSolution rs : p.getReferenceSolutions()) {
            
            pushBack(rs);
        }
        
        rowScales = scales.rowScales;
        
        colScales = scales.colScales;
    }
    
    @Override
    public double[] columnScales(SortedSet<String> vars) {
       
        return Util.extractValues(colScales, vars);
    }
    
    @Override
    public TreeMap<String,Double> colScales(SortedSet<String> vars) {
       
        return Util.extractKeyValues(colScales, vars);
    }
    
    @Override
    public TreeMap<String,Double> rowScales(SortedSet<String> cons) {
        
        return Util.extractKeyValues(rowScales, cons);
    }

    private TreeMap<String,Double> rowScalesSquared(SortedSet<String> cons) {
       
        TreeMap<String,Double> res = new TreeMap<>();

        for (String con : cons) {

            Double value = rowScales.get(con);
            
            assert value!=null : "Constraint \""+con+"\" not found";
            
            double w = value;
            
            res.put(con, w*w);
        }

        return res;
    }
    
    private TreeMap<String,Double> objectiveScales(SortedSet<String> cons) {
        
        TreeMap<String,Double> res = new TreeMap<>();
        
        int dimension = cons.size();

        for (String con : cons) {

            Double value = rowScales.get(con);
            
            assert value!=null : "Constraint \""+con+"\" not found";
            
            double w = value;
            
            res.put(con, (w*w)/dimension);
        }

        return res;
    }    
    
    private void pushBack(ReferenceSolution rs) {
        
        RealIntervalVector x = rs.getX();
        
        ArrayList<String> names = rs.getVarNames();
        
        TreeMap<String,Double> sol = new TreeMap<>();
        
        for (int i=0, size=names.size(); i<size; ++i) {
            
            String name = names.get(i);
          
            if (!name.startsWith("_")) {
                
                sol.put(name, x.get(i).getMid());
            }
        }
        
        refSols.add(sol);
    }
    
    static private void checkConsistency( SortedSet<String> selectedConstraints,
                                          TreeMap<String, Double> fixedVariableValueMap,
                                          TreeSet<String> varsToGuess,
                                          TreeMap<String, String[]> fullSparsityOnlyForDebugging )
    {
        
//        log.info("Fixed   "    + fixedVariableValueMap.keySet());
//        log.info("Guessed "    + varsToGuess );
//        log.info("Constraints" + selectedConstraints);
//        log.info("Vars in constraints according to sparsity");
//        for (String s : selectedConstraints)
//            log.info(s + "   " + Arrays.asList(fullSparsityOnlyForDebugging.get(s)));
        
        TreeSet<String> variablesInCons = new TreeSet<>();
        
        for (String s : selectedConstraints) {
            
            variablesInCons.addAll(Arrays.asList(fullSparsityOnlyForDebugging.get(s)));
        }

        for (String s : fixedVariableValueMap.keySet()) {
            
            boolean was_present = variablesInCons.remove(s); 
            
            assert was_present : "The selected constraints do not depend on \""+s+"\" but we are trying to fix it";            
        }
        
        for (String s : varsToGuess) {
            
            boolean was_present = variablesInCons.remove(s);
            
            assert was_present : "The selected constraints do not depend on \""+s+"\" but we are guessing it"; 
        }
        
        assert variablesInCons.isEmpty() : "The following variables should be either fixed or guessed: "+variablesInCons;
    }

    @Override
    public SubNLPProblem generateSubOptProblem(SortedSet<String> selectedConstraints,
                                               TreeMap<String, Double> fixedVariableValueMap,
                                               TreeSet<String> varsToGuess,
                                               TreeMap<String, String[]> fullSparsityOnlyForDebugging) 
    {
        checkConsistency(selectedConstraints, fixedVariableValueMap, varsToGuess, fullSparsityOnlyForDebugging);            

        try {

            Console out = new DefaultConsole();
            boolean debug = false;  // TODO Disable debug logging in a cleaner way

            // objective in the input prob is discarded!
            Dag subCSPDag = DagFactory.generateSubCSP(prob.getModel(), true,
                                                      selectedConstraints,
                                                      fixedVariableValueMap, out, debug);
            //RealInterval pow=new RealInterval(1.1);
            RealInterval pow=null; //use the squared L2 norm
            
            Dag subOptDag = DagFactory.CSPToOptProb(subCSPDag, true, 
                                                    rowScalesSquared(selectedConstraints),
                                                    pow, out, debug);

            DagProblem subOP = new DagProblem(new Problem(subOptDag), out, false);

            return new SubNLPProblemImpl(subCSPDag,
                                         subOP,
                                         selectedConstraints,
                                         fixedVariableValueMap.navigableKeySet(),
                                         varsToGuess,
                                         fullSparsityOnlyForDebugging);
        } 
        catch (Exception e) {
            e.printStackTrace();
            log.fatal("Subproblem generation failed");
            throw new AssertionError();
        }
    }
    
    @Override
    public SubNLPProblem generateSubOptProblem(TreeSet<String> constraints,
                                               TreeSet<String> variables,
                                               TreeMap<String,String[]> fullSparsityOnlyForDebugging) 
    {
        checkConsistency(constraints, variables, fullSparsityOnlyForDebugging);

        try {

            Console out = new DefaultConsole();
            boolean debug = false;  // TODO Disable debug logging in a cleaner way

            // objective in the input prob is discarded!
            Dag subCSPDag = DagFactory.generateSubCSP(prob.getModel(), true,
                                                      constraints,
                                                      new TreeMap<String,Double>(), out, debug);
            //RealInterval pow=new RealInterval(1.1);
            RealInterval pow=null; //use the squared L2 norm
            
            Dag subOptDag = DagFactory.CSPToOptProb(subCSPDag, true, 
                                                    objectiveScales(constraints),
                                                    pow, out, debug);

            DagProblem subOP = new DagProblem(new Problem(subOptDag), out, false);

            return new SubNLPProblemImpl(subCSPDag,
                                         subOP,
                                         constraints,
                                         fullSparsityOnlyForDebugging);
        } 
        catch (Exception e) {
            e.printStackTrace();
            log.fatal("Subproblem generation failed");
            throw new AssertionError();
        }      
    }

    private static void checkConsistency(TreeSet<String> constraints, TreeSet<String> variables, TreeMap<String, String[]> fullSparsityOnlyForDebugging) {

        TreeSet<String> variablesInCons = new TreeSet<>();
        
        for (String con : constraints) {
            
            variablesInCons.addAll( Arrays.asList(fullSparsityOnlyForDebugging.get(con)) );
        }

        for (String var : variables) {
            
            boolean was_present = variablesInCons.remove(var); 
            
            assert was_present : "The selected constraints do not depend on \""+var+"\"";            
        }
        
        assert variablesInCons.isEmpty() : "The following variables should be either fixed or guessed: "+variablesInCons;        
    }
    
    @Override
    public double[] lowerBoundsOfVars(TreeSet<String> vars) { // TODO Duplicate of SubNLPProblemImpl.getLowerBoundsOfGuessedVars()

        double[] bnd = new double[vars.size()];

        int k = 0;

        for (String var : vars) {

            bnd[k++] = prob.getModel().getVar(var).getBounds().getLower();
        }

        return bnd;
    }

    @Override
    public double[] upperBoundsOfVars(TreeSet<String> vars) {

        double[] bnd = new double[vars.size()];

        int k = 0;

        for (String var : vars) {

            bnd[k++] = prob.getModel().getVar(var).getBounds().getUpper();
        }

        return bnd;
    }

    @Override
    public int[] compareWithReferenceSolutions(List<Solution> allSolutions, double INFEASIBILITY_THRESHOLD, double maxDistance) {

        int[] count = new int[refSols.size()];

        for (Solution s : allSolutions) {

            if (s.getNormOfInfeasibility() >= INFEASIBILITY_THRESHOLD) {
                
                continue;                
            }

            int i = referenceSolutionIndex(s, maxDistance);

            if (i==-1) {
                
                log.info("New solution: "+s.toString());
            }
            else {
                
                ++count[i];
            }
        }

        return count;
    }
    
    @Override
    public ArrayList<TreeMap<String,Double>> getReferenceSolutions() {
     
        return refSols;
    }

    private int referenceSolutionIndex(Solution s, double threshold) {
        
        int i=0;

        for (TreeMap<String,Double> rs : refSols) {
            
                TreeMap<String,Double> sol = s.getAllVars();
                
                assert rs.size()==sol.size();
               
                Iterator<Double> it = sol.values().iterator();
                
                double sum = 0.0;
                
                for (double x : rs.values()) {
                    
                    double diff = x - it.next();
                    
                    sum += (diff*diff);
                }
                
                if (Math.sqrt(sum) < threshold)
                    
                    return i;
     
                ++i;
        }
       
        return -1;
    }

    @Override
    public int numberOfVars() {
        
        return prob.getModel().getVarNumber();
    }    
    
}
