//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//

package sdms.impl;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeMap;

import sdms.Plotter;
import sdms.ProgramTerminationRequested;
import sdms.Solution;
import sdms.Util;

final class PlotterImpl implements Plotter {
    
    private int counter = 0;
    private int nvar = -1;
    private boolean hasRefSols = false;
    
    private String prefix = "/tmp/";
    
    private String[] names;
    private ArrayList<String> filteredNames;
    
    private double LB = Double.NaN;
    private double UB = Double.NaN;

    private void initBratu(int n) {
        LB = 0;
        UB = 5;
        names = new String[n];
        for (int i=1; i<=n; ++i) {
            names[i-1] = "x["+i+"]";
        }
    }
    
    private void initJacobsenTorn() {
        names = new String[]{ "x[1]", "x[2]", "x[3]", "x[4]", "x[5]", "x[6]", "x[7]", "x[8]" };
        LB = 0;
        UB = 1;
        
        //names = new String[]{ "V[1]", "V[2]", "V[3]", "V[4]", "V[5]", "V[6]", "V[7]", "L[8]" };
        //LB = 2;
        //UB = 4;
    }
    
    private void initMSSTorn(int N) {
        LB = 0;
        UB = 1;
        names = new String[N];
        for (int i=1; i<=N; ++i) {
            names[i-1] = "x[1,"+i+"]";
        }   
    }
    
    private void initBlockEx() {
        LB = -1;
        UB =  3;
        int n = 20;
        names = new String[n];
        for (int i=1; i<=n; ++i) {
            names[i-1] = "x["+i+"]";
        }
    }    
   
    PlotterImpl(String file) {
        
        if (file.equals("data/mss20TornBlk/mss20Torn.mod") ||
            file.equals("data/mss20Torn/mss20Torn.mod")    || 
            file.equals("data/mss20TornLastTwoMerged/mss20Torn.mod")  || 
            file.equals("data/mss20TornScaled/mss20TornScaled.mod")   || 
            file.equals("data/mss20TornScaledD/mss20TornScaledD.mod") ||
            file.equals("data/mss20TornUnscaledD/mss20TornUnscaledD.mod") ) 
        {     
            initMSSTorn(20);
        }
        else if (file.equals("data/mss10Torn/mss10Torn.mod") || 
                 file.equals("data/mss10TornLastTwoMerged/mss10Torn.mod") || 
                 file.equals("data/mss10TornScaled/mss10TornScaled.mod") ) 
        {     
            initMSSTorn(10);
        }
        else if (file.equals("data/mss40TornScaled/mss40TornScaled.mod") ) 
        {     
           initMSSTorn(40);
        }
        else if (file.equals("data/mss50/mss50TornScaledD.mod")) 
        {     
           initMSSTorn(50);
        }        
        else if (file.equals("data/JacobsenTornNatruralOrdering/JacobsenTorn.mod")) {

            initJacobsenTorn();
        }
        else if (file.equals("data/bratu_020/bratu.mod")) {
            
            initBratu(20);
        }
        else if (file.equals("data/bratu250/bratu250.mod")) {
            
            initBratu(250);
        }
        else if (file.equals("data/blockEx/blockEx.mod")) {
            
            initBlockEx();
        }
        else {
            System.err.println("Unknown test file \""+file+"\"");
            throw new ProgramTerminationRequested();
        }
        
        filteredNames = new ArrayList<>(Util.filterNameArray(names));
    }

    @Override
    public void writePartialSolutions(List<Solution> solutions) {

        writeSolutionsToFile(convert2Map(solutions));
    }
    
    @Override
    public void writeNearestSolutions(List<TreeMap<String,Double>> solutions) {
        
        writeSolutionsToFile(solutions);
    }
    
    private void writeSolutionsToFile(List<TreeMap<String,Double>> solutions) {
        
        if (solutions.isEmpty()) {
            
            return;
        }
        
        writeToFile(prefix+"profiles_"+counter+".txt", solutions);
        
        ++counter;        
    }
    
    private static ArrayList<TreeMap<String,Double>> convert2Map(List<Solution> solutions) {

        ArrayList<TreeMap<String,Double>> sols = new ArrayList<>(solutions.size());
        
        for (Solution s : solutions) {
            
            sols.add(s.getAllVars());
        }
        
        return sols;
    }
    
    @Override
    public void writeReferenceSolutions(List<TreeMap<String,Double>> solutions) {
        
        assert !hasRefSols : "Reference solutions have already been written!";
    
        hasRefSols = true;
        
        if (solutions.isEmpty()) {
            
            return;
        }
        
        writeToFile(prefix+"refsols.txt", solutions);
    }    
    
    private void writeToFile(String fileName, List<TreeMap<String,Double>> solutions) {
        
        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {
            
            int ncols = solutions.size();
            
            int nrows= countResolvedVars(solutions);
            
            bw.write(nrows+"   "+ncols+"\n");
            
            bw.write(   LB+"   "+UB   +"\n");
            
            writeResolvedIndices(bw, solutions.get(0).keySet());
            
            writePartialSolutions(bw, solutions);
        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }
    }
    
    // Superfluous, same as solutions.get(0).getVarNames.size();
    private int countResolvedVars(List<TreeMap<String,Double>> solutions) {
        
        int nrows=0;
        
        TreeMap<String,Double> vars = solutions.get(0);
        
        for (String s : vars.keySet()) {
            
            if (filteredNames.contains(s)) {
                
                ++nrows;
            }
        }
        
        nvar = Math.max(nvar, nrows);
        
        return nrows;
    }
    
    private ArrayList<Integer> getResolvedIndices(Set<String> varNames) {
        
        ArrayList<Integer> indices = new ArrayList<>();
        
        for (int i=0; i<names.length; ++i) {
         
            boolean resolved = varNames.contains(filteredNames.get(i));
            
            if (resolved) {
                
                indices.add(i);
            }
        }
        
        return indices;
    }
     
    private void writeResolvedIndices(BufferedWriter bw, Set<String> varNames) throws IOException {
        
        ArrayList<Integer> indices = getResolvedIndices(varNames);
        
        for (Integer i : indices) {
        
            bw.write(i.toString());
        
            bw.newLine();
         }
        
        bw.newLine();       
    }
    
    private void writePartialSolutions(BufferedWriter bw, List<TreeMap<String,Double>> solutions) throws IOException {

        for (TreeMap<String,Double> val : solutions) {
            
            Double d;
            
            for (String name : filteredNames) { 
                
                if ((d=val.get(name))!=null) {
                
                    bw.write(d+"\n");
                }
            }
            
            bw.write("\n");
        }        
    }

    @Override
    public void done(String subProbsLogFile, String fullProbLogFile) {

        writeIndices();
        
        writeNames();
        
        writeCleanup(subProbsLogFile, fullProbLogFile);
        
        Util.copyFile("asy/plot.asy", prefix+"plot.asy", StandardCopyOption.REPLACE_EXISTING);
        
        Util.executeCmd("./asy/generate_plot.sh");
    }
    
    private void writeIndices() {
        
        String fileName = prefix+"indices.txt";
        
        assert nvar == names.length : "nvar = "+nvar+", names.length = "+names.length;

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {
            
            bw.write(counter+"\n");
            
            bw.write(names.length+"\n");

        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }        
    }
    
    private void writeNames() {
        
        String fileName = prefix+"names.txt";

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {
            
            for (String name : names) {
            
                bw.write(name+"\n");
            }
        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }             
    }
    
    private void writeCleanup(String subProbsLogFile, String fullProbLogFile) {
        
        String fileName = prefix+"cleanup.sh";

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {

            bw.write("rm -f names.txt\n");
            bw.write("rm -f indices.txt\n");
            bw.write("rm -f profiles_*.txt\n");
            bw.write("rm -f "+fullProbLogFile+"\n");
            bw.write("rm -f plot.asy\n");
            bw.write("rm -f plot.pdf\n");
            bw.write("rm -f refsols.txt\n");            
            bw.write("rm -f sdmsConsole.log\n");
            bw.write("rm -f sdms.log\n");
            bw.write("rm -f "+subProbsLogFile+"\n");            
            bw.write("rm -f success.txt\n");
            bw.write("rm -f cleanup.sh\n");
        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }
        
        String cmd = "chmod +x "+fileName;
        
        Util.executeCommand(cmd);        
    }
}
