//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import sdms.Plotter;
import sdms.Solution;
import sdms.Util;

final class PlotterTriangle implements Plotter {

    private int counter = 0;
    private boolean hasRefSols = false;
    
    private static final String prefix = "/tmp/";    
    
    PlotterTriangle() { }
    
    static final class Pair {
        final double x;
        final double y;
        Pair(double x, double y) { this.x=x; this.y=y; }
    }

    @Override
    public void writePartialSolutions(List<Solution> solutions) {

        writeSolutionsToFile(convert2Map(solutions));
    }
    
    @Override
    public void writeNearestSolutions(List<TreeMap<String,Double>> solutions) {
        
        writeSolutionsToFile(solutions);
    }
    
    private void writeSolutionsToFile(List<TreeMap<String,Double>> solutions) {
        
        if (solutions.isEmpty()) {

            return;
        }
        
        ArrayList<Pair[]> profiles = extractProfiles(solutions);
        
        String fileName = prefix+"triangle_"+counter+".txt";

        writeToFile(fileName, profiles);
        
        ++counter;        
    }
    
    private static ArrayList<TreeMap<String,Double>> convert2Map(List<Solution> solutions) {

        ArrayList<TreeMap<String,Double>> sols = new ArrayList<>(solutions.size());
        
        for (Solution s : solutions) {
            
            sols.add(s.getAllVars());
        }
        
        return sols;
    }
    
    @Override
    public void writeReferenceSolutions(List<TreeMap<String,Double>> solutions) {
        
        assert !hasRefSols : "Reference solutions have already been written!";
    
        hasRefSols = true;
        
        if (solutions.isEmpty()) {
            
            return;
        }
        
        ArrayList<Pair[]> profiles = extractProfiles(solutions);
        
        writeToFile(prefix+"refsols.txt", profiles);
    }

    private static ArrayList<Pair[]> extractProfiles(List<TreeMap<String,Double>> solutions) {
        
        ArrayList<Pair[]> profiles = new ArrayList<>(solutions.size());

        for (TreeMap<String,Double> s : solutions) {        

            double[] x1 = extract(Util.filter("x[1,"), s);
            double[] x2 = extract(Util.filter("x[2,"), s);
            double[] x3 = extract(Util.filter("x[3,"), s);
            
            profiles.add( normalize(x1, x2, x3) );
        }
        
        return profiles;
    }
    
    private static double[] extract(String prefix, TreeMap<String,Double> s) {
        
        TreeMap<Integer,Double> values = new TreeMap<>();
        
        for (Map.Entry<String,Double> e : s.entrySet()) {
            
            String name = e.getKey(); 
            
            if (name.startsWith(prefix)) {
                
                int index = extractIndex(name, prefix.length());
                
                Double old = values.put(index, e.getValue());
                
                assert old==null : "Name: \""+name+"\"";
            }            
        }

        return Util.values(values);
    }
    
    private static int extractIndex(String name, int prefixLen) {
        
        int end = name.indexOf(Util.filter("]"), prefixLen);
        
        assert end!=-1 : "Failed to find ] in "+name+", starting from position "+prefixLen;
        
        String indexAsString = name.substring(prefixLen, end);
        
        return Integer.parseInt(indexAsString);
    }
    
    private static Pair[] normalize(double[] x1, double[] x2, double[] x3) {
        
        int n = x1.length;
        
        assert x2.length == n;
        assert x3.length == n;
        
        Pair[] xNormalized = new Pair[n];
        
        for (int i=0; i<n; ++i) {
        
            double sum = x1[i] + x2[i] + x3[i];
            
            assert sum > 0.85 : "x1 + x2 + x3 = "+sum;
            
            xNormalized[i] = new Pair(x1[i]/sum, x2[i]/sum);
        }
        
        return xNormalized;
    }
    
    private static void writeToFile(String fileName, ArrayList<Pair[]> profiles) {

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {
            
            int ncols = profiles.size();
            
            int nrows = profiles.get(0).length;
            
            bw.write(nrows+"   "+ncols+"\n\n");
            
            writePartialSolutions(bw, profiles);
        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }
    }
    
    private static void writePartialSolutions(BufferedWriter bw, ArrayList<Pair[]> profiles) throws IOException {

        for (Pair[] prof : profiles) {

            for (Pair p : prof) { 

                bw.write(p.x+", "+p.y+"\n");
            }

            bw.write("\n");
        }        
    }
    
    @Override
    public void done(String subProbsLogFile, String fullProbLogFile) {

        writeCounter();
        
        writeCleanup(subProbsLogFile, fullProbLogFile);
        
        Util.copyFile("asy/plotTriangle.asy", prefix+"plotTriangle.asy", StandardCopyOption.REPLACE_EXISTING);
        
        Util.executeCmd("./asy/generate_triangle_plot.sh");
    }    

    private void writeCounter() {
        
        String fileName = prefix+"counter.txt";

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {
            
            bw.write(counter+"\n");
        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }        
    }
    
    private static void writeCleanup(String subProbsLogFile, String fullProbLogFile) {
        
        String fileName = prefix+"cleanup.sh";

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(fileName))) {

            bw.write("rm -f counter.txt\n");
            bw.write("rm -f "+fullProbLogFile+"\n");          
            bw.write("rm -f plotTriangle.asy\n");
            bw.write("rm -f plotTriangle.pdf\n");
            bw.write("rm -f refsols.txt\n"); 
            bw.write("rm -f sdms.log\n");            
            bw.write("rm -f sdmsConsole.log\n");
            bw.write("rm -f "+subProbsLogFile+"\n");             
            bw.write("rm -f success.txt\n");
            bw.write("rm -f triangle_*.txt\n");              
            bw.write("rm -f cleanup.sh\n");
        }
        catch (IOException e) {
            
            throw Util.logIOError(e, fileName, "writing");
        }
        
        String cmd = "chmod +x "+fileName;
        
        Util.executeCommand(cmd);        
    }
}
