//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ferenc Domes
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import dag.Dag;
import dag.DagCache;
import dag.methods.DagMethod;
import dag.methods.DagMethodException;
import dag.methods.propag.DagConstraintPropagator;
import dag.nodes.input.VarNode;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import problem.Problem;
import problem.ProblemParserException;
import problem.solution.ReferenceSolution;
import tools.SanityCheckException;
import tools.console.Console;
import tools.console.DefaultConsole;
import tools.math.RealVector;
import tools.math.function.Function;
import tools.math.function.FunctionGradCheck;
import tools.math.function.FunctionSolver;
import tools.math.function.FunctionSolverOptions;
import tools.math.interval.real.IMAMath;
import tools.opt.OptException;
import tools.opt.SolverResult;
import tools.opt.ProblemSolutionStatus;
import tools.profiler.Profiler;

/**
 *
 * @author dferi
 */
class Sdms {

  private static final Logger log = LogManager.getLogger();
  private static boolean CHECK_REFSOL_ORIG = false;
  private static boolean CHECK_REFSOL_SUBST = false;
  private static boolean CHECK_SUBST_MODEL = true;
  private static double CHECK_ERROR_TRESH = 1e-8;
//  private final static String FILENAME = MyUtils.getWorkingDir()
//          + MyUtils.getFileSep() + "data" + MyUtils.getFileSep() + "JacobsenJM2New.mod";
//  private final static TreeMap<String, Double> SEL_CONSTR_WEIGHT_MAP = new TreeMap<>();
//
//  static {
//    SEL_CONSTR_WEIGHT_MAP.put("M_eq_1_", 1.0);
//    SEL_CONSTR_WEIGHT_MAP.put("H_eq_2_", 1.0);
//  }
//  private final static TreeMap<String, Double> FIXED_VAR_VALUE_MAP = new TreeMap<>();
//
//  static {
//    FIXED_VAR_VALUE_MAP.put("x_1_", 1.0);
//  }

  static void setProblemCheck(boolean checkRefsolOnOriginal,
          boolean checkSubstitutedModelFeasibility,
          boolean checkRefsolOnSubstituted,
          double checkErrorTreshold) {
    CHECK_REFSOL_ORIG = checkRefsolOnOriginal;
    CHECK_REFSOL_SUBST = checkRefsolOnSubstituted;
    CHECK_SUBST_MODEL = checkSubstitutedModelFeasibility;
    CHECK_ERROR_TRESH = checkErrorTreshold;
  }

  static Problem readProblem(String file, boolean debug)
          throws IOException,
          ProblemParserException, ParseException,
          DagMethodException {
    Console out = debug ? new DefaultConsole() : null;
    //************* READ THE PROBLEM *************
    Profiler.turnOff();
    Problem p = new Problem(file);
    p.read(false, false, false, out);
    if (debug && log.isDebugEnabled()) {
      log.debug("**** INITIAL PROBLEM ****");
      log.debug(p.toString());
      for (ReferenceSolution rs : p.getReferenceSolutions()) {
        log.debug("Reference solution (original)=" + rs);
        if (CHECK_REFSOL_ORIG) {
          //check the reference solution of the original problem
          //use CP with a widened boy around the reference solution
          DagCache cache = new DagCache(p.getModel());
          cache.setVarBounds(IMAMath.widen(rs.getX(), CHECK_ERROR_TRESH, false));
          log.debug(cache);
          DagConstraintPropagator cp = new DagConstraintPropagator(cache,
                  out, DagMethod.OUTPUT_NONE, 0.01, 100);
          cp.internalStart();
          if (cache.isInfeasible()) {
            String msg = "No point in the "
                    + CHECK_ERROR_TRESH + " width box around the reference "
                    + "solution satisfies the constraints, violated node:"
                    + cache.getInfeasibleNode();
            log.fatal(msg);
            throw new SanityCheckException(msg);
          }
          log.debug("Propagated tiny box around the reference solution="
                  + cache.getVarBounds().getIntervalHull());
        }
      }
    }

    //************* BACKSUBSTITUTION *************
    //Assume that all variables without bounds can be backsubstituted
    ArrayList<String> substVars = new ArrayList<>();
    for (VarNode vn : p.getModel().getVars()) {
      if (vn.getBounds().isInfinite()) {
        substVars.add(vn.getName());
      }
    }
    //p.backSubstitution(new String[]{"_x_"}, new String[]{"generated_equ_name_"});
    p.backSubstitution(substVars.toArray(new String[substVars.size()]),
            new String[]{"generated_equ_name_"});
    Dag dag = p.getModel();
    if (debug && log.isDebugEnabled()) {
      log.debug("**** BACK SUSBSTITUTED PROBLEM ****");
      log.debug(dag.toAmplString());
      if (CHECK_SUBST_MODEL) {
        //crude test of model feasiblity
        DagCache cache = new DagCache(p.getModel());
        DagConstraintPropagator cp = new DagConstraintPropagator(cache,
                out, DagMethod.OUTPUT_NONE, 0.01, 100);
        cp.start();
        if (cache.isInfeasible()) {
          String msg = "The substituted problem was found "
                  + "infeasible, violated node:" + cache.getInfeasibleNode();
          log.fatal(msg);
          throw new SanityCheckException(msg);
        } else if (log.isDebugEnabled()) {
          log.debug("Constraint propagation results in \\x="
                  + cache.getVarBounds().getIntervalHull());
        }
      }
      if (CHECK_REFSOL_SUBST) {
        p.checkReferenceSolutions();
        for (ReferenceSolution rs : p.getReferenceSolutions()) {
          log.debug("Reference solution=" + rs);
          double viol = rs.getMaxConstrViol();
          log.debug("Reference solution max violation=" + viol
                  + " at " + rs.getMaxViolatedConstraint());
          log.debug("Reference solution sum violation=" + rs.getSumConstrViol());
          if (viol > CHECK_ERROR_TRESH) {
            throw new SanityCheckException("Substituted problem constraint "
                    + "max violation is too big, " + viol + " at '"
                    + rs.getMaxViolatedConstraint().getLabel() + "'");
          }
        }
      }
      log.debug("Dependcencies:" + dag.getConstraintDependencies());
    }
    return p;
  }

  /*
   static void oldMain(String[] args) throws ProblemParserException, DagMethodException, IOException, IllegalValueException, ParseException, OptException {

   Console out = new DefaultConsole();

   Problem p = readProblem(args);

   //ToDo: select suitable SEL_CONSTR and FIXED_VAR_VALUE_MAP
   //      DagProblem subProblem = generateSubOptProblem(p, SEL_CONSTR_WEIGHT_MAP,
   //              FIXED_VAR_VALUE_MAP, new DefaultConsole(), DEBUG);

   Dag subDag = DagFactory.generateSubCSP(p.getModel(), false, null, null,
   out, DEBUG);// Will fail with NullpointerException eventually
   subDag = DagFactory.CSPToOptProb(subDag, null, out, DEBUG);
   //create a new DagProblem for LMBopt
   DagProblem subProblem = new DagProblem(new Problem(subDag), out, false);

   //choose suitable starting point
   RealVector startPoint = subProblem.getP().getModel().getVarBounds().getMid();
   //startPoint = new RealVector(1, 0.8, 0.8);
   // Check the new subProblem.setXstart methods!

   //call LMBOPT on the new subproblem
   Lmbopt lmbopt = new Lmbopt(out);

   ProblemSolution ps = solveSubproblem(subProblem, startPoint, lmbopt, out);

   ps.toString();
   }*/
  static SolverResult solveSubproblem(Function subProblem,
          RealVector startPoint,
          FunctionSolver solver,
          FunctionSolverOptions solverOptions)
          throws ParseException, IOException, OptException {
    dbgGradient(subProblem, startPoint);

    return solveSubproblemImpl(subProblem, startPoint, solver, solverOptions);
  }
  static final int MAX_GRAD_CHECK_REPEAT = 5;

  @SuppressWarnings("null")
static void dbgGradient(Function function, RealVector x) {

    if (!log.isDebugEnabled()) {
      return;
    }

    double maxError = 1.0e-2;
    FunctionGradCheck fgc = null;

    for (int repeat = 0; repeat < MAX_GRAD_CHECK_REPEAT; repeat++) {
      fgc = FunctionGradCheck.check(function, x, 20, null, false);
      if (Double.isNaN(fgc.getEstimatedRelativeGradAccurarcy())) {
        break;
      }
    }

    double relAcc = fgc.getEstimatedRelativeGradAccurarcy();

    if (Double.isNaN(relAcc)) {

      log.warn("Gradient could not be checked.");

    } else if (fgc.getBad().size() > 1) {

      log.error("Gradient error detected!");

    } else if (relAcc > maxError) {

      log.warn("relAcc = "+relAcc);
      log.warn("The gradient is suspicious");

    } else {

      log.trace("Gradient seems OK!");
    }
  }

  static SolverResult solveSubproblemImpl(Function subProblem,
          RealVector startPoint, FunctionSolver solver,
          FunctionSolverOptions solverOptions) throws ParseException, IOException, OptException 
  {
      
//    startPoint.set(0, 0.513592183685151);
//    startPoint.set(1, 1.02339460776962);
//    startPoint.set(2, 1.52688722534979);
      
    //log.info("Starting point "+startPoint);  
    
    SolverResult ps = solver.solve(subProblem, startPoint, solverOptions);
    ProblemSolutionStatus pss = ps.getStatus();
    if (pss == ProblemSolutionStatus.ACCURACY_ACHIEVED && log.isDebugEnabled()) {
        log.info("**** LOCAL SOLUTION FOUND ****");
        log.info("x=" + ps.getBest());
        log.info("f(x)=" + ps.getBestF());
        log.info("grad evaluations=" + ps.getnObjGradEval());
    }
    else if (pss!=ProblemSolutionStatus.ACCURACY_ACHIEVED) {
        
//        log.warn("grad evaluations=" + ps.getnObjGradEval());
//        log.warn("Local solver returned: "+pss);
//        log.warn("This local solution will be thrown away");
//        throw new RuntimeException();
    }
    return ps;
  }
}
