//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.util.Collection;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeMap;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import sdms.Solution;
import sdms.Util;
import sdms.Violations;

// TODO Clean up the infeasibility norm tracking; for the purposes
// of the development, just recompute it no matter what; this of 
// course leads to an O(n^2) algorithm but we won't notice it 
// due to the tiny constants; remove this recomputed infeasibility 
// norm tracking and replace it with a recursive one once the
// algorithm is finished AND the O(n^2) nature is actually visible 

final class SolutionImpl implements Solution, Cloneable {
    
    private static final Logger log = LogManager.getLogger();
    
    private TreeMap<String, Double> values = null;
    private double infeasibilityBestObjective = Double.NaN;
    private double normOfInfeasibility     = Double.NaN;
    
    private double cumulativeInfeasibility = Double.NaN;
    private double scaledBoundViolSquared  = Double.NaN;
    
    private double infeasibilityInfinityNorm = Double.NaN;
    
    private double infNormScaledBoundViol = Double.NaN;
    
    private boolean fromForwardProjection = false;
    
    SolutionImpl() { }
    
    @Override
    public SolutionImpl clone() {
        SolutionImpl s = new SolutionImpl();
        s.values = new TreeMap<String,Double>(this.values);
        s.infeasibilityBestObjective = this.infeasibilityBestObjective;
        s.normOfInfeasibility        = this.normOfInfeasibility;
        s.cumulativeInfeasibility    = this.cumulativeInfeasibility;
        s.scaledBoundViolSquared     = this.scaledBoundViolSquared;
        return s;
    }
    
    SolutionImpl(TreeMap<String, Double> values, double infeasibilityBestObjective) {
        this.values = values;
        this.infeasibilityBestObjective = infeasibilityBestObjective;
        this.normOfInfeasibility = Math.sqrt(infeasibilityBestObjective); // Knows the norm
        
        this.cumulativeInfeasibility = Double.NaN; // only set by clone, and createEmptySolution
        this.scaledBoundViolSquared  = Double.NaN;
    }
    
    static Solution createEmptySolution() {
        SolutionImpl s = new SolutionImpl();
        s.values = new TreeMap<>();
        s.infeasibilityBestObjective = 0.0;
        s.normOfInfeasibility = 0.0;
        s.cumulativeInfeasibility = 0.0;
        s.scaledBoundViolSquared = Double.NaN;
        return s;
    }
    
    @Override
    public Solution extend(SortedSet<String> variablesToBeRemoved,
                           TreeMap<String, Double> subproblemSolutionForGuessedVars,
                           double infeasBestObjective) 
    {
        SolutionImpl extendedSolution = this.clone();
        
        for (String var : variablesToBeRemoved) {
        
            Double d = extendedSolution.values.remove(var);
            
            if (d==null) {
                
                log.fatal("Variable "+var+" is not present");
                
                throw new AssertionError();
            }
        }
                
        return extend(extendedSolution, subproblemSolutionForGuessedVars, infeasBestObjective);
    }
    
    @Override
    public Solution extend(TreeMap<String, Double> subproblemSolutionForGuessedVars, double infeasBestObjective) {
        
        return extend(this.clone(), subproblemSolutionForGuessedVars, infeasBestObjective);
    }

    private static Solution extend(SolutionImpl extendedSolution, TreeMap<String, Double> subproblemSolutionForGuessedVars, double infeasBestObjective) {
        
        if (log.isTraceEnabled()) {
            log.trace("The solution to be extended: "+extendedSolution.values);
        }
        
        extendedSolution.infeasibilityBestObjective = infeasBestObjective;
        extendedSolution.normOfInfeasibility = Math.sqrt(infeasBestObjective);
        
        for (Map.Entry<String,Double> e : subproblemSolutionForGuessedVars.entrySet()) {
            
            String var = e.getKey();
            
            if (extendedSolution.values.put(var, e.getValue()) != null) {
                
                log.fatal("Variable "+var+" is already in the map");
                
                throw new AssertionError();
            }
        }
        
        if (log.isTraceEnabled()) {
            log.trace("After the extension: "+extendedSolution.values);
        }
        
        return extendedSolution;
    }
    
    @Override
    public boolean isSuccessful() {
        
        return values!=null ? true : false;
    }

    @Override
    public TreeMap<String, Double> getValues(SortedSet<String> subsetOfVars) {
        
        TreeMap<String,Double> val = new TreeMap<>();
        
        for (String s : subsetOfVars) {
            
            Double d = values.get(s);
            
            assert d!=null : "Value for "+s+" not found";
            
            val.put(s, d);
        }
        
        return val;
    }
    
    @Override
    public double[] values(Collection<String> vars) {

        return Util.extractValues(values, vars);
    }
    
    @Override
    public TreeMap<String,Double> dbgGetValues(String... names) {
        
        TreeMap<String,Double> val = new TreeMap<>();
        
        for (String s : names) {
            
            Double d = values.get(s);
            
            if (d!=null) {
                
                val.put(s, d);
            }
        }
        
        return val;
    }
    
    @Override
    public TreeMap<String,Double> getValues(SortedSet<String> subsetOfVars, TreeMap<String,Double> artificallyFixed) {
        
        if (artificallyFixed.isEmpty()) {
            
            return getValues(subsetOfVars);
        }

        TreeMap<String,Double> val = new TreeMap<>();
        
        for (String s : subsetOfVars) {
            
            Double d1 = values.get(s);
            Double d2 = artificallyFixed.get(s);
            
            assert (d1!=null) || (d2!=null) : "Value for "+s+" not found";
            assert (d1==null) || (d2==null) : "Value for "+s+" found twice";
            
            val.put(s, (d2==null) ? d1 : d2);
        }
        
        return val;
    }
    
    @Override
    public double getInfeasibilityBestObjective() {

        return infeasibilityBestObjective;
    }

    @Override
    public String toString() {
        // TODO When the solution comes from IPOPT, the infeasibility norms are in infinity norm and should be set properly!
        return (values==null) ? "[no solution]" : String.format("%g", infeasibilityBestObjective) + "   " + values.toString();
    }

    @Override
    public TreeMap<String, Double> getAllVars() {
       
        return values;
    }
    
    @Override
    public SortedSet<String> getVarNames() {
        
        return values.navigableKeySet();
    }

    @Override
    public int dimension() {
        
        return values.size();
    }

    @Override
    public Collection<Double> getValues() {
        
        return values.values();
    }

    @Override
    public double getInfNormScaledBoundViol() {
        assert infNormScaledBoundViol >= 0.0;
        return infNormScaledBoundViol;
    }
    
    @Override
    public void setInfNormScaledBoundViol(double infNorm) {
        assert 0.0 <= infNorm : "Invalid infinity norm for bound violation: "+infNorm;
        infNormScaledBoundViol = infNorm;
    }
    
    @Override
    public void setScaledBoundconstraintViolation(double violSquared) {
        assert 0.0 <= violSquared : "Invalid argument for bound violation: "+violSquared;
        assert Double.isNaN(scaledBoundViolSquared);
        scaledBoundViolSquared = violSquared;
    }
    
    @Override
    public double getScaledBoundconstraintViolationSquared() {
        assert !Double.isNaN(scaledBoundViolSquared);
        return scaledBoundViolSquared;
    }    
    
    @Override
    public void incrementCumulativeInfeasibilityWithBestObjAndBoundViolSquared() {

        assert 0.0 <= cumulativeInfeasibility :    "Invalid cumulative infeasibility norm: "+cumulativeInfeasibility; 
        assert 0.0 <= infeasibilityBestObjective : "Invalid best objective: "+infeasibilityBestObjective;
        assert 0.0 <= scaledBoundViolSquared :     "Invalid bound violation: "+scaledBoundViolSquared;
        
        cumulativeInfeasibility += (infeasibilityBestObjective + scaledBoundViolSquared);
        
        infeasibilityBestObjective = Double.NaN;
        normOfInfeasibility        = Double.NaN;
        scaledBoundViolSquared     = Double.NaN;        
    }

    @Override
    public double getNormOfInfeasibility() {
        assert 0.0<=normOfInfeasibility : "Invalid infeasibility norm: "+normOfInfeasibility;
        return normOfInfeasibility;
    }

    @Override
    public double getCumulativeInfeasibility() {
        assert 0.0 <= cumulativeInfeasibility : "Invalid cumulative infeasibility norm: "+cumulativeInfeasibility; 
        return cumulativeInfeasibility;
    }
    
    @Override
    public double getInfeasibilityInfinityNorm() {
        assert 0.0 <= infeasibilityInfinityNorm; 
        return infeasibilityInfinityNorm;
    }    

    @Override
    public void setCumulativeInfeasibilitySquared(double viol2) {
        assert Double.isNaN(scaledBoundViolSquared);
        // FIXME Why were these two necessary?
        //assert Double.isNaN(normOfInfeasibility);
        //assert Double.isNaN(infeasibilityBestObjective);
        assert cumulativeInfeasibility >= 0;
        
        assert viol2 >= 0 : "Trying to set "+viol2+" as cumulative infeasibility";

        cumulativeInfeasibility = viol2;
    }
    
    @Override
    public void setAllViolations(Violations viol) {
        
        setCumulativeInfeasibilitySquared(viol.L2SquaredConsAndBound);
        
        infeasibilityInfinityNorm = viol.LinfConsAndBound;
    }
    
    @Override
    public void markAsFromForwardProjection() {
        assert fromForwardProjection==false;
        fromForwardProjection = true;
    }
    
    @Override
    public boolean isFromForwardProjection() {
        return fromForwardProjection;
    }
}
