//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.TreeMap;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import sdms.Solver;
import sdms.Solution;
import sdms.SubNLPProblem;
import tools.math.RealVector;

final class SolutionInfo {
    
    final int funEval;
    final int gradEval;
    final double bestObj;
    final RealVector solution;
    final boolean error;
    
    SolutionInfo(int funEval, int gradEval, double bestObj, RealVector solution, boolean error) {
        this.funEval  = funEval;
        this.gradEval = gradEval;
        this.bestObj  = bestObj;
        this.solution = solution;
        this.error    = error;
    }
    
    static SolutionInfo createFailed() {
        
        return new SolutionInfo(0);
    }
    
    private SolutionInfo(int fEval) { 
        funEval = fEval;
        gradEval = 0;
        bestObj = Double.NaN;
        solution = null;
        this.error = true;
    }
}

abstract class SolverBase implements Solver {

    private static final Logger log = LogManager.getLogger();
    
    private ArrayList<Integer> cntF   = new ArrayList<>();
    private ArrayList<Integer> cntJac = new ArrayList<>();
    private ArrayList<Integer> nRows  = new ArrayList<>();
    private ArrayList<Integer> nCols  = new ArrayList<>();

    private int evalF = 0;
    private int gradF = 0;
    
    abstract protected SolutionInfo runSolver(SubNLPProblemImpl subProb, RealVector startingPoint) throws Exception;
    
    abstract protected String solverOptionsToString();

    @Override
    final public Solution solveSubNLPProblem(SubNLPProblem subProblem,  double[] startingPoint) {

        return solveSubNLPProblemPrivate((SubNLPProblemImpl) subProblem, startingPoint); // TODO Factory must make sure that the cast succeeds 
    }
    
    @Override
    final public Solution solveSubNLPProblem(SubNLPProblem subProblem, double[] startingPoint, double[] solutionOutParam) {
        
        SubNLPProblemImpl subNLPProblemImpl = (SubNLPProblemImpl) subProblem;

        SolutionInfo solInfo = invokeSolver(subNLPProblemImpl, startingPoint); 

        if (solInfo.error) {

            return new SolutionImpl();
        } 
        else {
            
            TreeMap<String,Double> sol = subNLPProblemImpl.map(solInfo.solution, solutionOutParam);

            return new SolutionImpl(sol, solInfo.bestObj);
        }        
    }
    
    private Solution solveSubNLPProblemPrivate(SubNLPProblemImpl subProblemImpl,  double[] startingPoint) {

        SolutionInfo solInfo = invokeSolver(subProblemImpl, startingPoint); 

        if (solInfo.error) {

            return new SolutionImpl();
        } 
        else {

            log.debug("Local solver has converged, quality should be checked");
            
            TreeMap<String,Double> sol = subProblemImpl.map(solInfo.solution);
            
            double infeas = solInfo.bestObj; // In case of a CSP, IPOPT doesn't have objective, infinity norm is returned instead
           
            if (log.isTraceEnabled()) log.trace("Infeas: "+infeas+"  "+sol.toString());

            return new SolutionImpl(sol, infeas);
        }
    }
    
    private SolutionInfo invokeSolver(SubNLPProblemImpl subProblemImpl, double[] startingPoint) {
        
        RealVector startingPointInInternalPermutation = subProblemImpl.permuteFromAlphabeticalToInternalOrder(startingPoint);

        evalF = gradF = 0;

        try {
            
            //System.out.println("Subproblem\n"+subProblemImpl.getDagCSP().toString());
            //System.out.println("Start point");
            //System.out.println(subProblemImpl.map(startingPointInInternalPermutation));

            SolutionInfo solInfo = runSolver(subProblemImpl, startingPointInInternalPermutation);
            
            evalF = solInfo.funEval;

            gradF = solInfo.gradEval;
            
            updateCumulativeCounters(subProblemImpl);
            
            return solInfo;
        } 
        catch (Exception e) {

            e.printStackTrace();
            
            System.err.println("Subproblem\n"+subProblemImpl.getDagProblem().toString(false, false));
            System.err.println("Start point");
            System.err.println(subProblemImpl.map(startingPointInInternalPermutation));
            System.err.println("Solver option");
            System.err.println(solverOptionsToString());

            log.error("Unexpected error during subproblem solution");

            return SolutionInfo.createFailed();
        }
    }
    
    private void updateCumulativeCounters(SubNLPProblemImpl subProblemImpl) {
        
        cntF.add(  evalF);
        cntJac.add(gradF);
        
        nRows.add(subProblemImpl.getNumberOfConstraints());
        nCols.add(subProblemImpl.notFixedVars());
    }
    
    @Override 
    final public void dumpStatistics(String file) {
        
        try (BufferedWriter bw = new BufferedWriter(new FileWriter(file))) {
            
            bw.write("eval_F\teval_Jac\tn_rows\tn_cols\n");
            
            for (int i=0, n=cntF.size(); i<n; ++i)
                
                bw.write(cntF.get(i)+"\t"+cntJac.get(i)+"\t"+nRows.get(i)+"\t"+nCols.get(i)+"\n");
            
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
