//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.util.ArrayList;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.coinor.Ipopt;

import dag.Dag;
import dag.DagOptProgram;
import dag.Objective;
import tools.math.RealVector;
import tools.math.interval.real.RealIntervalVector;
import tools.opt.OptIpoptSolver;
import tools.opt.OptSolverOptions;
import tools.opt.ProblemSolutionStatus;
import tools.opt.SolverResult;
import tools.options.IllegalValueException;
import tools.options.Options;

final class SolverCSP extends SolverBase {
    
    private static final Logger log = LogManager.getLogger();
    
    OptIpoptSolver ipopt;
    
    SolverResult sr;
 
    SolverCSP(OptIpoptSolver ipopt) {

        this.ipopt = ipopt;
        Options globalOpts=OptIpoptSolver.getIpoptOptions();
        try {
          globalOpts.setValue(Ipopt.KEY_FILE_PRINT_LEVEL, 0);
          OptIpoptSolver.setIpoptOptions(globalOpts);
        } catch (IllegalValueException ex) {
          log.catching(ex);
        }
    }
    
    static private OptSolverOptions createSolverOptions() throws Exception {

        OptSolverOptions opts=new OptSolverOptions(log.isDebugEnabled() ? 1 : 0, 1e-5, 200, true, 0.0);
      
        if(log.isTraceEnabled()){
          log.debug(opts.toString(true));
        }
        
        return opts;
    }

    @Override
    protected SolutionInfo runSolver(SubNLPProblemImpl subProb,  RealVector startingPoint) throws Exception {

        OptSolverOptions opts = createSolverOptions();
        
        //boolean overDetermined = subProb.getNumberOfGuessedVars() > subProb.getNumberOfConstraints();
        
        Dag dag = subProb.getDagCSP();

        RealIntervalVector startBounds = dag.getVarBounds();

        ArrayList<Objective> objs = dag.getObjectives();

        Objective objective = null;
        
        if (objs.isEmpty()) {
            // That's fine, objective is null
        }        
        else if (objs.size() == 1) {

            objective=objs.get(0);
        } 
        else {

            throw new AssertionError("ipopt can only solve single objective problems");
        }
        
        DagOptProgram dop=new DagOptProgram(dag, objective, startBounds);

        //System.out.println("Starting point: "+subProb.map(startingPoint));
        //System.out.println("Starting point: "+Util.toHexString(subProb.map(startingPoint)));
        
        sr= ipopt.solve(dop, startingPoint, opts);
        
        //System.out.println("Solution: "+Util.toHexString(subProb.map(sr.getBest())));
        
        boolean error = processStatus();

        // TODO tell Feri to fix it, should return infinity norm constraint violation
        return new SolutionInfo(sr.getnConEval(),
                                Math.max(sr.getnJacEval()-2, 0), // HACK accounts for scaling (2 if gradient-based, 5 if MC19 is used)
                                error ? Double.NaN : Evaluator.infinityNormOfConstraintViolation(subProb.getDagCSP(),sr.getBest()),
                                sr.getBest(),
                                error);
    }
    
    private boolean processStatus() {
        
        ProblemSolutionStatus status = sr.getStatus();
        
        boolean error = false;

        switch (status) {
        case ACCURACY_ACHIEVED:
        case INFEASIBLE_PROBLEM:   // Infeasible problems are also OK 
        case MAX_GRAD_EVAL_REACHED:
            error = false;
            break;
        case EARLY_STOP:
        case LINE_SEARCH_STALLED:
        case RESTORATION_FAILED:
        case ERROR:
        default:
            log.warn("Ipopt failure: "+status);
            error = true;
            break;
        }

        logResults();

        log.debug("Error during solution? "+error);
        
        return error;
    }
    
    private void logResults() {
        
        if (log.isTraceEnabled()) {
            log.trace("  status: "+sr.getStatus());
            log.trace("  x=" + sr.getBest());
            log.trace("  fun evals=" + sr.getnConEval());
            log.trace("  grad evals=" + sr.getnJacEval());
            log.trace("  obj=" + sr.getBestF());
            //log.trace("  feas=" + SolutionChecker.computeGeneralConstraintViolation(
            //        subProb.getDagCSP(),new RealVector(sr.getBest())).getMax());            
        }
    }

    @Override
    protected String solverOptionsToString() {

        try {
            return createSolverOptions().toString();
        }
        catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }
    }
    
}
