//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import lmbopt.fun.DagProblem;

import org.apache.logging.log4j.*;

import dag.Dag;
import sdms.LatinHypercubeSampling;
import sdms.Solution;
import sdms.SubNLPProblem;
import sdms.Util;
import sdms.Violations;
import tools.math.IndexList;
import tools.math.RealVector;
import tools.math.interval.real.RealInterval;
import tools.math.interval.real.RealIntervalVector;

// TODO Setting / fixing / relaxing varbounds should use the correct
// functions of the DAG; the duplication from these set/fix/relax 
// functions should be removed

final class SubNLPProblemImpl implements SubNLPProblem {
    
    private static final Logger log = LogManager.getLogger();
    
    private final DagProblem optimizationProblem;
    
    private final Dag dagOpt;
    
    private final Dag CSPDag;
    
    private final TreeMap<String,String[]> fullProblemSparsity;

    /** Alphabetically ordered as constructed from a TreeSet */
    private ArrayList<String> guessedVars;
    
    private IndexList indicesOfGuessedVarsInDag;
    
    private TreeSet<String> selectedConstraints;
    
    private TreeSet<String> fixedVars;
    
    SubNLPProblemImpl(Dag asCSPProblem,
                      DagProblem asOptimizationProblem, 
                      SortedSet<String> selectedConstraints, 
                      SortedSet<String> fixedVars, 
                      TreeSet<String> varsToGuess,
                      TreeMap<String,String[]> fullProblemSparsityOnlyForDebugging)
    {
        CSPDag = asCSPProblem;
        
        optimizationProblem = asOptimizationProblem;
        
        dagOpt = optimizationProblem.getP().getModel();
        
        this.selectedConstraints = new TreeSet<>(selectedConstraints);
        
        this.fixedVars = new TreeSet<>(fixedVars);
        
        this.fullProblemSparsity = fullProblemSparsityOnlyForDebugging;
        
        setGuessedVarNames(varsToGuess);
    }
    
    SubNLPProblemImpl(Dag asCSPProblem,
                      DagProblem asOptimizationProblem, 
                      TreeSet<String> constraints, 
                      TreeMap<String,String[]> sparsityOnlyForDebugging)
    {
        CSPDag = asCSPProblem;
        optimizationProblem = asOptimizationProblem;
        dagOpt = optimizationProblem.getP().getModel();
        selectedConstraints = constraints;
        fullProblemSparsity = sparsityOnlyForDebugging;
        
        fixedVars                 = null;
        guessedVars               = null;
        indicesOfGuessedVarsInDag = null;
    }

    @Override
    public void setVarRoles(SortedSet<String> fixed, SortedSet<String> guessed) {
    
        fixedVars = new TreeSet<String>(fixed);
        
        setGuessedVarNames(guessed);
    }
    
    private void setGuessedVarNames(SortedSet<String> vars) {
        
        guessedVars = new ArrayList<String>(vars);
        
        indicesOfGuessedVarsInDag = dagOpt.getVarIndices(guessedVars);
       
        if (log.isDebugEnabled()) {
        
            dbgIndices();
        
            dbgSparsity();
        }
    }

    RealVector permuteFromAlphabeticalToInternalOrder(double[] values) {
        
        assert values.length == indicesOfGuessedVarsInDag.size() : "values size: "+values.length+", guessed: "+indicesOfGuessedVarsInDag.size();

        RealVector startPoint = dagOpt.getVarBounds().getMid(); // Keep fixed variables

        for (int i=0; i<values.length; ++i) {

            int index = indicesOfGuessedVarsInDag.get(i);

            startPoint.set(index, values[i]);  // Set only the not yet fixed variables
        }

        if (log.isDebugEnabled()) {
            log.debug("Guessed vars:   "+guessedVars);
            log.debug("Guessed values: "+Util.arrayToString(values));
            log.debug("All variables:  "+dagOpt.getVarNames());
            log.debug("Initial guess:  "+startPoint);
            log.debug("Bounds:         "+dagOpt.getVarBounds());
        }
        return startPoint;
    }

    /** Maps the solution vector back from DAG permutation to original var names */
    TreeMap<String, Double> map(RealVector x) {
        
        TreeMap<String, Double> result = new TreeMap<>();
        
        for (int i=0; i<indicesOfGuessedVarsInDag.size(); ++i) {
            
            result.put(guessedVars.get(i), x.get(indicesOfGuessedVarsInDag.get(i)));            
        }
        
        return result;
    }
    
    TreeMap<String, Double> map(RealVector x, double[] solutionOutParam) {
        
        assert solutionOutParam.length==indicesOfGuessedVarsInDag.size();
        
        TreeMap<String, Double> result = new TreeMap<>();
        
        for (int i=0; i<indicesOfGuessedVarsInDag.size(); ++i) {
            
            int index = indicesOfGuessedVarsInDag.get(i);
            
            double value = x.get(index);
            
            solutionOutParam[i] = value;
            
            result.put(guessedVars.get(i), value);            
        }
        
        return result;
    }
    
    @Override
    public double[] getLowerBoundsOfGuessedVars() {
        
        return lowerBnd(dagOpt, guessedVars);
    }

    @Override
    public double[] getUpperBoundsOfGuessedVars() {
        
        return upperBnd(dagOpt, guessedVars);
    }
    

    @Override
    public double[] lowerBoundsOf(SortedSet<String> names) {
        
        return lowerBnd(dagOpt, names);
    }

    @Override
    public double[] upperBoundsOf(SortedSet<String> names) {

        return upperBnd(dagOpt, names);
    }
    
    private static double[] lowerBnd(Dag dag, Collection<String> names) {
        
        double[] bnd = new double[names.size()];
        
        int k=0;
        
        for (String var : names) {
            
            bnd[k++] = dag.getVar(var).getBounds().getLower();
        }
        
        return bnd;
    }
    
    private static double[] upperBnd(Dag dag, Collection<String> names) {
        
        double[] bnd = new double[names.size()];
        
        int k=0;
        
        for (String var : names) {
            
            bnd[k++] = dag.getVar(var).getBounds().getUpper();
        }
        
        return bnd;
    }

    @Override
    public TreeMap<String,Double> getLowerBounds(SortedSet<String> names) {

        TreeMap<String,Double> lo = new TreeMap<>();

        for (String var : names) {

            lo.put(var, dagOpt.getVar(var).getBounds().getLower());
        }

        return lo;
    }

    @Override
    public TreeMap<String,Double> getUpperBounds(SortedSet<String> names) {
        
        TreeMap<String,Double> up = new TreeMap<>();

        for (String var : names) {

            up.put(var, dagOpt.getVar(var).getBounds().getUpper());
        }

        return up;
    }
    
    @Override
    public TreeMap<String,Double> allLowerBounds() {
        
        ArrayList<String> names = dagOpt.getVarNames();
        
        TreeMap<String,Double> bnds = new TreeMap<>();
        
        for (String name : names) {
            
            bnds.put(name, dagOpt.getVar(name).getBounds().getLower());
        }
        
        return bnds;
    }
    
    @Override
    public TreeMap<String,Double> allUpperBounds() {
        
        ArrayList<String> names = dagOpt.getVarNames();
        
        TreeMap<String,Double> bnds = new TreeMap<>();
        
        for (String name : names) {
            
            bnds.put(name, dagOpt.getVar(name).getBounds().getUpper());
        }
        
        return bnds;        
    }

    private void dbgIndices() {
        
        log.trace("Guessed vars:  "+guessedVars);
        log.trace("Indices in DAG "+indicesOfGuessedVarsInDag);
        
        TreeSet<Integer> uniqueIndices = new TreeSet<Integer>();
        
        for (int i : indicesOfGuessedVarsInDag) {
            
            if (!uniqueIndices.add(i)) {
                log.error("Index "+i+" found multiple times");
            }           
        }
     
        assert uniqueIndices.size()==guessedVars.size();

        ArrayList<String> guessedNamesList = CSPDag.getVarNames(indicesOfGuessedVarsInDag);
        
        TreeSet<String> uniqueNames = new TreeSet<>(guessedNamesList);
        
        assert uniqueNames.size()==guessedNamesList.size();
        
        assert guessedVars.size()==guessedNamesList.size();
        
        assert guessedVars.containsAll(uniqueNames);
     
        log.trace("Indices seem OK!");
    }
    
    private void dbgSparsity() {
        
        if (fullProblemSparsity==null) // patched subproblems no longer have the same sparsity pattern
            return;
        
        log.trace("Checking constraint dependecies based on AMPL and DAG sparsity patterns");
        
        TreeSet<String> constraintDependencies = new TreeSet<>();
        
        for (String con : selectedConstraints) {
            
            String[]        varsAmpl = fullProblemSparsity.get(con);
            
            log.trace("Checking "+con+" sparsity");
            
            TreeSet<String> varsInDag = CSPDag.getConstraintDependingOn(CSPDag.getConstraint(con));
            
            assert varsInDag.containsAll(Arrays.asList(varsAmpl)) : 
                "Constraint "+con+" depends on vars in dag: "+varsInDag+", in AMPL "+Arrays.toString(varsAmpl);
            
            assert varsInDag.size()==varsAmpl.length;
            
            constraintDependencies.addAll(varsInDag);           
        }
        
        assert constraintDependencies.containsAll(guessedVars);
        
        boolean modified = constraintDependencies.removeAll(guessedVars); 
        
        assert modified;
        
        assert fixedVars.equals(constraintDependencies) : 
            "fixedVars = "+fixedVars+"\nguessedVars = "+guessedVars+"\nconstraintDependecies = "+constraintDependencies;
        
        log.trace("Subproblem sparsity seems OK!");
    }

    DagProblem getDagProblem() {
        
        return optimizationProblem;
    }
    
    Dag getDagCSP() {
        
        return CSPDag;
    }

    @Override
    public int getNumberOfGuessedVars() {
        
        return guessedVars.size();
    }

    @Override
    public int getNumberOfConstraints() {
        
        return selectedConstraints.size();
    }

    @Override
    public void fixVars(TreeMap<String,Double> x) {
        
        for (Map.Entry<String,Double> e : x.entrySet()) {
            
            String name  = e.getKey();
            
            RealInterval fixAt = new RealInterval( e.getValue() );
            
            CSPDag.getVar(name).setBounds(fixAt);
            
            dagOpt.getVar(name).setBounds(fixAt);
        }
        
        Iterator<Double> itr = x.values().iterator();
        
        for (int i : dagOpt.getVarIndices(x.keySet())) {
            
            RealInterval fixAt = new RealInterval( itr.next() );            
            
            optimizationProblem.getVarBounds().set(i, fixAt);
        }       
    }
    
    @Override
    public void setBounds(TreeMap<String,Double> lo, TreeMap<String,Double> up) {
        
        assert lo.size()==up.size();
        
        Iterator<Entry<String,Double>> itr = up.entrySet().iterator();
        
        for (Map.Entry<String,Double> lb : lo.entrySet()) {
            
            Entry<String,Double> ub = itr.next();
            
            String name = lb.getKey();
            
            assert name.equals(ub.getKey());
            
            RealInterval bnd = new RealInterval(lb.getValue(), ub.getValue());
            
            CSPDag.getVar(name).setBounds(bnd);
            
            dagOpt.getVar(name).setBounds(bnd);
        }
        
        Iterator<Double> lb = lo.values().iterator();
        
        Iterator<Double> ub = up.values().iterator();
        
        for (int i : dagOpt.getVarIndices(lo.keySet())) {
            
            RealInterval bnd = new RealInterval( lb.next(), ub.next() );            
            
            optimizationProblem.getVarBounds().set(i, bnd);
        } 
    }
    
    int notFixedVars() {
        
        RealIntervalVector bnds = CSPDag.getVarBounds();
        
        int n = bnds.getNNZ();
        
        for (RealInterval i : bnds.getData().values()) {
            
            if (Double.compare(i.getLower(),i.getUpper())==0) 
                --n;
        }
        
        return n;
    }
    
    /** [l, u] --> [l-c*(u-l), u+c*(u-l)] */
    @Override
    public void relaxBounds(SortedSet<String> names, double[] lb, double[] ub, double c) {
        
        assert names.size() == lb.length : "Names size: "+names.size()+", "+lb.length;
        assert lb.length==ub.length : "Length mismatch: "+lb.length+" and "+ub.length;
        
        int k=0;
        
        for (String name : names) {
            
            double l = lb[k];
            
            double u = ub[k];
            
            ++k;
            
            double diam = u - l;
            
            RealInterval newBounds = new RealInterval(l-c*diam, u+c*diam);

            CSPDag.getVar(name).setBounds(newBounds);
            
            dagOpt.getVar(name).setBounds(newBounds);
        }
        
        IndexList il = dagOpt.getVarIndices(names);
        
        k = 0;
        
        for (int i : il) {
            
            double l = lb[k];
            
            double u = ub[k];
            
            ++k;
            
            double diam = u - l;
            
            RealInterval newBounds = new RealInterval(l-c*diam, u+c*diam);       
            
            optimizationProblem.getVarBounds().set(i, newBounds);
        } 
    }
    
    @Override
    public void dbgAllVarsCovered(SortedSet<String> x, SortedSet<String> y) {

        dbgAllVarsCovered(dagOpt.getVarNames(),  x, y);
    }
    
    private static void dbgAllVarsCovered(ArrayList<String> names, SortedSet<String> x, SortedSet<String> y) {
        
        SortedSet<String> allNames = new TreeSet<>();
        
        for (String name : names) {
            
            boolean added = allNames.add(name);
            
            assert added : "Variable found twice: "+name;
        }
        
        for (String xName : x) {
            
            boolean removed = allNames.remove(xName);
            
            assert removed : "Variable not found: "+xName;
        }
        
        for (String yName : y) {
            
            boolean removed = allNames.remove(yName);
            
            assert removed : "Variable not found: "+yName;
        }
        
        // although implicitly, overlap between x and y is also checked at this point
        
        assert allNames.isEmpty() : "Variables not covered: "+allNames;
    }
    
    @Override
    public TreeMap<String,Double> dbgResiduals(TreeMap<String,Double> x) {
        
        return Evaluator.evaluateResidual(CSPDag, x);
    }
    
    @Override
    public String toString() {
        
        StringBuilder sb = new StringBuilder("As CSP:\n");
        sb.append(CSPDag.toFullString());
        sb.append("\n");
        sb.append("As optimization problem:\n");
        sb.append(dagOpt.toFullString());
        
        return sb.toString();
    }
    
    //-----------------------------------------------------------------------------------------------------------
    
    @Override
    public ArrayList<TreeMap<String, Double>> sampleUnderDeterminedWithQRP(Random rnd, 
                                                                           SortedSet<String> vars, 
                                                                           TreeMap<String,Double> rowScales, 
                                                                           TreeMap<String,Double> colScales, 
                                                                           int m, 
                                                                           int toFix) 
    {
        return Evaluator.sampleUnderDeterminedWithQR(CSPDag, new ArrayList<>(vars), rowScales, colScales, rnd, m, toFix);
    }
    
    @Override
    public ArrayList<TreeMap<String,Double>> findLeastInfluentialWithQRP(TreeMap<String,Double> rowScales, 
                                                                         TreeMap<String,Double> columnScales, 
                                                                         TreeMap<String,Double> xPart, 
                                                                         TreeSet<String> yVars,
                                                                         int m, 
                                                                         TreeSet<String> yLast,
                                                                         int nToFix,
                                                                         Random rnd ) 
    {
        double[] lb = lowerBoundsOf(yVars);
        
        double[] ub = upperBoundsOf(yVars);
        
        ArrayList<double[]> samples = LatinHypercubeSampling.generateLHSamples(rnd, m, lb, ub);
        
        return Evaluator.findLeastInfluentialWithQRP(CSPDag, rowScales, columnScales, xPart, new ArrayList<>(yVars), samples, yLast, nToFix);
    }
    
    @Override
    public Violations allViolations(TreeMap<String,Double> x, TreeMap<String,Double> rowScales, TreeMap<String,Double> colScales) {
        
        return Evaluator.allConstraintAndVarboundViolations(CSPDag, x, rowScales, colScales);
    }

    @Override
    public ArrayList<Solution> denseNewtonIterate(ArrayList<Solution> allPartialSolutions) {
        
        double[] lb = lowerBnd(CSPDag, allPartialSolutions.get(0).getVarNames());
        double[] ub = upperBnd(CSPDag, allPartialSolutions.get(0).getVarNames());        
        
        return Evaluator.applyDenseNewton(CSPDag, lb, ub, allPartialSolutions);
    }
    
    @Override
    public void showConditionNumbers(ArrayList<Solution> solutions) {
        
        Evaluator.showConditionNumbers(CSPDag, solutions);
    }
}

class Sample {

    double[] x;
    double   f;

    Sample(double[] x, double f) { 
        this.x=x; 
        this.f=f; 
    }

    static class ByFunctionValue implements Comparator<Sample> {
        @Override
        public int compare(Sample a, Sample b) {
            return Double.compare(a.f, b.f);
        }
    }
}

class ValueName {

    double gradValue;
    String name;
    double xValue;

    ValueName(Double gradValue, String name, double xValue) {
        this.gradValue = gradValue; 
        this.name = name;
        this.xValue = xValue;
    }

    static class ByAbsGradVal implements Comparator<ValueName> {
        @Override
        public int compare(ValueName a, ValueName b) {
            return Double.compare(Math.abs(a.gradValue), Math.abs(b.gradValue));
        }
    }
}
