//  Copyright (C) University of Vienna, 2013-2015
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import sdms.ProgramTerminationRequested;
import sdms.SubproblemSelector;

final class SubproblemSelectorImpl implements SubproblemSelector {
    
    static final Logger log = LogManager.getLogger();
    
    int subProblemIndex  = -1;
    int numberOfSubprobs = 0;
    
    ArrayList<String> allVariables;
    
    ArrayList<ArrayList<String>> subproblemGuessedVars;
    ArrayList<ArrayList<String>> subproblemFixedVars;    
    
    ArrayList<String> allConstraints;
    ArrayList<String[]> subproblemCons;
    
    TreeSet<String> varsToGuess;
    TreeSet<String> fixedVars;
    TreeSet<String> selectedCons;
    
    ArrayList<String[]> sparsity;
    TreeMap<String,String[]> sparsityMap;
   
    SubproblemSelectorImpl(ArrayList<String>   allVariables, 
                           ArrayList<String[]> subproblemVars,
                           ArrayList<String>   allConstraints,    
                           ArrayList<String[]> subproblemCons,
                           ArrayList<String[]> sparsity) 
    {    
        int sizeSubprobVar = subproblemVars.size();
        int sizeSubprobCon = subproblemCons.size();
        
        assert sizeSubprobVar!=0 : "there should be at least one subproblem";
        assert sizeSubprobCon!=0 : "there should be at least one subproblem";

        numberOfSubprobs = sizeSubprobVar;
        
        this.allVariables   = allVariables;
        this.allConstraints = allConstraints;
        this.subproblemCons = subproblemCons;
        
        int nCons = allConstraints.size();
        
        this.sparsity = sparsity;
        
        sparsityMap = new TreeMap<>();
        
        for (int i=0; i<nCons; ++i) {
            
            sparsityMap.put(allConstraints.get(i), sparsity.get(i));
        }
        
        initalizeVars(subproblemVars);
    }
    
    private void initalizeVars(ArrayList<String[]> subproblemVars) {

        int[] firstGuessedVarIndex = initializeFirstGuessedVarIndices(subproblemVars);
        
        subproblemFixedVars   = new ArrayList<ArrayList<String>>();

        subproblemGuessedVars = new ArrayList<ArrayList<String>>();
 
        for (int subProb=0; subProb<numberOfSubprobs; ++subProb) {

            String[] vars = subproblemVars.get(subProb); 

            int firstIndex = firstGuessedVarIndex[subProb];

            ArrayList<String> fixedVars = new ArrayList<String>(Arrays.asList(vars).subList(1, firstIndex));

            subproblemFixedVars.add(fixedVars);

            ArrayList<String> guessedVars = new ArrayList<String>(Arrays.asList(vars).subList(firstIndex, vars.length));

            subproblemGuessedVars.add(guessedVars);
        }
    }
    
    private int[] initializeFirstGuessedVarIndices(ArrayList<String[]> subproblemVars) {
        
        int[] firstGuessedVarIndex = new int[numberOfSubprobs];

        int subProb = 0;

        for (String[] vars : subproblemVars) {

            try {

                firstGuessedVarIndex[subProb++] = Integer.valueOf(vars[0]) + 1;
            }
            catch (NumberFormatException e) {

                log.fatal("Failed to parse \""+vars[0]+"\" as integer at subproblem "+subProblemIndex);

                throw new ProgramTerminationRequested();
            }
        }
        
        if (firstGuessedVarIndex[0]!=0) {
            
            log.warn("Forcing zero as first guessed var index at the first subproblem and ignoring the original "+firstGuessedVarIndex[0]);
            
            firstGuessedVarIndex[0] = 1; // must be off by one, see subList() calls in initializeVars()
        }
        
        return firstGuessedVarIndex;
    }
    
    @Override
    final public boolean isLastSubproblem() { 
        
        return subProblemIndex >= numberOfSubprobs - 1; 
    }
    
    @Override
    final public boolean setNext() {
        
        boolean hasNext = ++subProblemIndex < numberOfSubprobs;
        
        if (hasNext) {
            
            log.info("Subproblem "+(subProblemIndex+1)+" of "+numberOfSubprobs);
            
            setNextVars();

            setNextCons();
            
            if (!isLastSubproblem()) {// TODO The last subproblem size should be shown somewhere! Polymorphism?
                log.info("Rows: "+selectedCons);
                log.info("Cols: "+varsToGuess);
                log.info("Fixed: "+fixedVars);
                log.info("Subproblem size "+(selectedCons.size())+"x"+(varsToGuess.size()));
            }
        }
        else {
            
            log.info("No more subproblems");
        }
                
        return hasNext;
    }
 
    @Override
    final public TreeSet<String> getFirstVariables() {

        // TODO In the first block, the fixed vars have different meaning then the others.
        //      Revise this as it is confusing.
        return new TreeSet<>(subproblemFixedVars.get(0));
    }

    @Override
    final public TreeSet<String> getNextConstraints() {
        
        return selectedCons;
    }
 
    @Override
    final public TreeSet<String> getVariablesToGuess() {
        
        return varsToGuess;
    }
    
    @Override
    final public TreeSet<String> getFixedVars() {
        
        return fixedVars;
    }

    @Override
    final public TreeMap<String, String[]> dbgFullProbSparsity() {

        return sparsityMap;
    }

    @Override
    final public void setFullProblem() {
        
        selectedCons = new TreeSet<>(allConstraints);
        
        fixedVars = new TreeSet<>();
        
        varsToGuess = new TreeSet<>(allVariables);
    }
    
    //---------------------------------------------------------------
    // Moved from implementation class after cleanup
    
    TreeSet<String> consProcessedSoFar = new TreeSet<>();
    
    private void setNextVars() {
       
        fixedVars = new TreeSet<String>(subproblemFixedVars.get(subProblemIndex));
        
        varsToGuess = new TreeSet<String>(subproblemGuessedVars.get(subProblemIndex));
 
        log.debug("Fixed variables:   "+fixedVars);
        
        log.debug("Guessed variables: "+varsToGuess);
    }

    private void setNextCons() {

        selectedCons = new TreeSet<>(Arrays.asList(subproblemCons.get(subProblemIndex)));
        
        consProcessedSoFar.addAll(selectedCons);
    }
    
    @Override
    final public TreeSet<String> getProcessedConstraintsSoFar() {
        
        return consProcessedSoFar;
    }
    
    private int getBegin(int maxStepLength) {

        assert maxStepLength > 0;
       
        return Math.max(0, subProblemIndex - maxStepLength); 
    }
    
    @Override    
    final public TreeSet<String> allConsBetweenCurrentAndPast(int maxStepLength) {

        TreeSet<String> result = new TreeSet<>(); 
        
        for (int i=getBegin(maxStepLength); i<=subProblemIndex; ++i) {
            
            result.addAll(Arrays.asList(subproblemCons.get(i)));
        }

        return result;  
    }

    @Override
    final public boolean hasPrecedingSubproblem(int maxStepLength) {

        return subProblemIndex - maxStepLength > 0;
    }
    
    @Override
    final public TreeSet<String> xVarsBetweenCurrentAndPast(int maxStepLength) {

        TreeSet<String> result = new TreeSet<>();

        for (int i=getBegin(maxStepLength); i<=subProblemIndex; ++i) {

            result.addAll(subproblemFixedVars.get(i));
        }

        result.removeAll(yVarsBetweenCurrentAndPast(maxStepLength));
        
        assert !result.isEmpty();
        
        return result;
    }
    
    @Override
    final public TreeSet<String> yVarsBetweenCurrentAndPast(int maxStepLength) {
        
        TreeSet<String> result = new TreeSet<>();
        
        for (int i=getBegin(maxStepLength); i<=subProblemIndex; ++i) {
            
            result.addAll(subproblemGuessedVars.get(i));
        }

        return result;
    }    
}
