//  Copyright (C) University of Vienna, 2013-2016
//  All rights reserved.
//
//  Author: Ali Baharev
//
//  Please refer to the LICENSE.txt file in the top directory
//
package sdms.impl;

import dag.Constraint;
import dag.Dag;
import dag.DagFactory;
import dag.DagOptProgram;
import dag.eval.interval.ConstraintIntervalEvalInfo;
import dag.eval.interval.DagIntervalEvalResult;
import dag.methods.relax.DagRelaxations;
import dag.nodes.input.ParamNode;
import dag.nodes.input.VarNode;

import java.util.ArrayList;

import problem.Problem;
import problem.searchspace.SearchPoint;
import tools.math.RealMatrix;
import tools.math.RealVector;
import tools.math.interval.real.IMAMath;
import tools.math.interval.real.RealInterval;
import tools.math.interval.real.RealIntervalMatrix;
import tools.math.interval.real.RealIntervalVector;
import tools.opt.OptIpoptSolver;
import tools.opt.OptSolverOptions;
import tools.opt.SolverResult;
import tools.opt.lp.IntervalLinearProgram;

class Wishlist {

    static { ClassLoader.getSystemClassLoader().setDefaultAssertionStatus(true); }    

    private Wishlist() { }
    
    public static void main(String[] args) throws Throwable {
        
        String file = "data/zzz_misc/infinityInValue.mod";
        
        Sdms.setProblemCheck(false, false, true, 1e-6);
        
        Problem prob = Sdms.readProblem(file, false);
        
        System.out.println(prob);
        
        Dag dagCSP = prob.getModel();
        
        //-----------------------------------------------------------------------------------------------------------------------------
        
        String[] varNames = new String[]{ "x_1_18_", "x_1_19_", "x_1_20_", "x_2_19_", "x_3_19_", "T_19_", "x_2_18_", "x_2_20_", "x_3_18_", "x_3_20_" };
        
        int[] indices = new int[varNames.length];
        
        int k=0;
        
        for (String name : varNames) {
            
            indices[k++] = dagCSP.getVar(name).getIndex();
        }

        RealVector startingPoint = new RealVector(varNames.length);

        double[] x = new double[] {
                0.07707386234781144,
                0.1796277195151431,
                0.3855369954215693,
                0.42581536689465926,
                0.8716470847886628,
                344.6267815882165,
                0.2689514850633305,
                0.17031109009390644,
                0.9974659911538793,
                0.3448175279769417   };
        
        k=0;
        for (int i : indices) {
            
            startingPoint.set(i, x[k++]);    
        }
        
        //----------------------------------------------------------------------------------------------------------------------------
        //RealIntervalVector xx = new RealIntervalVector(x);

        Dag dag=prob.getModel();
        RealIntervalVector xb=dag.getVarBounds();
        SolverResult solverRes = OptIpoptSolver.getInstance().solve(
                new DagOptProgram(dag,
                        dag.hasObjective() ? dag.getObjectives().get(0) : null,
                                xb), new RealVector(x), new OptSolverOptions(1, 1e-5, 500, false, 0.0));
        //RealVector best=solverRes.getBest();
        if (solverRes.isFeasible()) {
            //System.err.print("Infeas:"+best);
        }else{
        }
        
        double[] bad = new double[] {
                0.07707386234781144,
                1.8648296726872E-17,
                0.3855369954215693,
                -0.31651405741779903, 
                0.8716470847886628,
                344.6267815882165,
                0.2689514850633305,
                21.84961462310037, 
                0.9974659911538793,
                28.498794350554068  };

        DagIntervalEvalResult evalRes = dagCSP.evalRan(new RealIntervalVector(bad), dagCSP.computeParamBounds(new RealIntervalVector(bad), ParamNode.PARAM_EVAL_TYPE_LOCSOL));
        
        System.out.println(evalRes);
        
        for (ConstraintIntervalEvalInfo cei : evalRes.values()) {

            RealInterval leftHandSide = cei.getRangeHull();
            
            RealInterval rightHandSide = cei.getConstraint().getBounds();

            ASSERT_NARROW(cei.getConstraint(), leftHandSide);
            
            ASSERT_NARROW(cei.getConstraint(), rightHandSide);
        }

        return;
    }
    
    static private void ASSERT_NARROW(Constraint c, RealInterval x) {
        
        Double xL = x.getLower();
        
        Double xU = x.getUpper();

        assert (-Double.MAX_VALUE < xL) && (xU < Double.MAX_VALUE);
        
        assert xL <= xU;
        
        double absWidth = xU - xL;
        
        double relWidth = absWidth / ((1e-300) + Math.max(Math.abs(xL), Math.abs(xU)));
        
        double width = Math.min(relWidth, absWidth);
        
        if (width >= 1.0e-2) {
            System.err.println("Catastrophic loss of precision in the "
                    + "constraint '"+c.getLabel()+"': ["+String.format("%g", x.getLower())+", "+String.format("%g", x.getUpper())+"]  ("+width+")");
//            throw new AssertionError();
        }
    }

    @SuppressWarnings("unused")
    public static void mainOld(String[] args) throws Throwable {
        
        String file = "data/dummy/dummy.mod";
        
        // The dummy example is a 2D manifold in 4D, given implicitly by the equations:
        // exp(x1) / exp(x2) = 2;
        // x1*x2*x3*x4       = 0.01;
        
        Sdms.setProblemCheck(false, false, true, 1e-6);
        
        Problem prob = Sdms.readProblem(file, false);
        System.out.println(prob);
        
        /*
          computeLinearRelaxation(Dag dag,
          TreeMap<Constraint, RealInterval> toRelax, RealIntervalVector x,
          RealVector z, boolean centered, boolean intervalCoeffs,
          boolean normalizeRelations, boolean debug)
         */
        IntervalLinearProgram ilp= DagRelaxations.computeLinearRelaxation(
                prob.getModel(),  null, prob.getModel().getVarBounds(), null, 
                false, false, false, false);
        System.out.println("* RELAXATION *");
        System.out.println(ilp);
        
        
        //----------------------------------------------------------------------------------
        // Please provide a function that can be used to append linear constraints
        // c_ij are the coefficients (real numbers) coming as a matrix / vectors (e.g. from QR factorization)
        // v_i are (most likely new) variables, potentially with bound constraints
        //
        // x1 = c11*v1 + c12*v2 + c13
        // x2 = c21*v1 + c22*v2 + c23
        // x3 = c31*v1 + c32*v2 + c33
        // x4 = c41*v1 + c42*v2 + c43
        //
        // The purpose of this is that the 2D manifold (dummy.mod) will be described in terms of (v1, v2)
        
        // With GRAPH_REF? Or deep copy?
        DagFactory df=new DagFactory(prob.getModel());
        df.fullCopy(true); //deep dag copy ?
        
        RealMatrix E=new RealMatrix(new double[][]{{1,2,3},{0,0,0}});
        RealVector c=new RealVector(new double[]{-1,3});
        RealVector xref_back=new RealVector(3.5,-1,11,-3.3);
        Dag dag=df.getResult();
        ArrayList<VarNode> x=new ArrayList<>();
        x.add(dag.getVar("x3"));
        x.add(dag.getVar("x1"));
        ArrayList<VarNode> v=new ArrayList<>();
        v.add(dag.addVar("v1", new RealInterval(0,3), "artifical variable"));
        v.add(dag.addVar("v2", new RealInterval(-1,0.5), "artifical variable"));
        v.add(dag.getVar("x4"));
        df.addLinearConstraints(new RealIntervalMatrix(E), new RealIntervalVector(c), 
                x, v, "linc", "artificial linear constraint");
        System.out.println("* LINEAR CONSTRAINTS ADDED *");
        System.out.println(dag);

        
        //------------------------------------------------------------------------------------
        // We may want to backsubstitute these linear constraints, that is, all occurrence of 
        // x_i should be replaced with the corresponding linear constraints
        // Please provide a high-level function for that
        
        Problem pnew=new Problem(dag);
        pnew.backSubstitution(new String[]{"x1","x3"},new String[]{"linc_"});
        System.out.println("* BACKSUBST *");
        System.out.println(dag);
        System.out.println("* TEST BACKSUBST *");
        System.out.println("x="+xref_back);
        SearchPoint sp=new SearchPoint(xref_back);
        sp.check(dag, 0);
        DagIntervalEvalResult res=dag.evalRan(new RealIntervalVector(xref_back), 
                dag.computeParamBounds(new RealIntervalVector(xref_back), 
                ParamNode.PARAM_EVAL_TYPE_LOCSOL));
        RealIntervalVector fx=res.getRangeHull(dag.getConstraints());
        System.out.println("f(x)="+fx);
        System.out.println("f(x)-\\F="+
                IMAMath.dist(dag.getConstraintBounds(),
                res.getRangeHull(dag.getConstraints()).getMid(), true));
        System.out.println("Matlab check:\n"+dag.toMatlabCheckString(xref_back));
        
        
        //-----------------------------------------------------------------------------------
        // Unrelated to the above: please provide a high-level function for overriding or disabling 
        // variable bounds.
        //
        // Say, for example I would like to fix x2 to 1.5, I would like to do something like:
        // csp.setVarBounds("x2", 1.5, 1.5);  // or
        // csp.setVarBounds(indexOfx2InDag, 1.5, 1.5)
        //
        // similarly, x1 has initially 0 as lower and 1 as upper bound
        // and I would like to drop the upper bound and keep only the lower bound
        // but change it to 0.5
        //
        // csp.setVarBound("x1", 0.5, NO_UPPER_BOUND); // or 
        // csp.setVarBound(indexOfx1InDag, 0.5, NO_UPPER_BOUND);
        
        dag.getVar("x2").setBounds(new RealInterval(1.5,1.5));
        dag.getVar("x4").setBounds(new RealInterval(0.5,Double.POSITIVE_INFINITY));
        System.out.println("* CHANGE VAR BOUNDS *");
        System.out.println(dag);
    }

}
